From stdpp Require Export gmultiset strings.
From iris.algebra Require Import frac.
From iris.base_logic.lib Require Export invariants.
From iris.bi Require Import fractional.
From orc11 Require Import base.
From gpfsl Require Export vprop.
Set Default Proof Using "Type".

Definition lftN : namespace := nroot .@ "lft".

Implicit Types (Lat : latticeT) (E : coPset).

Module Type lifetime_sig.
  (** CMRAs needed by the lifetime logic  *)
  (* We can't instantie the module parameters with inductive types, so we
     have aliases here. *)
  Parameter lftG' : ∀ Lat E0, gFunctors → Set.
  Global Notation lftG := lftG'.
  Existing Class lftG'.
  Parameter lftPreG' : ∀ Lat, gFunctors → Set.
  Global Notation lftPreG := lftPreG'.
  Existing Class lftPreG'.

  (** Definitions *)
  Parameter lft : Type.
  Parameter static : lft.
  Global Declare Instance lft_intersect : Meet lft.

  Section defs.
  Context {Σ : gFunctors} {Lat}.
  Notation monPred := (monPredSI (lat_bi_index Lat) (uPredSI (iResUR Σ))).

  Parameter lft_ctx : ∀ `{invG Σ, lftG Lat E0 Σ}, iProp Σ.

  Parameter lft_tok : ∀ `{lftG Lat E0 Σ} (q : Qp) (κ : lft), monPred.
  Parameter lft_dead : ∀ `{lftG Lat E0 Σ} (κ : lft), monPred.

  Parameter lft_incl : ∀ `{invG Σ, lftG Lat E0 Σ} (κ κ' : lft), monPred.
  Parameter bor : ∀ `{invG Σ, lftG Lat E0 Σ} (κ : lft) (P : monPred), monPred.

  Parameter bor_idx : Type.
  Parameter idx_bor_own : ∀ `{lftG Lat E0 Σ} (i : bor_idx), monPred.
  Parameter idx_bor :
    ∀ `{invG Σ, lftG Lat E0 Σ} (κ : lft) (i : bor_idx) (P : monPred), monPred.

  Parameter in_at_bor : ∀ `{invG Σ, lftG Lat E0 Σ} (κ : lft) (P : monPred), monPred.
  End defs.

  (** Notation *)
  Notation "q .[ κ ]" := (lft_tok q κ)
      (format "q .[ κ ]", at level 0) : bi_scope.
  Notation "[† κ ]" := (lft_dead κ) (format "[† κ ]"): bi_scope.

  Notation "&{ κ }" := (bor κ) (format "&{ κ }") : bi_scope.
  Notation "&{ κ , i }" := (idx_bor κ i) (format "&{ κ , i }") : bi_scope.
  Notation "&in_at{ κ }" := (in_at_bor κ) (format "&in_at{ κ }") : bi_scope.

  Infix "⊑" := lft_incl (at level 70) : bi_scope.

  Section properties.
  Context `{invG Σ, lftG Lat E0 Σ}.

  (** Instances *)
  Global Declare Instance lft_inhabited : Inhabited lft.
  Global Declare Instance bor_idx_inhabited : Inhabited bor_idx.

  Global Declare Instance lft_intersect_comm : Comm eq lft_intersect.
  Global Declare Instance lft_intersect_assoc : Assoc eq lft_intersect.
  Global Declare Instance lft_intersect_inj_1 κ : Inj eq eq (lft_intersect κ).
  Global Declare Instance lft_intersect_inj_2 κ : Inj eq eq (λ κ', lft_intersect κ' κ).
  Global Declare Instance lft_intersect_left_id : LeftId eq static lft_intersect.
  Global Declare Instance lft_intersect_right_id : RightId eq static lft_intersect.

  Global Declare Instance lft_ctx_persistent : Persistent lft_ctx.
  Global Declare Instance lft_dead_persistent κ : Persistent [†κ].
  Global Declare Instance lft_incl_persistent κ κ' : Persistent (κ ⊑ κ').
  Global Declare Instance idx_bor_persistent κ i P : Persistent (&{κ, i}P).
  Global Declare Instance in_at_bor_persistent : Persistent (&in_at{κ}P).

  Global Declare Instance lft_tok_timeless q κ : Timeless q.[κ].
  Global Declare Instance lft_dead_timeless κ : Timeless [†κ].
  Global Declare Instance idx_bor_own_timeless i : Timeless (idx_bor_own i).

  Global Declare Instance idx_bor_params : Params (@idx_bor) 6.
  Global Declare Instance bor_params : Params (@bor) 5.
  Global Declare Instance in_at_bor_params : Params (@in_at_bor) 5.

  Global Declare Instance bor_ne κ n : Proper (dist n ==> dist n) (bor κ).
  Global Declare Instance bor_contractive κ : Contractive (bor κ).
  Global Declare Instance bor_proper κ : Proper ((≡) ==> (≡)) (bor κ).
  Global Declare Instance idx_bor_ne κ i n : Proper (dist n ==> dist n) (idx_bor κ i).
  Global Declare Instance idx_bor_contractive κ i : Contractive (idx_bor κ i).
  Global Declare Instance idx_bor_proper κ i : Proper ((≡) ==> (≡)) (idx_bor κ i).
  Global Declare Instance in_at_bor_ne κ n : Proper (dist n ==> dist n) (in_at_bor κ).
  Global Declare Instance in_at_bor_contractive κ : Contractive (in_at_bor κ).
  Global Declare Instance in_at_bor_proper : Proper ((≡) ==> (≡)) (in_at_bor κ).

  Global Declare Instance lft_tok_fractional κ : Fractional (λ q, q.[κ])%I.
  Global Declare Instance lft_tok_as_fractional κ q :
    AsFractional q.[κ] (λ q, q.[κ])%I q.

  (** Laws *)
  Parameter lft_tok_sep : ∀ q κ1 κ2, q.[κ1] ∗ q.[κ2] ⊣⊢ q.[κ1 ⊓ κ2].
  Parameter lft_tok_split_obj :
    ∀ `{LatBottom Lat} q1 q2 κ, (q1 + q2).[κ] -∗ q1.[κ] ∗ <obj> q2.[κ].
  Parameter lft_dead_or : ∀ κ1 κ2, [†κ1] ∨ [†κ2] ⊣⊢ [† κ1 ⊓ κ2].
  Parameter lft_tok_dead : ∀ q κ, q.[κ] -∗ [† κ] -∗ False.
  Parameter lft_tok_static : ∀ q, (q.[static])%I.
  Parameter lft_dead_static : [† static] -∗ False.

  Parameter lft_create : ∀ `{LatBottom Lat} E, ↑lftN ⊆ E →
    ⎡lft_ctx⎤ ={E}=∗ ∃ κ, 1.[κ] ∗ □ (1.[κ] ={↑lftN ∪ E0, E0}▷=∗ [†κ]).
  Parameter bor_create : ∀ `{LatBottom Lat} E κ P,
    ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ ▷ P ={E}=∗ &{κ}P ∗ ([†κ] ={E}=∗ ▷ P).
  Parameter bor_fake : ∀ `{LatBottom Lat} E κ P,
    ↑lftN  ⊆ E → ⎡lft_ctx⎤ -∗ <subj>[†κ] ={E}=∗ &{κ}P.

  Parameter bor_sep : ∀ `{LatBottom Lat} E κ P Q, ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗
    &{κ}(P ∗ Q) ={E}=∗ &{κ}P ∗ &{κ}Q.
  Parameter bor_combine : ∀ `{LatBottom Lat} E κ P Q,
    ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ &{κ}P -∗ &{κ}Q ={E}=∗  &{κ}(P ∗ Q).

  Parameter bor_unfold_idx : ∀ κ P, (&{κ}P)%I ⊣⊢ ∃ i, (&{κ,i}P)%I ∗ (idx_bor_own i).

  Parameter idx_bor_shorten : ∀ κ κ' i P, κ ⊑ κ' -∗ &{κ',i}P -∗ &{κ,i}P.
  Parameter idx_bor_iff : ∀ κ i P P', ▷ □ (P ↔ P') -∗ &{κ,i}P -∗ &{κ,i}P'.

  Parameter idx_bor_unnest : ∀ `{LatBottom Lat} E κ κ' i P,
    ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ &{κ,i}P -∗ &{κ'}(idx_bor_own i) ={E}=∗ &{κ ⊓ κ'}P.
  Parameter in_at_bor_shorten : ∀ κ κ' P, κ ⊑ κ' -∗ &in_at{κ'}P -∗ &in_at{κ}P.

  Parameter in_at_bor_iff : ∀ κ P P',
    ▷ □ (P ↔ P') -∗ &in_at{κ}P -∗ &in_at{κ}P'.
  Parameter in_at_bor_share : ∀ E κ P, &{κ}P ={E}=∗ &in_at{κ}P.

  Parameter idx_bor_acc : ∀ E q κ i P Vtok Vbor, ↑lftN ⊆ E →
    lft_ctx -∗ &{κ,i}P Vtok -∗ idx_bor_own i Vbor -∗ q.[κ] Vtok ={E}=∗
      ∃ V, ⌜V ⊑ Vbor ⊔ Vtok⌝ ∗ ▷ P V ∗
        (∀ Vtok', ⌜Vtok ⊑ Vtok'⌝ -∗ ▷ P (Vtok' ⊔ V)
                  ={E}=∗ idx_bor_own i (Vtok' ⊔ V) ∗ q.[κ] Vtok').
  Parameter bor_acc_strong : ∀ E q κ P, ↑lftN ⊆ E →
    ⎡lft_ctx⎤ -∗ &{κ}P -∗ q.[κ] ={E}=∗ ∃ κ', (κ ⊑ κ') ∗ ▷ P ∗
      ∀ Q, ▷ (▷ Q -∗ <subj>[†κ'] ={E0}=∗ ▷ P) -∗ ▷ Q ={E}=∗ &{κ'}Q ∗ q.[κ].
  (* Atomic accessor for full borrows.

     In the context of view predicates, this accessor is a bit
     different when compared to the sequentially consistent case.
     Indeed, first, even if we can change the underlying predicate
     (i.e., going from P to Q), we must preserve the underlying
     resource, so that we need to preserve an existentially quantified
     assertion P'.

     Note that the "consequence" view shift is only executed if the
     borrow is actually opened afterwards (if the borrow is not
     opened, then P' still holds and can be used in the
     inheritance). Hence, we actually get that the consequence view
     shift happens at a view which is larger than the current view.
     See the comment in accessors.v for why the consequence VS only
     gets the dead token under the subjectively modality. *)
  Parameter bor_acc_atomic_strong : ∀ E κ P, ↑lftN ⊆ E →
    ⎡lft_ctx⎤ -∗ &{κ}P ={E,E∖↑lftN}=∗
      (∃ P' κ', κ ⊑ κ' ∗ ▷ (⎡P'⎤ ∧ P) ∗
         ∀ Q,
           ▷ (▷ Q -∗ <subj>[†κ'] ={E0}=∗ ▷ P) -∗
           ▷ (⎡P'⎤ ∧ Q)
           ={E∖↑lftN,E}=∗ &{κ'}Q) ∨
      (∃ κ', (κ ⊑ κ') ∗ <subj>[†κ'] ∗ |={E∖↑lftN,E}=> True).
  Parameter in_at_bor_acc : ∀ E q κ P, ↑lftN ⊆ E →
    ⎡lft_ctx⎤ -∗ &in_at{κ}P -∗ q.[κ] ={E,E∖↑lftN}=∗ ∃ Vb,
      (monPred_in Vb → ▷ P) ∗ ((monPred_in Vb → ▷ P) ={E∖↑lftN,E}=∗ q.[κ]).

  (* Because Coq's module system is horrible, we have to repeat
     properties of lft_incl here unless we want to prove them twice
     (both here and in model.primitive) *)
  Parameter lft_intersect_acc : ∀ κ κ' q q', q.[κ] -∗ q'.[κ'] -∗
    ∃ q'', q''.[κ ⊓ κ'] ∗ (q''.[κ ⊓ κ'] -∗ q.[κ] ∗ q'.[κ']).
  Parameter lft_intersect_incl_l : ∀ κ κ', (κ ⊓ κ' ⊑ κ)%I.
  Parameter lft_intersect_incl_r : ∀ κ κ', (κ ⊓ κ' ⊑ κ')%I.
  Parameter lft_incl_refl : ∀ κ, (κ ⊑ κ)%I.
  Parameter lft_incl_trans : ∀ κ κ' κ'', κ ⊑ κ' -∗ κ' ⊑ κ'' -∗ κ ⊑ κ''.
  Parameter lft_incl_glb : ∀ κ κ' κ'', κ ⊑ κ' -∗ κ ⊑ κ'' -∗ κ ⊑ κ' ⊓ κ''.
  Parameter lft_intersect_mono : ∀ κ1 κ1' κ2 κ2',
    κ1 ⊑ κ1' -∗ κ2 ⊑ κ2' -∗ κ1 ⊓ κ2 ⊑ κ1' ⊓ κ2'.
  Parameter lft_incl_acc : ∀ E κ κ' q,
    ↑lftN ⊆ E → κ ⊑ κ' -∗ q.[κ] ={E}=∗ ∃ q', q'.[κ'] ∗ (q'.[κ'] ={E}=∗ q.[κ]).
  Parameter lft_incl_intro : ∀ κ κ',
    □ (∀ q, q.[κ] ={↑lftN}=∗ ∃ q', q'.[κ'] ∗ (q'.[κ'] ={↑lftN}=∗ q.[κ]))
      -∗ κ ⊑ κ'.
  End properties.

  Parameter lftΣ : latticeT → gFunctors.
  Global Declare Instance subG_lftPreG Lat Σ : subG (lftΣ Lat) Σ → lftPreG Lat Σ.

  Parameter lft_init : ∀ `{invG Σ, !lftPreG Lat Σ} E E0,
      ↑lftN ⊆ E → ↑lftN ## E0 →
      (|={E}=> ∃ _ : lftG Lat E0 Σ, lft_ctx)%I.
End lifetime_sig.
