From lrust.lifetime Require Export primitive.
From lrust.lifetime Require Export faking reborrow.
From iris.algebra Require Import csum auth frac gmap agree.
From iris.proofmode Require Import tactics.
Set Default Proof Using "Type*".

Lemma uPred_and_split {M} (P Q1 Q2 : uPred M) :
  P ∧ (Q1 ∗ Q2) -∗ ∃ P1 P2, bi_persistently (P1 ∗ P2 → P) ∗ (Q1 ∧ P1) ∗ (Q2 ∧ P2).
Proof.
  uPred.unseal. split. intros ? x ?[HP(a1&a2&Hx&HQ1&HQ2)].
  exists (uPred_ownM a1), (uPred_ownM a2). uPred.unseal.
  eexists ε, _. rewrite left_id. split; [done|]. split.
  { intros ? x' ?? Hx'v (a1'&a2'& Hx' &(?&Ha1')&(?&Ha2')).
    rewrite Hx' Ha1' Ha2' assoc [_ ⋅ a2]comm assoc [a2 ⋅ a1]comm.
    eapply dist_le in Hx; last done. rewrite -Hx -assoc.
    eapply uPred_mono; [done|by eexists _|done]. }
  exists a1, a2. split; [done|].
  split; (split; [done|]); exists ε; by rewrite right_id.
Qed.

Lemma monPred_and_split {M I} (P Q1 Q2 : monPred I (uPredI M)) :
  P ∧ (Q1 ∗ Q2) -∗
  ∃ P1 P2, bi_persistently (⎡P1 ∗ P2⎤ → P) ∗ (Q1 ∧ ⎡P1⎤) ∗ (Q2 ∧ ⎡P2⎤).
Proof.
  iStartProof (uPred _). iIntros (i). rewrite uPred_and_split.
  iDestruct 1 as (P1 P2) "(#? & H1 & H2)". iExists P1, P2.
  iSplit; [by iIntros (? ->) "!#"|iFrame].
Qed.

Section borrow.
Context `{invG Σ, lftG Lat E0 Σ, LatBottom Lat}.
Implicit Types κ : lft.

Lemma bor_sep E κ P Q :
  ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ &{κ}(P ∗ Q) ={E}=∗ &{κ}P ∗ &{κ}Q.
Proof.
  iIntros (HE) "#LFT Hbor". iInv mgmtN as (A I) "(>HA & >HI & Hinv)" "Hclose".
  rewrite {1}/bor. iDestruct "Hbor" as (κ') "[#Hκκ' Hbor]".
  rewrite /raw_bor /idx_bor_own. iDestruct "Hbor" as (s) "[Hbor Hslice]".
  iDestruct "Hslice" as (P') "[#HPP' Hslice]".
  iDestruct "Hbor" as (V') "[#HV' Hbor]".
  iDestruct (own_bor_auth with "HI Hbor") as %Hκ'.
  rewrite big_sepS_later big_sepS_elem_of_acc //. rewrite [lft_inv _ κ']/lft_inv.
  iDestruct "Hinv" as "[Hinv Hclose']".
  iDestruct "Hinv" as (Vκ) "[>% [[Hinv >%]|[Hinv >%]]]"; last first.
  { unfold lft_inv_dead.
    iDestruct "Hinv" as (Pinh Vinh) "(>% & Hdead & Hcnt & Hinh)".
    iMod (raw_bor_fake with "[Hdead]") as "[Hdead Hbor1]"; [solve_ndisj|by iNext|].
    iMod (raw_bor_fake with "Hdead") as "[Hdead Hbor2]"; first solve_ndisj.
    iMod ("Hclose" with "[-Hbor1 Hbor2]") as "_".
    { iExists A, I. iFrame. rewrite big_sepS_later. iApply "Hclose'".
      iExists _. iSplit; [done|]. iRight. iSplit; [|done]. iNext. iExists _, _. by iFrame. }
    iSplitL "Hbor1"; rewrite /bor; iExists _; by iFrame. }
  rewrite ->lft_inv_alive_unfold. unfold lft_bor_alive.
  iDestruct "Hinv" as (Pb Pi) "(H & Hvs & Hinh)".
  iDestruct "H" as (B) "[[Hbox >Hown] HB]".
  iDestruct (own_bor_valid_2 with "Hown Hbor")
      as %[EQB%to_borUR_included _]%auth_valid_discrete_2.
  iAssert (▷ ⌜V' ⊑ Vκ⌝)%I with "[#]" as ">%".
  { iNext. rewrite big_sepM_lookup //. simpl. by iDestruct "HB" as %?. }
  iMod (slice_delete_full _ _ true with "Hslice Hbox")
    as (Pb0) "(HP' & #HPb0 & Hbox)"; [solve_ndisj| |].
  { rewrite lookup_fmap EQB //. }
  iAssert (▷ (⎡P' V'⎤ ∧ (P ∗ Q)))%I with "[HP']" as "HP'".
  { iNext. iSplit; [done|]. iApply "HPP'". by iApply monPred_in_elim. }
  rewrite monPred_and_split. iDestruct "HP'" as (P0 Q0) "(#HP' & HP & HQ)".
  iCombine "HP HQ HV' HPP'" as "HH". iClear "HPP'".
  iDestruct (monPred_in_intro with "HH") as (V) "(#HV & HP & HQ & % & #HPP')".
  set (P1 := (((monPred_in V) ∨ ⎡P0⎤) ∧ (monPred_in V → P))%I).
  set (Q1 := (((monPred_in V) ∨ ⎡Q0⎤) ∧ (monPred_in V → Q))%I).
  iMod (slice_insert_full _ _ _ _ _ P1 V' with "[HP] Hbox") as (γP HγP) "[#? Hbox]".
    solve_ndisj.
  { iSplit; [iRight; by iDestruct "HP" as "[_ $]"|].
    iIntros (???); by iDestruct "HP" as "[$ _]". }
  iMod (slice_insert_full _ _ _ _ _ Q1 V' with "[HQ] Hbox") as (γQ HγQ) "[#? Hbox]".
    solve_ndisj.
  { iSplit; [iRight; by iDestruct "HQ" as "[_ $]"|].
    iIntros (???); by iDestruct "HQ" as "[$ _]". }
  iMod (own_bor_update_2 with "Hown Hbor") as "Hbor".
  { etrans; last etrans.
    - apply auth_update_dealloc. by apply delete_singleton_local_update, _.
    - apply auth_update_alloc,
         (alloc_singleton_local_update _ γP (Excl (Bor_in V'))); last done.
      rewrite /to_borUR -fmap_delete lookup_fmap fmap_None
              -fmap_None -lookup_fmap fmap_delete //.
    - apply cmra_update_op; last done.
      apply auth_update_alloc,
        (alloc_singleton_local_update _ γQ (Excl (Bor_in V'))); last done.
      revert HγQ. change (Some V') with (bor_filled (Bor_in V')).
      by rewrite -!fmap_delete -!fmap_insert !lookup_fmap !fmap_None. }
  rewrite !own_bor_op. iDestruct "Hbor" as "[[H● H◯Q] H◯P]".
  iAssert (&{κ}P)%I with "[H◯P]" as "$".
  { rewrite /bor /raw_bor /idx_bor_own /=. iExists κ'. iFrame "Hκκ'". iExists γP.
    iSplitL; [by auto with iFrame|]. iExists _. iFrame "#". iNext. iAlways.
    iSplit; [|by auto]. iIntros "[_ H]". by iApply "H". }
  iAssert (&{κ}Q)%I with "[H◯Q]" as "$".
  { rewrite /bor /raw_bor /idx_bor_own /=. iExists κ'. iFrame "Hκκ'". iExists γQ.
    iSplitL; [by auto with iFrame|]. iExists _. iFrame "#". iNext. iAlways.
    iSplit; [|by auto]. iIntros "[_ H]". by iApply "H". }
  iMod ("Hclose" with "[-]"); [|done]. iModIntro ⎡_⎤%I. iExists A, I. iFrame.
  rewrite big_sepS_later. iApply "Hclose'". iExists Vκ. iSplit; [done|].
  iLeft. iFrame "%". rewrite lft_inv_alive_unfold /lft_bor_alive.
  iExists _, Pi. iSplitR "Hvs Hinh".
  - iExists (<[γQ := Bor_in V']>(<[γP := Bor_in V']>(delete s B))).
    rewrite /to_borUR !fmap_insert !fmap_delete !big_sepM_insert /=.
    + iFrame. rewrite (big_sepM_delete _ _ _ _ EQB) /=. iDestruct "HB" as "[>% $]". auto.
    + by rewrite -fmap_None -lookup_fmap fmap_delete.
    + by rewrite -(fmap_None bor_filled) -lookup_fmap fmap_insert fmap_delete.
  - iFrame "Hinh". iNext. iApply (lft_vs_cons with "[] Hvs").
    iIntros (V'' ?) "HD %% (HQ & HP & ?) %% ?". iFrame "HD". iModIntro. iNext. iRewrite "HPb0".
    iFrame. rewrite !bi.and_or_r. iDestruct "HPP'" as "[_ HPP']".
    iDestruct "HP" as "[[% HP]|HP]"; iDestruct "HQ" as "[[% HQ]|HQ]".
    + iApply "HPP'". iSplitL "HP"; [iApply "HP"|iApply "HQ"]; solve_lat.
    + iDestruct "HQ" as "[_ HQ]". iApply "HPP'".
      iSplitL "HP"; [iApply "HP"|iApply "HQ"]; solve_lat.
    + iDestruct "HP" as "[_ HP]". iApply "HPP'".
      iSplitL "HP"; [iApply "HP"|iApply "HQ"]; solve_lat.
    + iDestruct "HP" as "[HP _]". iDestruct "HQ" as "[HQ _]".
      iApply ("HP'" $! V' with "[$HP $HQ]").
Qed.

Lemma bor_combine E κ P Q :
  ↑lftN ⊆ E →
  ⎡lft_ctx⎤ -∗ &{κ}P -∗ &{κ}Q ={E}=∗ &{κ}(P ∗ Q).
Proof.
  iIntros (?) "#LFT HP HQ". rewrite {1 2}/bor.
  iDestruct "HP" as (κ1) "[#Hκ1 Hbor1]". iDestruct "HQ" as (κ2) "[#Hκ2 Hbor2]".
  iMod (raw_bor_shorten _ _ (κ1 ⊓ κ2) with "LFT Hbor1") as "Hbor1".
    done. by apply gmultiset_union_subseteq_l.
  iMod (raw_bor_shorten _ _ (κ1 ⊓ κ2) with "LFT Hbor2") as "Hbor2".
    done. by apply gmultiset_union_subseteq_r.
  iInv mgmtN as (A I) "(>HA & >HI & Hinv)" "Hclose". unfold raw_bor, idx_bor_own.
  iDestruct "Hbor1" as (j1) "[Hbor1 Hslice1]". iDestruct "Hbor2" as (j2) "[Hbor2 Hslice2]".
  iDestruct "Hbor1" as (V1) "[#HV1 Hbor1]". iDestruct "Hbor2" as (V2) "[#HV2 Hbor2]".
  iDestruct "Hslice1" as (P') "[#HPP' Hslice1]".
  iDestruct "Hslice2" as (Q') "[#HQQ' Hslice2]".
  iDestruct (own_bor_auth with "HI Hbor1") as %Hκ'.
  rewrite big_sepS_later big_sepS_elem_of_acc  //
          /lfts_inv /lft_inv /lft_inv_dead /lft_alive_in.
  iDestruct "Hinv" as "[Hinv Hclose']".
  iDestruct "Hinv" as (Vκ) "[>% [[Hinv >%]|[Hinv >%]]]"; last first.
  { iDestruct "Hinv" as (Pinh Vinh) "(>% & Hdead & Hcnt & Hinh)".
    iMod (raw_bor_fake with "[Hdead]") as "[Hdead Hbor]"; [solve_ndisj|by iNext|].
    iMod ("Hclose" with "[-Hbor]") as "_".
    { iExists A, I. iFrame. rewrite big_sepS_later. iApply "Hclose'".
      iExists _. iFrame "%". iRight. iFrame "%". iExists _, _. iNext. by iFrame. }
    unfold bor. iExists _. iFrame. iApply (lft_incl_glb with "Hκ1 Hκ2"). }
  rewrite ->lft_inv_alive_unfold. unfold lft_bor_alive.
  iDestruct "Hinv" as (Pb Pi) "(H & Hvs & Hinh)".
  iDestruct "H" as (B) "[[Hbox >Hown] HB]".
  iDestruct (own_bor_valid_2 with "Hown Hbor1")
      as %[EQB1%to_borUR_included _]%auth_valid_discrete_2.
  iDestruct (own_bor_valid_2 with "Hown Hbor2")
      as %[EQB2%to_borUR_included _]%auth_valid_discrete_2.
  iAssert ⌜j1 ≠ j2⌝%I with "[#]" as %Hj1j2.
  { iIntros (->).
    iDestruct (own_bor_valid_2 with "Hbor1 Hbor2") as %Hj1j2%auth_valid_discrete.
    exfalso. revert Hj1j2. by rewrite /= op_singleton singleton_valid. }
  iMod (slice_combine _ _ true with "Hslice1 Hslice2 Hbox")
    as (γ) "(% & Hslice & Hbox)"; first solve_ndisj.
  { done. }
  { by rewrite lookup_fmap EQB1. }
  { by rewrite lookup_fmap EQB2. }
  iCombine "Hown" "Hbor1" as "Hbor1". iCombine "Hbor1" "Hbor2" as "Hbor".
  rewrite -!own_bor_op. iMod (own_bor_update with "Hbor") as "Hbor".
  { etrans; last etrans.
    - apply cmra_update_op; last done.
      apply auth_update_dealloc. by apply delete_singleton_local_update, _.
    - apply auth_update_dealloc. by apply delete_singleton_local_update, _.
    - apply auth_update_alloc,
         (alloc_singleton_local_update _ γ (Excl (Bor_in (V1 ⊔ V2)))); last done.
      rewrite /to_borUR -!fmap_delete lookup_fmap fmap_None
              -fmap_None -lookup_fmap !fmap_delete //. }
  rewrite own_bor_op. iDestruct "Hbor" as "[H● H◯]".
  iAssert (&{κ}(P ∗ Q))%I with "[H◯ Hslice]" as "$".
  { rewrite /bor /raw_bor /idx_bor_own. iExists (κ1 ⊓ κ2).
    iSplit; first by iApply (lft_incl_glb with "Hκ1 Hκ2").
    iExists γ. iSplit.
    - iExists _. simpl. iFrame. iCombine "HV1 HV2" as "HV12".
      iDestruct (monPred_in_intro with "HV12") as (V12) "(HV12' & % & %)".
      iApply (monPred_in_elim _ V12). done. rewrite monPred_at_in. solve_lat.
    - iExists _. iFrame. iNext. iAlways.
      by iSplit; iIntros "[HP HQ]"; iSplitL "HP"; (iApply "HPP'" || iApply "HQQ'"). }
  iMod ("Hclose" with "[-]"); [|done]. iExists A, I. iFrame. rewrite big_sepS_later.
  iApply "Hclose'". iExists _. iSplit; [done|]. iLeft. iSplit; [|done].
  rewrite lft_inv_alive_unfold /lft_bor_alive. iExists Pb, Pi. iFrame. iExists _.
  rewrite -!fmap_delete -!fmap_insert. iFrame "H●".
  rewrite !fmap_insert !fmap_delete /=. iFrame "Hbox".
  rewrite !big_sepM_insert /=; last first.
  { by rewrite -fmap_None -lookup_fmap !fmap_delete. }
  rewrite (big_sepM_delete _ _ _ _ EQB1)
          (big_sepM_delete _ (delete j1 _) j2 (Bor_in V2)) /=; first last.
  { rewrite lookup_delete_ne //. }
  iDestruct "HB" as "(>% & >% & $)". solve_lat.
Qed.
End borrow.
