From iris.base_logic.lib Require Export invariants.
From iris.algebra Require Import auth gmap agree.
From iris.bi Require Import big_op.
From iris.proofmode Require Import tactics.
From gpfsl Require Import vprop.
From gpfsl.lang Require Import lattice_cmra.
Set Default Proof Using "Type".
Import uPred.

Implicit Types Lat : latticeT.

(** The CMRAs we need. *)
Class boxG Lat Σ :=
  boxG_inG :> inG Σ (prodR
    (authR (optionUR (exclR (optionC (latC Lat)))))
    (optionR (agreeR (laterC (Lat -c> iPreProp Σ))))).

Definition boxΣ Lat : gFunctors := #[
   GFunctor (authR (optionUR (exclR (optionC (latC Lat)))) *
   optionRF (agreeRF (▶ (Lat -c> ∙))) ) ].

Instance subG_stsΣ Lat Σ : subG (boxΣ Lat) Σ → boxG Lat Σ.
Proof. solve_inG. Qed.

Section box_defs.
  Context `{invG Σ, boxG Lat Σ} (N : namespace).
  Notation iProp := (iProp Σ).
  Notation monPred := (monPred (lat_bi_index Lat) (uPredSI (iResUR Σ))).

  Definition slice_name := gname.

  Definition box_own_auth (γ : slice_name) (a : auth _) : iProp :=
    own γ (a, ε).

  Definition box_own_prop (γ : slice_name) (P : monPred) : iProp :=
    own γ (ε, Some (to_agree (Next (iProp_unfold ∘ P : _ -c> _)))).

  Definition slice_inv (γ : slice_name) (P : monPred) : iProp :=
    (∃ o, box_own_auth γ (● Excl' o) ∗ from_option (P ∘ of_latT) True o)%I.

  Definition slice (γ : slice_name) (P : monPred) : iProp :=
    (box_own_prop γ P ∗ inv N (slice_inv γ P))%I.

  Definition box (f : gmap slice_name (option Lat)) (P : monPred) : iProp :=
    (∃ Φ : slice_name → monPred,
      ▷ (P ≡ [∗ map] γ ↦ _ ∈ f, Φ γ) ∗
      [∗ map] γ ↦ o ∈ f, box_own_auth γ (◯ Excl' (to_latT <$> o)) ∗
                         box_own_prop γ (Φ γ) ∗ inv N (slice_inv γ (Φ γ)))%I.
End box_defs.

Instance: Params (@box_own_prop) 4.
Instance: Params (@box_own_auth) 4.
Instance: Params (@slice_inv) 4.
Instance: Params (@slice) 6.
Instance: Params (@box) 6.

Section box.
Context `{invG Σ, boxG Lat Σ} (N : namespace).

Notation monPred := (monPred (lat_bi_index Lat) (uPredSI (iResUR Σ))).
Implicit Types P Q : monPred.

Global Instance box_own_prop_ne γ : NonExpansive (box_own_prop γ).
Proof. unfold box_own_prop. intros ????. (do 5 f_equiv)=>? /=. by do 2 f_equiv. Qed.
Global Instance box_own_prop_contractive γ : Contractive (box_own_prop γ).
Proof.
  unfold box_own_prop. intros ????.
  do 4 f_equiv. f_contractive=>? /=. by do 2 f_equiv.
Qed.

Global Instance box_own_auth_proper γ : Proper ((≡) ==> (≡)) (box_own_auth γ).
Proof. solve_proper. Qed.

Global Instance box_inv_ne γ : NonExpansive (slice_inv γ).
Proof. unfold slice_inv, from_option=>??? Heq. repeat (apply Heq || f_equiv). Qed.

Global Instance slice_ne γ : NonExpansive (slice N γ).
Proof. solve_proper. Qed.
Global Instance slice_contractive γ : Contractive (slice N γ).
Proof. solve_contractive. Qed.
Global Instance slice_proper γ : Proper ((≡) ==> (≡)) (slice N γ).
Proof. apply ne_proper, _. Qed.

Global Instance slice_persistent γ P : Persistent (slice N γ P).
Proof. apply _. Qed.

Global Instance box_contractive f : Contractive (box N f).
Proof. unfold box=>??? Heq. repeat (apply Heq || f_contractive || f_equiv). Qed.
Global Instance box_ne f : NonExpansive (box N f).
Proof. apply (contractive_ne _). Qed.
Global Instance box_proper f : Proper ((≡) ==> (≡)) (box N f).
Proof. apply ne_proper, _. Qed.

Lemma box_own_auth_agree γ o1 o2 :
  box_own_auth γ (● Excl' o1) ∗ box_own_auth γ (◯ Excl' o2) ⊢ o1 ≡ o2.
Proof.
  rewrite /box_own_prop -own_op own_valid prod_validI /= and_elim_l.
  iDestruct 1 as %[[[[]|] EQ%(inj Some)] _]%auth_valid_discrete; by inversion_clear EQ.
Qed.

Lemma box_own_auth_update γ o1 o2 o3 :
  box_own_auth γ (● Excl' o1) ∗ box_own_auth γ (◯ Excl' o2)
  ==∗ box_own_auth γ (● Excl' o3) ∗ box_own_auth γ (◯ Excl' o3).
Proof.
  rewrite /box_own_auth -!own_op. apply own_update, prod_update; last done.
  by apply auth_update, option_local_update, exclusive_local_update.
Qed.

Lemma box_own_agree γ Q1 Q2 :
  box_own_prop γ Q1 ∗ box_own_prop γ Q2 ⊢ ▷ (Q1 ≡ Q2).
Proof.
  rewrite /box_own_prop -own_op own_valid prod_validI /= and_elim_r.
  rewrite option_validI /= agree_validI agree_equivI later_equivI /=.
  iIntros "#HQ". iApply monPred_equivI. iIntros (V).
  iApply iProp_unfold_equivI. rewrite ofe_fun_equivI. iNext. iApply "HQ".
Qed.

Lemma box_alloc : box N ∅ True%I.
Proof.
  iIntros. iExists (λ _, True%I); iSplit; [|done]. by rewrite big_sepM_empty True_emp.
Qed.

Lemma slice_insert_empty E q f Q P :
  ▷?q box N f P ={E}=∗ ∃ γ, ⌜f !! γ = None⌝ ∗
    slice N γ Q ∗ ▷?q box N (<[γ:=None]> f) (Q ∗ P).
Proof.
  iDestruct 1 as (Φ) "[#HeqP Hf]".
  iMod (own_alloc_strong (● Excl' None ⋅ ◯ Excl' None,
    Some (to_agree (Next ((λ V, iProp_unfold (Q V)) : _ -c> _)))) (dom _ f))
    as (γ) "[Hdom Hγ]"; first done.
  rewrite pair_split. iDestruct "Hγ" as "[[Hγ Hγ'] #HγQ]".
  iDestruct "Hdom" as % ?%not_elem_of_dom.
  iMod (inv_alloc N _ (slice_inv γ Q) with "[Hγ]") as "#Hinv".
  { iNext. iExists None; eauto. }
  iModIntro; iExists γ; repeat iSplit; auto.
  iNext. iExists (<[γ:=Q]> Φ); iSplit.
  - iNext. iRewrite "HeqP". rewrite -big_sepM_fn_insert' //.
  - rewrite (big_sepM_fn_insert (λ _ _ P',  _ ∗ _ _ P' ∗ _ _ (_ _ P')))%I //.
    iFrame. auto.
Qed.

Lemma slice_delete_empty q f P Q γ :
  f !! γ = Some None →
  slice N γ Q -∗ ▷?q box N f P -∗ ∃ P',
    ▷?q ▷ (P ≡ (Q ∗ P')) ∗ ▷?q box N (delete γ f) P'.
Proof.
  iIntros (?) "#[HγQ _] H". iDestruct "H" as (Φ) "[#HeqP Hf]".
  iDestruct (big_opM_delete _ f _ None with "Hf") as "[[_ [HΦ _]] Hf]"; [done|].
  set (P' := ([∗ map] γ'↦_ ∈ delete γ f, Φ γ')%I). iExists P'. iSplitR "Hf"; iNext.
  - iDestruct (box_own_agree γ Q (Φ γ) with "[#]") as "HeqQ"; first by eauto. iNext.
    iRewrite "HeqP". iRewrite "HeqQ". rewrite big_opM_delete //.
  - iExists Φ; eauto.
Qed.

Lemma slice_fill E q f γ P Q V :
  ↑N ⊆ E →
  f !! γ = Some None →
  slice N γ Q -∗ ▷ Q V -∗ ▷?q box N f P ={E}=∗ ▷?q box N (<[γ:=Some V]> f) P.
Proof.
  iIntros (??) "#[HγQ Hinv] HQ H"; iDestruct "H" as (Φ) "[#HeqP Hf]".
  iInv N as (o') "[>Hγ _]" "Hclose".
  iDestruct (big_opM_delete _ f _ None with "Hf")
    as "[[>Hγ' #[HγΦ Hinv']] ?]"; first done.
  iMod (box_own_auth_update γ o' None (Some $ to_latT V) with "[$Hγ $Hγ']") as "[Hγ Hγ']".
  iMod ("Hclose" with "[Hγ HQ]") as "_"; [iNext; iExists (Some $ to_latT V); by iFrame|].
  iModIntro; iNext; iExists Φ; iSplit.
  - by rewrite big_opM_insert_override.
  - rewrite -insert_delete big_opM_insert ?lookup_delete //. auto with iFrame.
Qed.

Lemma slice_empty E q f P Q γ V :
  ↑N ⊆ E →
  f !! γ = Some (Some V) →
  slice N γ Q -∗ ▷?q box N f P ={E}=∗ ▷ Q V ∗ ▷?q box N (<[γ:=None]> f) P.
Proof.
  iIntros (??) "#[HγQ Hinv] H"; iDestruct "H" as (Φ) "[#HeqP Hf]".
  iInv N as (o) "[>Hγ HQ]" "Hclose".
  iDestruct (big_opM_delete _ f with "Hf") as "[[>Hγ' #[HγΦ Hinv']] ?]"; [done|].
  iDestruct (box_own_auth_agree γ o (Some $ to_latT V) with "[-]") as %EQ; [by iFrame|].
  inversion_clear EQ. setoid_subst. iFrame "HQ".
  iMod (box_own_auth_update γ with "[$Hγ $Hγ']") as "[Hγ Hγ']".
  iMod ("Hclose" with "[Hγ]") as "_"; [iNext; iExists None; by repeat iSplit|].
  iModIntro; iNext; iExists Φ; iSplit.
  - by rewrite big_opM_insert_override.
  - rewrite -insert_delete big_opM_insert ?lookup_delete //. auto with iFrame.
Qed.

Lemma slice_insert_full E q f P Q V :
  ↑N ⊆ E →
  ▷ Q V -∗ ▷?q box N f P ={E}=∗ ∃ γ, ⌜f !! γ = None⌝ ∗
    slice N γ Q ∗ ▷?q box N (<[γ:=Some V]> f) (Q ∗ P).
Proof.
  iIntros (?) "HQ Hbox".
  iMod (slice_insert_empty with "Hbox") as (γ) "(% & #Hslice & Hbox)".
  iExists γ. iFrame "%#". iMod (slice_fill with "Hslice HQ Hbox"); first done.
  by apply lookup_insert. by rewrite insert_insert.
Qed.

Lemma slice_delete_full E q f P Q γ V :
  ↑N ⊆ E →
  f !! γ = Some (Some V) →
  slice N γ Q -∗ ▷?q box N f P ={E}=∗
  ∃ P', ▷ Q V ∗ ▷?q ▷ (P ≡ (Q ∗ P')) ∗ ▷?q box N (delete γ f) P'.
Proof.
  iIntros (??) "#Hslice Hbox".
  iMod (slice_empty with "Hslice Hbox") as "[$ Hbox]"; try done.
  iDestruct (slice_delete_empty with "Hslice Hbox") as (P') "[Heq Hbox]".
  { by apply lookup_insert. }
  iExists P'. iFrame. rewrite -insert_delete delete_insert ?lookup_delete //.
Qed.

Lemma box_fill E f P V :
  ↑N ⊆ E →
  box N f P -∗ ▷ P V ={E}=∗ box N (const (Some V) <$> f) P.
Proof.
  iIntros (?) "H HP"; iDestruct "H" as (Φ) "[#HeqP Hf]".
  iExists Φ; iSplitR.
  { iModIntro. iNext. iRewrite "HeqP". by rewrite big_opM_fmap. }
  iAssert ([∗ map] γ↦_ ∈ f, ▷ Φ γ V)%I with "[HP]" as "HP'".
  { rewrite -big_sepM_later -monPred_at_big_sepM. iNext. by iRewrite "HeqP" in "HP". }
  iCombine "Hf" "HP'" as "Hf".
  rewrite -big_sepM_sepM big_opM_fmap; iApply (fupd_big_sepM _ _ f).
  iApply (@big_sepM_impl with "[$Hf]").
  iAlways; iIntros (γ b' ?) "[(Hγ' & #$ & #$) HΦ]".
  iInv N as (b) "[>Hγ _]" "Hclose".
  iMod (box_own_auth_update γ with "[Hγ Hγ']") as "[Hγ $]"; first by iFrame.
  iApply "Hclose". iNext; iExists (Some $ to_latT V). by iFrame.
Qed.

Lemma box_empty E f P V :
  ↑N ⊆ E →
  map_Forall (λ _, from_option (⊑ V) False) f →
  box N f P ={E}=∗ ▷ P V ∗ box N (const None <$> f) P.
Proof.
  iDestruct 1 as (Φ) "[#HeqP Hf]".
  iAssert (([∗ map] γ↦b ∈ f, ▷ Φ γ V) ∗
    [∗ map] γ↦b ∈ f, box_own_auth γ (◯ Excl' None) ∗  box_own_prop γ (Φ γ) ∗
      inv N (slice_inv γ (Φ γ)))%I with "[> Hf]" as "[HΦ ?]".
  { rewrite -big_opM_opM -fupd_big_sepM. iApply (@big_sepM_impl with "[$Hf]").
    iAlways; iIntros (γ o ?) "(Hγ' & #HγΦ & #Hinv)".
    assert (Ho : from_option (⊑ V) False o) by eauto.
    destruct o as [V'|]=>//. rewrite -Ho.
    iInv N as (b) "[>Hγ HΦ]" "Hclose".
    iDestruct (box_own_auth_agree γ b (Some _) with "[-]") as %EQ; first by iFrame.
    inversion_clear EQ. setoid_subst.
    iMod (box_own_auth_update γ (Some _) (Some _) None with "[$Hγ $Hγ']") as "[Hγ $]".
    iMod ("Hclose" with "[Hγ]") as "_"; [unfold slice_inv; by auto with iFrame|].
    iFrame "HγΦ Hinv". by iApply "HΦ". }
  iModIntro; iSplitL "HΦ".
  - rewrite -big_sepM_later -monPred_at_big_sepM. iNext. by iRewrite "HeqP".
  - iExists Φ; iSplit; [|by rewrite big_opM_fmap].
    iNext. iRewrite "HeqP". by rewrite big_opM_fmap.
Qed.

Lemma slice_iff E q f P Q Q' γ o :
  ↑N ⊆ E → f !! γ = Some o →
  ▷ □ (∀ V, Q V ↔ Q' V) -∗ slice N γ Q -∗ ▷?q box N f P ={E}=∗ ∃ γ' P',
    ⌜delete γ f !! γ' = None⌝ ∗ ▷?q ▷ □ (∀ V, P V ↔ P' V) ∗
    slice N γ' Q' ∗ ▷?q box N (<[γ' := o]>(delete γ f)) P'.
Proof.
  iIntros (??) "#HQQ' #Hs Hb". destruct o.
  - iMod (slice_delete_full with "Hs Hb") as (P') "(HQ & #Heq & Hb)"; try done.
    iDestruct ("HQQ'" with "HQ") as "HQ'".
    iMod (slice_insert_full with "HQ' Hb") as (γ') "(% & #Hs' & Hb)"; try done.
    iExists γ', _. iFrame "∗#%". iIntros "!>". do 2 iNext. iAlways. iIntros (V).
    iRewrite "Heq". by iSplit; iIntros "[? $]"; iApply "HQQ'".
  - iDestruct (slice_delete_empty with "Hs Hb") as (P') "(#Heq & Hb)"; [done|].
    iMod (slice_insert_empty with "Hb") as (γ') "(% & #Hs' & Hb)"; try done.
    iExists γ', _. iFrame "∗#%". iIntros "!>". do 2 iNext. iAlways. iIntros (V).
    iRewrite "Heq". by iSplit; iIntros "[? $]"; iApply "HQQ'".
Qed.

Lemma slice_combine E q f P Q1 Q2 γ1 γ2 V1 V2 :
  ↑N ⊆ E → γ1 ≠ γ2 → f !! γ1 = Some (Some V1) → f !! γ2 = Some (Some V2) →
  slice N γ1 Q1 -∗ slice N γ2 Q2 -∗ ▷?q box N f P ={E}=∗ ∃ γ,
    ⌜delete γ2 (delete γ1 f) !! γ = None⌝ ∗ slice N γ (Q1 ∗ Q2) ∗
    ▷?q box N (<[γ := Some (V1 ⊔ V2)]>(delete γ2 (delete γ1 f))) P.
Proof.
  iIntros (????) "#Hslice1 #Hslice2 Hbox".
  iMod (slice_delete_full with "Hslice1 Hbox") as (P1) "(HQ1 & #Heq1 & Hbox)"; try done.
  iMod (slice_delete_full with "Hslice2 Hbox") as (P2) "(HQ2 & #Heq2 & Hbox)"; first done.
  { by simplify_map_eq. }
  iMod (slice_insert_full _ _ _ _ (Q1 ∗ Q2)%I (V1 ⊔ V2) with "[$HQ1 $HQ2] Hbox")
    as (γ) "(% & #Hslice & Hbox)"; first done.
  iExists γ. iFrame "%#". iModIntro. iNext.
  iApply (internal_eq_rewrite_contractive with "[]"); [|done].
  iNext. iRewrite "Heq1". iRewrite "Heq2". by rewrite assoc.
Qed.
End box.

Typeclasses Opaque slice box.
