From iris.algebra Require Import csum auth frac gmap agree gset.
From lrust.lifetime.model Require Import boxes.
From lrust.lifetime Require Export lifetime_sig.
From gpfsl.lang Require Export lattice_cmra.
Set Default Proof Using "Type".
Import uPred.
Implicit Types (Lat : latticeT) (E : coPset).

Definition borN : namespace := lftN .@ "bor".
Definition inhN : namespace := lftN .@ "inh".
Definition mgmtN : namespace := lftN .@ "mgmt".

Definition atomic_lft := positive.
(* HACK : We need to make sure this is not in the top-level context,
   so that it does not conflict with the *definition* of [lft] that we
   use in lifetime.v. *)
Module Export lft_notation.
  Notation lft := (gmultiset atomic_lft).
End lft_notation.

Definition static : lft := (∅ : gmultiset _).
Instance lft_intersect : Meet lft := union.

Inductive bor_state {Lat} :=
  | Bor_in (V : Lat)
  | Bor_open (q : frac) (Vtok : Lat) (V : Lat)
  | Bor_rebor (κ : lft) (V : Lat).
Arguments bor_state : clear implicits.

Canonical Structure bor_stateC Lat := leibnizC (bor_state Lat).

Record lft_names := LftNames {
  bor_name : gname;
  cnt_name : gname;
  inh_name : gname
}.
Canonical lft_namesC := leibnizC lft_names.

Definition lft_stateR Lat := csumR (prodR fracR (latR Lat)) (agreeR (latC Lat)).
Definition alftUR Lat := gmapUR atomic_lft (lft_stateR Lat).
Definition ilftUR := gmapUR lft (agreeR lft_namesC).
Definition borUR Lat := gmapUR slice_name (exclR (bor_stateC Lat)).
Definition inhUR := gset_disjUR slice_name.

Class lftG Lat E0 Σ := LftG {
  lft_box :> boxG Lat Σ;
  alft_inG :> inG Σ (authR (alftUR Lat));
  alft_name : gname;
  ilft_inG :> inG Σ (authR ilftUR);
  ilft_name : gname;
  lft_bor_inG :> inG Σ (authR (borUR Lat));
  lft_cnt_inG :> inG Σ (authR natUR);
  lft_inh_inG :> inG Σ (authR inhUR);
  E0_lftN_disj : ↑lftN ## E0;
}.
Definition lftG' := lftG.

Class lftPreG Lat Σ := LftPreG {
  lft_preG_box :> boxG Lat Σ;
  alft_preG_inG :> inG Σ (authR (alftUR Lat));
  ilft_preG_inG :> inG Σ (authR ilftUR);
  lft_preG_bor_inG :> inG Σ (authR (borUR Lat));
  lft_preG_cnt_inG :> inG Σ (authR natUR);
  lft_preG_inh_inG :> inG Σ (authR inhUR);
}.
Definition lftPreG' := lftPreG.

Definition lftΣ Lat : gFunctors :=
  #[ boxΣ Lat; GFunctor (authR (alftUR Lat)); GFunctor (authR ilftUR);
     GFunctor (authR (borUR Lat)); GFunctor (authR natUR);  GFunctor (authR inhUR) ].
Instance subG_lftPreG :
  subG (lftΣ Lat) Σ → lftPreG Lat Σ.
Proof. solve_inG. Qed.

Definition bor_filled {Lat} (s : bor_state Lat) : option Lat :=
  match s with Bor_in V => Some V | _ => None end.

Definition to_lft_stateR {Lat} (bV : bool * Lat) : lft_stateR Lat :=
  if bV.1 then Cinl (1%Qp, to_latT bV.2) else Cinr $ to_agree $ to_latT bV.2.
Definition to_alftUR {Lat} : gmap atomic_lft (bool * Lat) → alftUR Lat :=
  fmap to_lft_stateR.
Definition to_ilftUR : gmap lft lft_names → ilftUR := fmap to_agree.
Definition to_borUR {Lat} : gmap slice_name (bor_state Lat) → borUR Lat :=
  fmap Excl.

Section defs.
  Context {Σ Lat} `{invG Σ, lftG Lat E0 Σ}.

  Notation iProp := (iProp Σ).
  Notation monPred := (monPredSI (lat_bi_index Lat) (uPredSI (iResUR Σ))).

  Definition lft_tok (q : Qp) (κ : lft) : monPred :=
    ([∗ mset] Λ ∈ κ, ∃ V, monPred_in V ∗
     ⎡own alft_name (◯ {[ Λ := Cinl (q, to_latT V) ]})⎤)%I.

  Definition lft_dead (κ : lft) : monPred :=
    (∃ Λ V, ⌜Λ ∈ κ⌝ ∗ monPred_in V ∗
             ⎡own alft_name (◯ {[ Λ := Cinr (to_agree $ to_latT V) ]})⎤)%I.

  Definition own_alft_auth (A : gmap atomic_lft (bool * Lat)) : iProp :=
    own alft_name (● to_alftUR A).
  Definition own_ilft_auth (I : gmap lft lft_names) : iProp :=
    own ilft_name (● to_ilftUR I).

  Definition own_bor (κ : lft) (x : authR (borUR Lat)) : iProp :=
    (∃ γs,
      own ilft_name (◯ {[ κ := to_agree γs ]}) ∗
      own (bor_name γs) x)%I.

  Definition own_cnt (κ : lft) (x : authR natUR) : iProp :=
    (∃ γs,
      own ilft_name (◯ {[ κ := to_agree γs ]}) ∗
      own (cnt_name γs) x)%I.

  Definition own_inh (κ : lft) (x : authR inhUR) : iProp :=
    (∃ γs,
      own ilft_name (◯ {[ κ := to_agree γs ]}) ∗
      own (inh_name γs) x)%I.

  Definition bor_cnt (κ : lft) (s : bor_state Lat) : monPred :=
    match s with
    | Bor_in V => monPred_in V
    | Bor_open q Vtok V => monPred_in V ∗ ⎡lft_tok q κ Vtok⎤
    | Bor_rebor κ' V => monPred_in V ∗ ⌜κ ⊂ κ'⌝ ∗ ⎡own_cnt κ' (◯ 1)%nat⎤
    end%I.

  Definition lft_bor_alive (κ : lft) (Pi : monPred) : monPred :=
    (∃ B : gmap slice_name (bor_state Lat),
      ⎡box borN (bor_filled <$> B) Pi ∗ own_bor κ (● to_borUR B)⎤ ∗
      [∗ map] s ∈ B, bor_cnt κ s)%I.

  Definition lft_bor_dead (κ : lft) : iProp :=
    (∃ (B: gmap slice_name Lat) (Pb : monPred),
       own_bor κ (● ((λ V, Excl (Bor_in V)) <$> B)) ∗
       box borN (const None <$> B) Pb)%I.

   Definition lft_inh (κ : lft) (s : option Lat) (Pi : monPred) : iProp :=
     (∃ E : gset slice_name,
       own_inh κ (● GSet E) ∗
       box inhN (to_gmap s E) Pi)%I.

   Definition lft_vs_inv_go (κ : lft) (lft_inv_alive : ∀ κ', κ' ⊂ κ → Lat → iProp)
       (Vs : lft → Lat) (I : gmap lft lft_names) : iProp :=
     (lft_bor_dead κ ∗
       own_ilft_auth I ∗
       [∗ set] κ' ∈ dom _ I, ∀ Hκ : κ' ⊂ κ, lft_inv_alive κ' Hκ (Vs κ'))%I.

   Definition lft_vs_go (κ : lft) (lft_inv_alive : ∀ κ', κ' ⊂ κ → monPred)
              (Pb Pi : monPred) : monPred :=
     (∃ Vκ n, monPred_in Vκ ∗
        ⎡own_cnt κ (● n) ∗
         ∀ (I : gmap lft lft_names) (Vs : lft → Lat),
           ⌜Vκ ⊑ Vs κ⌝ -∗
           ⌜∀ κ', κ' ∈ dom (gset _) I → κ' ⊆ κ → Vs κ ⊑ Vs κ'⌝ -∗
           lft_vs_inv_go κ lft_inv_alive Vs I -∗ ▷ Pb (Vs κ) -∗ (∃ V, lft_dead κ V)
             ={↑borN ∪ E0}=∗
           lft_vs_inv_go κ lft_inv_alive Vs I ∗ ▷ Pi (Vs κ) ∗ own_cnt κ (◯ n)⎤)%I.

   Definition lft_inv_alive_go (κ : lft)
              (lft_inv_alive : ∀ κ', κ' ⊂ κ → monPred) : monPred :=
    (∃ Pb Pi,
      lft_bor_alive κ Pb ∗
      lft_vs_go κ lft_inv_alive Pb Pi ∗
      ⎡lft_inh κ None Pi⎤)%I.

  Definition lft_inv_alive (κ : lft) : monPred :=
    Fix_F _ lft_inv_alive_go (gmultiset_wf κ).
  Definition lft_vs_inv (κ : lft) (Vs : lft → Lat) (I : gmap lft lft_names) : iProp :=
    lft_vs_inv_go κ (λ κ' _, lft_inv_alive κ') Vs I.
  Definition lft_vs (κ : lft) (Pb Pi : monPred) : monPred :=
    lft_vs_go κ (λ κ' _, lft_inv_alive κ') Pb Pi.

  Definition lft_inv_dead (κ : lft) : monPred :=
    (∃ Pi V, monPred_in V ∗
      ⎡lft_bor_dead κ ∗ own_cnt κ (● 0)%nat ∗ lft_inh κ (Some V) Pi⎤)%I.

  Definition lft_has_view (A : gmap atomic_lft (bool * Lat)) (κ : lft) (Vκ : Lat) :=
    ∀ Λ : atomic_lft, Λ ∈ κ → from_option ((Vκ ⊑) ∘ snd) False (A !! Λ).
  Definition lft_alive_in (A : gmap atomic_lft (bool * Lat)) (κ : lft) : Prop :=
    ∀ Λ : atomic_lft, Λ ∈ κ → fst <$> A !! Λ = Some true.
  Definition lft_dead_in (A : gmap atomic_lft (bool * Lat)) (κ : lft) : Prop :=
    ∃ (Λ : atomic_lft), Λ ∈ κ ∧ fst <$> A !! Λ = Some false.

  Definition lft_inv (A : gmap atomic_lft (bool * Lat)) (κ : lft) : iProp :=
    (∃ Vκ, ⌜lft_has_view A κ Vκ⌝ ∗
       (lft_inv_alive κ Vκ ∗ ⌜lft_alive_in A κ⌝ ∨
        lft_inv_dead κ Vκ ∗ ⌜lft_dead_in A κ⌝))%I.

  Definition lfts_inv : iProp :=
    (∃ (A : gmap atomic_lft (bool * Lat)) (I : gmap lft lft_names),
      own_alft_auth A ∗
      own_ilft_auth I ∗
      [∗ set] κ ∈ dom _ I, lft_inv A κ)%I.

  Definition lft_ctx : iProp := inv mgmtN lfts_inv.

  Definition lft_incl (κ κ' : lft) : monPred :=
    (□ (∀ q, lft_tok q κ ={↑lftN}=∗ ∃ q',
       lft_tok q' κ' ∗ (lft_tok q' κ' ={↑lftN}=∗ lft_tok q κ)))%I.

  Definition bor_idx := (lft * slice_name)%type.

  Definition idx_bor_own (i : bor_idx) : monPred :=
    (∃ V, monPred_in V ∗ ⎡own_bor (i.1) (◯ {[ i.2 := Excl (Bor_in $ to_latT V) ]})⎤)%I.

  Definition idx_bor (κ : lft) (i : bor_idx) (P : monPred) : monPred :=
    (lft_incl κ (i.1) ∗ ∃ (P' : monPred), ▷ □ (P' ↔ P) ∗ ⎡slice borN (i.2) P'⎤)%I.

  Definition raw_bor (κ : lft) (P : monPred) : monPred :=
    (∃ s : slice_name, idx_bor_own (κ, s) ∗
       ∃ (P' : monPred), ▷ □ (P' ↔ P) ∗ ⎡slice borN s P'⎤)%I.

  Definition bor (κ : lft) (P : monPred) : monPred :=
    (∃ κ', lft_incl κ κ' ∗ raw_bor κ' P)%I.
End defs.

Instance: Params (@lft_bor_alive) 6.
Instance: Params (@lft_inh) 6.
Instance idx_bor_params : Params (@idx_bor) 6.
Instance: Params (@raw_bor) 5.
Instance bor_params : Params (@bor) 5.

Notation "q .[ κ ]" := (lft_tok q κ)
    (format "q .[ κ ]", at level 0) : bi_scope.
Notation "[† κ ]" := (lft_dead κ) (format "[† κ ]"): bi_scope.

Notation "&{ κ }" := (bor κ) (format "&{ κ }") : bi_scope.
Notation "&{ κ , i }" := (idx_bor κ i) (format "&{ κ , i }") : bi_scope.

Infix "⊑" := lft_incl (at level 70) : bi_scope.

(* TODO: Making all these things opaque is rather annoying, we should
   find a way to avoid it, or *at least*, to avoid having to manually unfold
   this because iDestruct et al don't look through these names any more. *)
Typeclasses Opaque lft_tok lft_dead bor_cnt lft_bor_alive lft_bor_dead
  lft_inh lft_inv_alive lft_vs_inv lft_vs lft_inv_dead lft_inv lft_incl
  idx_bor_own idx_bor raw_bor bor.

Section basic_properties.
Context `{invG Σ, lftG Lat E0 Σ}.
Implicit Types κ : lft.

Notation iProp := (iProp Σ).
Notation monPred := (monPred (lat_bi_index Lat) (uPredSI (iResUR Σ))).

(* Unfolding lemmas *)
Lemma lft_vs_inv_go_ne κ (f f' : ∀ κ', κ' ⊂ κ → monPred) Vs I n :
  (∀ κ' (Hκ : κ' ⊂ κ), f κ' Hκ ≡{n}≡ f' κ' Hκ) →
  lft_vs_inv_go κ f Vs I ≡{n}≡ lft_vs_inv_go κ f' Vs I.
Proof.
  intros Hf. apply sep_ne, sep_ne, big_opS_ne=> // κ'.
  apply forall_ne=> Hκ. by rewrite Hf.
Qed.

Lemma lft_vs_go_ne κ (f f' : ∀ κ', κ' ⊂ κ → monPred) Pb Pb' Pi Pi' n :
  (∀ κ' (Hκ : κ' ⊂ κ), f κ' Hκ ≡{n}≡ f' κ' Hκ) →
  Pb ≡{n}≡ Pb' → Pi ≡{n}≡ Pi' →
  lft_vs_go κ f Pb Pi ≡{n}≡ lft_vs_go κ f' Pb' Pi'.
Proof.
  intros Hf HPb HPi. unfold lft_vs_go. do 14 f_equiv.
  - by apply lft_vs_inv_go_ne.
  - do 4 f_equiv=>//; [by apply lft_vs_inv_go_ne|]. by repeat f_equiv.
Qed.

Lemma lft_inv_alive_go_ne κ (f f' : ∀ κ', κ' ⊂ κ → monPred) n :
  (∀ κ' (Hκ : κ' ⊂ κ), f κ' Hκ ≡{n}≡ f' κ' Hκ) →
  lft_inv_alive_go κ f ≡{n}≡ lft_inv_alive_go κ f'.
Proof.
  intros Hf. apply exist_ne=> Pb; apply exist_ne=> Pi. by rewrite lft_vs_go_ne.
Qed.

Lemma lft_inv_alive_unfold κ :
  lft_inv_alive κ ⊣⊢ ∃ Pb Pi,
    lft_bor_alive κ Pb ∗ lft_vs κ Pb Pi ∗ ⎡lft_inh κ None Pi⎤.
Proof.
  apply equiv_dist=> n. rewrite /lft_inv_alive -Fix_F_eq.
  apply lft_inv_alive_go_ne=> κ' Hκ.
  apply (Fix_F_proper _ (λ _, dist n)).
  unfold pointwise_relation. auto using lft_inv_alive_go_ne.
Qed.
Lemma lft_vs_inv_unfold κ (I : gmap lft lft_names) Vs :
  lft_vs_inv κ Vs I ⊣⊢ lft_bor_dead κ ∗
    own_ilft_auth I ∗
    [∗ set] κ' ∈ dom _ I, ⌜κ' ⊂ κ⌝ → lft_inv_alive κ' (Vs κ').
Proof.
  rewrite /lft_vs_inv /lft_vs_inv_go. by setoid_rewrite pure_impl_forall.
Qed.
Lemma lft_vs_unfold κ Pb Pi :
  lft_vs κ Pb Pi ⊣⊢ ∃ (Vκ : Lat) (n : nat), monPred_in Vκ ∗
    ⎡own_cnt κ (● n) ∗
     ∀ (I : gmap lft lft_names) (Vs : lft → Lat),
       ⌜Vκ ⊑ Vs κ⌝ -∗
       ⌜∀ κ', κ' ∈ dom (gset _) I → κ' ⊆ κ → Vs κ ⊑ Vs κ'⌝ -∗
       lft_vs_inv κ Vs I -∗ ▷ Pb (Vs κ) -∗ (∃ V, lft_dead κ V) ={↑borN ∪ E0}=∗
       lft_vs_inv κ Vs I ∗ ▷ Pi (Vs κ) ∗ own_cnt κ (◯ n)⎤.
Proof. done. Qed.

Global Instance own_bor_proper κ : Proper ((≡) ==> (≡)) (own_bor κ).
Proof. solve_proper. Qed.
Global Instance own_cnt_proper κ : Proper ((≡) ==> (≡)) (own_cnt κ).
Proof. solve_proper. Qed.
Global Instance own_inh_proper κ : Proper ((≡) ==> (≡)) (own_inh κ).
Proof. solve_proper. Qed.

Global Instance lft_bor_alive_ne κ : NonExpansive (lft_bor_alive κ).
Proof. solve_proper. Qed.
Global Instance lft_bor_alive_proper κ : Proper ((≡) ==> (≡)) (lft_bor_alive κ).
Proof. apply ne_proper, _. Qed.

Global Instance lft_inh_ne κ s : NonExpansive (lft_inh κ s).
Proof. solve_proper. Qed.
Global Instance lft_inh_proper κ s : Proper ((≡) ==> (≡)) (lft_inh κ s).
Proof. apply (ne_proper _). Qed.

Global Instance lft_vs_ne κ : NonExpansive2 (lft_vs κ).
Proof. intros P P' HP Q Q' HQ. by apply lft_vs_go_ne. Qed.
Global Instance lft_vs_proper κ : Proper ((≡) ==> (≡) ==> (≡)) (lft_vs κ).
Proof. apply (ne_proper_2 _). Qed.

Global Instance idx_bor_ne κ i n : Proper (dist n ==> dist n) (idx_bor κ i).
Proof. solve_proper. Qed.
Global Instance idx_bor_contractive κ i : Contractive (idx_bor κ i).
Proof. solve_contractive. Qed.
Global Instance idx_bor_proper κ i : Proper ((≡) ==> (≡)) (idx_bor κ i).
Proof. apply (ne_proper _). Qed.

Global Instance raw_bor_ne i n : Proper (dist n ==> dist n) (raw_bor i).
Proof. solve_proper. Qed.
Global Instance raw_bor_contractive i : Contractive (raw_bor i).
Proof. solve_contractive. Qed.
Global Instance raw_bor_proper i : Proper ((≡) ==> (≡)) (raw_bor i).
Proof. apply (ne_proper _). Qed.

Global Instance bor_ne κ n : Proper (dist n ==> dist n) (bor κ).
Proof. solve_proper. Qed.
Global Instance bor_contractive κ : Contractive (bor κ).
Proof. solve_contractive. Qed.
Global Instance bor_proper κ : Proper ((≡) ==> (≡)) (bor κ).
Proof. apply (ne_proper _). Qed.

(*** PersistentP and TimelessP and instances  *)
Global Instance lft_tok_timeless q κ : Timeless q.[κ].
Proof. rewrite /lft_tok. apply _. Qed.

Global Instance lft_dead_persistent κ : Persistent [†κ].
Proof. rewrite /lft_dead. apply _. Qed.
Global Instance lft_dead_timeless κ : Timeless [†κ].
Proof. rewrite /lft_dead. apply _. Qed.

Global Instance lft_incl_persistent κ κ' : Persistent (κ ⊑ κ').
Proof. rewrite /lft_incl. apply _. Qed.

Global Instance idx_bor_persistent κ i P : Persistent (&{κ,i}P).
Proof. rewrite /idx_bor. apply _. Qed.
Global Instance idx_bor_own_timeless i : Timeless (idx_bor_own i).
Proof. rewrite /idx_bor_own. apply _. Qed.

Global Instance lft_ctx_persistent : Persistent lft_ctx.
Proof. rewrite /lft_ctx. apply _. Qed.
End basic_properties.
