From Coq Require Import Qcanon.
From lrust.typing Require Export type.
Set Default Proof Using "Type".

Section util.
  Context `{!typeG Σ}.

  (* Delayed sharing is used by various types; in particular own and uniq.
     It comes in two flavors: Borrows of "later something" and borrows of
     "borrowed something".
     TODO: Figure out a nice way to generalize that; the two proofs below are too
     similar. *)

  (* This is somewhat the general pattern here... but it doesn't seem
     easy to make this usable in Coq, with the arbitrary quantifiers
     and things.  Also, it actually works not just for borrows but for
     anything that you can split into a timeless and a persistent
     part.

  Lemma delay_borrow_step :
    lfeE ⊆ N → (∀ x, Persistent (Post x)) →
    lft_ctx -∗ &{κ} P -∗
      □ (∀ x, &{κ} P -∗ Pre x -∗ Frame x ={F1 x,F2 x}▷=∗ Post x ∗ Frame x) ={N}=∗ 
      □ (∀ x, Pre x -∗ Frame x ={F1 x,F2 x}▷=∗ Post x ∗ Frame x).
   *)

  Lemma delay_sharing_later N κ l ty tid :
    lftE ⊆ N →
    ⎡lft_ctx⎤ -∗ &{κ}(▷ l ↦∗: ty_own ty tid) ={N}=∗
       □ ∀ (F : coPset) (q : Qp),
       ⌜↑shrN ∪ lftE ⊆ F⌝ -∗ (q).[κ] ={F,F ∖ ↑shrN}▷=∗ ty.(ty_shr) κ tid l ∗ (q).[κ].
  Proof.
    iIntros (?) "#LFT Hbor". rewrite bor_unfold_idx.
    iDestruct (monPred_in_intro with "Hbor") as (V) "[#Hin Hbor]".
    iDestruct "Hbor" as (i) "(#Hpb&Hpbown)".
    iMod (inv_alloc shrN _ (idx_bor_own i V ∨ ty_shr ty κ tid l V)%I
          with "[Hpbown]") as "#Hinv".
    { iModIntro. eauto. }
    iIntros "!> !# * % Htok".
    iMod (inv_open with "Hinv") as "[INV Hclose]"; first solve_ndisj.
    iDestruct "INV" as "[>Hbtok|#Hshr]".
    - rewrite -(Qp_div_2 q).
      iDestruct (lft_tok_split_obj with "Htok") as "[$ Htok]".
      rewrite monPred_objectively_unfold. iSpecialize ("Htok" $! V).
      iApply (monPred_in_elim with "Hin"). iModIntro.
      iMod (bor_later_tok with "LFT [Hbtok] Htok") as "Hdelay"; first solve_ndisj.
      { rewrite bor_unfold_idx. iExists _. auto. }
      iModIntro. iNext. iMod "Hdelay" as "[Hb Htok]".
      iMod (ty.(ty_share) with "LFT Hb Htok") as "[#$ $]"; first solve_ndisj.
      iApply "Hclose". auto.
    - iMod fupd_intro_mask' as "Hclose'"; first solve_ndisj. iModIntro.
      iNext. iMod "Hclose'" as "_". iMod ("Hclose" with "[]") as "_".
      + iModIntro. eauto.
      + iFrame. iModIntro. by iApply monPred_in_elim.
  Qed.

  Lemma delay_sharing_nested N κ κ' κ'' l ty tid :
    lftE ⊆ N →
    ⎡lft_ctx⎤ -∗ ▷ (κ'' ⊑ κ ⊓ κ') -∗ &{κ'}(&{κ}(l ↦∗: ty_own ty tid)) ={N}=∗
       □ ∀ (F : coPset) (q : Qp),
       ⌜↑shrN ∪ lftE ⊆ F⌝ -∗ (q).[κ''] ={F,F ∖ ↑shrN}▷=∗ ty.(ty_shr) κ'' tid l ∗ (q).[κ''].
  Proof.
    iIntros (?) "#LFT Hincl Hbor". rewrite bor_unfold_idx.
    iCombine "Hincl Hbor" as "Hinclbor".
    iDestruct (monPred_in_intro with "Hinclbor") as (V) "(#Hin & #Hincl & Hbor)".
    iDestruct "Hbor" as (i) "(#Hpb&Hpbown)".
    iMod (inv_alloc shrN _ (idx_bor_own i V ∨ ty_shr ty κ'' tid l V)%I
          with "[Hpbown]") as "#Hinv".
    { iModIntro. auto. }
    iIntros "!> !# * % Htok".
    iMod (inv_open with "Hinv") as "[INV Hclose]"; first solve_ndisj.
    iDestruct "INV" as "[>Hbtok|#Hshr]".
    - rewrite -(Qp_div_2 q).
      iDestruct (lft_tok_split_obj with "Htok") as "[$ Htok]".
      rewrite monPred_objectively_unfold. iSpecialize ("Htok" $! V).
      iApply (monPred_in_elim with "Hin"). iModIntro.
      iMod (bor_unnest with "LFT [Hbtok]") as "Hb"; first solve_ndisj.
      { iApply bor_unfold_idx. eauto. }
      iModIntro. iNext. iMod "Hb".
      iMod (ty.(ty_share) with "LFT [Hb] Htok") as "[#Hshr $]"; first solve_ndisj.
      { by iApply bor_shorten. }
      iMod ("Hclose" with "[]") as "_"; auto.
    - iMod fupd_intro_mask' as "Hclose'"; last iModIntro; first solve_ndisj.
      iNext. iMod "Hclose'" as "_". iMod ("Hclose" with "[]") as "_".
      + iModIntro. eauto.
      + iFrame. iModIntro. by iApply monPred_in_elim.
  Qed.
End util.
