From Coq.QArith Require Import Qcanon.
From iris.program_logic Require Import weakestpre.
From iris.proofmode Require Import tactics.
From iris.algebra Require Import ofe.
From iris.bi.lib Require Import fractional.
From gpfsl.lang Require Export lifting proofmode.
From gpfsl.gps Require Export gps_cmra.
From gpfsl.gps Require Import model.
From gpfsl Require Import vprop.

Set Default Proof Using "Type".

Lemma alloc_local_singleton_seen_local l t m V:
  alloc_local l {[t := m]} V → seen_local l t V.
Proof. destruct 1 as [?[?[[? _]%lookup_singleton_Some [_ ?]]]]. by subst. Qed.

Notation gpsExAgG Σ := (inG Σ (agreeR (leibnizC time))).

Section SimpleRules.
  Context `{!noprolG Σ, !gpsG Σ Prtcl} (IP : interpC Σ Prtcl) (l : loc).
  Local Notation iProp := (iProp Σ).
  Local Notation vProp := (vProp Σ).

  Implicit Types
           (s : pr_stateT Prtcl) (ζ : stateT Prtcl) (e : eventT Prtcl)
           (t : time) (v : val) (V : view) (E : coPset) (q: Qp).

  Notation F := (IP true l).        (* full interp *)
  Notation F_read := (IP false l).  (* read interp *)

  Program Definition gps_inv_vProp IW γ γF b := MonPred (gps_inv IP IW l γ γF b) _.
  Next Obligation. rewrite /gps_inv. solve_proper. Qed.

  Section SingleWriter.
    Definition GPS_INV_def IW γ : vProp := gps_inv_vProp IW γ γ true.
    Definition GPS_INV_aux : seal (@GPS_INV_def). by eexists. Qed.
    Definition GPS_INV := unseal (@GPS_INV_aux).
    Definition GPS_INV_eq : @GPS_INV = _ := seal_eq _.

    (* DEALLOC *)
    Lemma GPS_INV_dealloc IW b γ E (SUB: ↑histN ⊆ E):
      ⎡ hist_inv ⎤ -∗ ▷?b GPS_INV IW γ ={E}=∗
        ∃ t s v, l ↦ v ∗ ▷?b F γ t s v.
    Proof.
      constructor. iIntros (?) "HInv". iIntros (? _) "Inv". rewrite GPS_INV_eq /=.
      by iApply (GPSRaw_dealloc with "HInv Inv").
    Qed.

    Definition GPS_SWReader_def t s v γ : vProp := (PrtSeen l t s v γ)%I.
    Definition GPS_SWReader_aux : seal (@GPS_SWReader_def). by eexists. Qed.
    Definition GPS_SWReader := unseal (@GPS_SWReader_aux).
    Definition GPS_SWReader_eq : @GPS_SWReader = _ := seal_eq _.

    Global Instance GPS_SWReader_persistent : Persistent (GPS_SWReader t s v γ).
    Proof. rewrite GPS_SWReader_eq. by apply _. Qed.
    Global Instance GPS_SWReader_timeless : Timeless (GPS_SWReader t s v γ).
    Proof. rewrite GPS_SWReader_eq. by apply _. Qed.

    Program Definition GPS_SWWriter_def t s v γ : vProp :=
      (∃ ζ, PrtSeenView l t s v γ
            ∗ ⎡ Writer γ ζ ∗ exwr γ t 1%Qp ⎤
            ∗ ⌜∀ t', is_Some (ζ !! t') → t' ⊑ t⌝)%I.
    Definition GPS_SWWriter_aux : seal (@GPS_SWWriter_def). by eexists. Qed.
    Definition GPS_SWWriter := unseal (@GPS_SWWriter_aux).
    Definition GPS_SWWriter_eq : @GPS_SWWriter = _ := seal_eq _.

    Global Instance GPS_SWWriter_timeless : Timeless (GPS_SWWriter t s v γ).
    Proof. rewrite GPS_SWWriter_eq. by apply _. Qed.

    (* SW DEALLOC *)
    Lemma GPS_SWWriter_dealloc IW b t s v γ E (SUB: ↑histN ⊆ E) :
      ⎡ hist_inv ⎤ -∗ ▷?b GPS_INV IW γ -∗ GPS_SWWriter t s v γ ={E}=∗
        l ↦ v ∗ ▷?b F γ t s v.
    Proof.
      constructor. iIntros (?) "HInv". iIntros (? _) "Inv". iIntros (V->) "W".
      rewrite GPS_INV_eq GPS_SWWriter_eq /=.
      iDestruct "W" as (ζ) "(PS & (W & ex) & MAX)". iDestruct "MAX" as %MAX.
      iDestruct "PS" as (Vv Va ?) "(na & R & SL)". iDestruct "SL" as %[SL LeV].
      by iApply (GPSRaw_SW_dealloc _ l _ b ζ t s v with "HInv Inv W R na").
    Qed.

    Lemma GPS_SWWriter_Reader t s v γ:
      GPS_SWWriter t s v γ -∗ GPS_SWReader t s v γ.
    Proof.
      rewrite GPS_SWWriter_eq GPS_SWReader_eq.
      iDestruct 1 as (ζ) "(S & _)". by iApply PrtSeenView_PrtSeen.
    Qed.

    Lemma GPS_SWReaders_agree IW t1 t2 s1 s2 v1 v2 γ Vb E:
      GPS_SWReader t1 s1 v1 γ -∗ GPS_SWReader t2 s2 v2 γ
      -∗ (monPred_in Vb → ▷ GPS_INV IW γ)
      ={E}=∗ (monPred_in Vb → ▷ GPS_INV IW γ) ∗
              ⌜(s1 ⊑ s2 ∧ t1 ⊑ t2) ∨ (s2 ⊑ s1 ∧ t2 ⊑ t1)⌝.
    Proof.
      rewrite GPS_SWReader_eq. iStartProof iProp.
      iIntros (?). iDestruct 1 as (???) "(_ & R1 & _)".
      iIntros (??). iDestruct 1 as (???) "(_ & R2 & _)".
      iIntros (V ?) "Inv". set Vo := V ⊔ Vb.
      iAssert ((▷ GPS_INV IW γ) Vo)%I with "[Inv]" as "Inv".
      { iApply "Inv". iPureIntro; solve_lat. }
      rewrite GPS_INV_eq monPred_at_later.
      iMod (Reader_singleton_agree with "Inv R1 R2") as "[Inv $]".
      iModIntro. iIntros (V' ??).
      rewrite -(monPred_mono _ Vo V'); [done|solve_lat].
    Qed.

    Lemma GPS_SWWriter_exclusive t1 t2 s1 s2 v1 v2 γ :
      GPS_SWWriter t1 s1 v1 γ -∗ GPS_SWWriter t2 s2 v2 γ -∗ False.
    Proof.
      rewrite GPS_SWWriter_eq.
      iDestruct 1 as (?) "(_ & (W1 & _) & _)".
      iDestruct 1 as (?) "(_ & (W2 & _) & _)".
      by iDestruct (Writer_Exclusive with "W1 W2") as %?.
    Qed.

    Lemma GPS_SWWriter_latest IW t t' s s' v v' γ Vb E:
      GPS_SWWriter t s v γ -∗ GPS_SWReader t' s' v' γ
      -∗ (monPred_in Vb → ▷ GPS_INV IW γ)
      ={E}=∗ GPS_SWWriter t s v γ ∗ (monPred_in Vb → ▷ GPS_INV IW γ) ∗
            ⌜s' ⊑ s ∧ t' ⊑ t⌝.
    Proof.
      rewrite GPS_SWWriter_eq GPS_SWReader_eq. iStartProof iProp.
      iIntros (?). iDestruct 1 as (ζ) "(Seen1 & (W & ox) & MAX)".
      iIntros (? ->). iDestruct 1 as (???) "(_ & R2 & _)".
      iIntros (V ->) "Inv".
      set Vo := V ⊔ Vb.
      iAssert ((▷ GPS_INV IW γ) Vo)%I with "[Inv]" as "Inv".
      { iApply "Inv". iPureIntro; solve_lat. }
      rewrite GPS_INV_eq monPred_at_later.
      iDestruct "Seen1" as (V1 Va1 ?) "(naV & R1 & F1)".
      iMod (Writer_stsorted with "Inv W") as "(Inv & W & S)".
      iDestruct "S" as %SORTED. iDestruct "MAX" as %MAX.
      iDestruct "Inv" as (μ C tx Va ???) "(hist & at & aw & na & >oA & REST)".
      iDestruct (gps_ownA_Reader_singleton_sub with "oA R1") as %HR1.
      iDestruct (gps_ownA_Reader_singleton_sub with "oA R2") as %HR2.
      iDestruct (Writer_exact_2 with "oA W") as %?.
      iModIntro. iSplitR "hist at aw na oA REST"; last iSplit.
      - iExists ζ. iFrame (MAX) "W ox". iExists V1, Va1, _. by iFrame.
      - iIntros (V' ??). rewrite -(monPred_mono _ Vo V'); [|solve_lat].
        iExists μ, C, tx, Va, _, _, _. by iFrame.
      - iPureIntro. subst ζ. have ?: t' ⊑ t. by apply MAX; eexists.
        split; [by apply (SORTED _ _ _ _ HR2 HR1)|done].
    Qed.

    (* INIT *)
    Lemma GPS_SWRaw_Init_vs IW b s v:
      l ↦ v -∗ (∀ t γ, ▷?b F γ t s v)
      ==∗ ∃ γ t, GPS_SWWriter t s v γ ∗ ▷?b GPS_INV IW γ.
    Proof.
      iStartProof iProp. iIntros (?) "own".
      iIntros (V->) "F".
      rewrite own_loc_na_eq GPS_SWWriter_eq.
      iDestruct "own" as (t m) "[own [% LE]]".
      iDestruct "own" as (???) "[[ALL [% AT]] [own [at [aw na]]]]".
      iDestruct "AT" as %(AT & AW & NA).
      iDestruct "ALL" as %ALL. iDestruct "LE" as %LE. iSpecialize ("F" $! t).
      iMod (GPSRaw_Init_vs IP _ _ b t v s m with "own at aw na F")
        as (γ) "(Inv & ox & W & R & na)"; [done..|].
      iModIntro. iExists γ, t. iSplitR "Inv".
      - iExists _. iFrame "W ox". iSplit.
        + iExists _, _, _. iFrame "R na". iPureIntro. split; [|by rewrite LE].
          by eapply alloc_local_singleton_seen_local.
        + iPureIntro. by move => t' [? / lookup_singleton_Some [-> ]].
      - iNext.
        have Ext2: default ∅ (mrel m) ⊔ {[l := [{t,∅,∅,∅}]]} ⊔ V ⊑ V.
        { apply lat_join_lub; [apply lat_join_lub|done]; [by rewrite LE|].
          destruct ALL as [t' [m' [[??]%lookup_singleton_Some [ND SL]]]].
          subst t' m'. by eapply seen_local_view_sqsubseteq. }
        rewrite GPS_INV_eq. iApply (monPred_mono _ _ _ Ext2). by iFrame.
    Qed.

    Lemma GPS_SWRaw_Init IW b v s tid E:
      ↑histN ⊆ E →
      {{{ l ↦ ? ∗ (∀ t γ, ▷?b F γ t s v) }}}
        #l <- v in tid @ E
      {{{ γ t, RET #☠; GPS_SWWriter t s v γ ∗ ▷?b GPS_INV IW γ }}}.
    Proof.
      iIntros (? Φ) "(own & F) Post". iApply wp_fupd. wp_write.
      iMod (GPS_SWRaw_Init_vs with "own F") as (γ t) "?". by iApply "Post".
    Qed.

    Lemma GPS_SWRaw_Init_vs_strong IW b v s (Q: time → gname → vProp) E :
      l ↦ v -∗
      (∀ t γ, GPS_SWWriter t s v γ ={E}=∗ ▷?b F γ t s v ∗ Q t γ )
      ={E}=∗ ∃ γ t, ▷?b GPS_INV IW γ ∗ Q t γ.
    Proof.
      iStartProof iProp. iIntros (?) "own".
      iIntros (V->) "F". rewrite own_loc_na_eq.
      iDestruct "own" as (t m) "[own [% LE]]".
      iDestruct "own" as (???) "[[ALL [% AT]] [own [at [aw na]]]]".
      iDestruct "AT" as %(AT & AW & NA).
      iDestruct "ALL" as %ALL. iDestruct "LE" as %LE.
      have SL:= alloc_local_singleton_seen_local _ _ _ _ ALL.
      iMod (GPSRaw_Init_strong_vs IP _ _ b t v s m _ V _ _ _ id (Q t)
            with "own at aw na [F]") as (γ) "(Inv & Q)"; [done..| |].
      { iIntros (γ) "ex W R na". simpl. setoid_rewrite monPred_at_wand.
        iMod ("F" $! t γ V with "[//] [-]") as "$"; [|done].
        rewrite GPS_SWWriter_eq.
        iExists _. iFrame "ex W". iSplit.
        - iExists _,_,_. iFrame "na R". iPureIntro. split; [done|by rewrite LE].
        - iPureIntro. move => ? [? /lookup_singleton_Some [-> //]]. }
      iModIntro. iExists γ, t. rewrite GPS_INV_eq. by iFrame.
    Qed.

    (* SW READ *)
    Lemma GPS_SWRaw_SWRead IW R o t s v γ Vb tid E:
      ↑histN ⊆ E → o = Relaxed ∨ o = AcqRel →
      {{{ GPS_SWWriter t s v γ ∗ (monPred_in Vb → ▷ GPS_INV IW γ)
        ∗ ▷ <obj> (F γ t s v ={E}=∗ F γ t s v ∗ R) }}}
        Read o #l in tid @ E
      {{{ RET v; GPS_SWWriter t s v γ ∗ (monPred_in Vb → ▷ GPS_INV IW γ) ∗ R }}}.
    Proof.
      iIntros (SUB OR Φ). iStartProof iProp. iIntros (V1) "(W & Inv & D)".
      set Vo := V1 ⊔ Vb.
      iAssert ((▷ GPS_INV IW γ) Vo)%I with "[Inv]" as "Inv".
      { iApply "Inv". iPureIntro; solve_lat. } rewrite GPS_INV_eq monPred_at_later.
      iIntros (V2 ?) "Post". rewrite wp_eq. iIntros (𝓥3 Ext3) "H𝓥3 s".
      have Ext3' : V1 ⊑ 𝓥3.(cur) by etrans. rewrite 2!(monPred_mono _ _ _ Ext3').
      rewrite GPS_SWWriter_eq.
      iDestruct "W" as (ζ) "(Prt & (W & ox) & MAX)". iDestruct "MAX" as %MAX.
      iDestruct "Prt" as (V Va ?) "(#naV & #R & SL)". iDestruct "SL" as %[SL LE].
      iSpecialize ("D" $! ∅).
      iApply program_logic.weakestpre.wp_fupd.
      iApply (GPSRaw_ReadSW IP l _ R t s v V o ζ γ _ _ _ Vo Va _ 𝓥3 _
                SL LE MAX SUB OR with "[$s $D $Inv $R $naV $ox $W]").
      iNext. iIntros (𝓥4) "(s' & Ext4 & Inv & ox & W & Fp)".
      iDestruct "Ext4" as %Ext4.
      iMod (own_lat_auth_update with "H𝓥3") as "[$ #H𝓥4]"; [done|]. iModIntro.
      assert (V2 ⊑ 𝓥4.(cur)) by (etrans; [eassumption|by f_equiv]).
      iApply "Post". iFrame "Fp". iSplitR "Inv".
      - iExists ζ. iFrame (MAX) "W ox". iExists V, Va, _.
        rewrite (monPred_mono _ _ _ (cur_mono _ _ Ext4)).
        iFrame "naV R". iPureIntro. split; [|by rewrite -Ext4].
        eapply seen_local_mono; [apply Ext4|done].
      - iIntros (V' Ext' ?). iNext.
        rewrite -(monPred_mono _ (Vo ⊔ 𝓥4.(cur)) V'); [done|].
        apply lat_join_lub; [|done].
        rewrite /Vo. apply lat_join_lub; [by rewrite -Ext' -Ext4|done].
    Qed.

    (* Reader READ *)
    Lemma GPS_SWRaw_Read IW R o t s v γ tid Vb E:
      ↑histN ⊆ E → o = Relaxed ∨ o = AcqRel →
      {{{ GPS_SWReader t s v γ ∗ (monPred_in Vb → ▷ GPS_INV IW γ)
          ∗ (▷ ∀ t' s' v', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
               <obj> ((F_read γ t' s' v' ={E}=∗ F_read γ t' s' v' ∗ R t' s' v') ∧
                      (F γ t' s' v' ={E}=∗ F γ t' s' v' ∗ R t' s' v') ∧
                      (IW l γ t' s' v' ={E}=∗ IW l γ t' s' v' ∗ R t' s' v'))) }}}
        Read o #l in tid @ E
      {{{ t' s' v', RET v'; ⌜s ⊑ s' ∧ t ⊑ t'⌝
          ∗ GPS_SWReader t' s' v' γ ∗ (monPred_in Vb → ▷ GPS_INV IW γ)
          ∗ if (decide (AcqRel ⊑ o)) then R t' s' v' else ▽{tid} R t' s' v' }}}.
    Proof.
      iIntros (SUB OR Φ). iStartProof iProp. iIntros (V1) "(R & Inv & D)".
      set Vo := V1 ⊔ Vb.
      iAssert ((▷ GPS_INV IW γ) Vo)%I with "[Inv]" as "Inv".
      { iApply "Inv". iPureIntro; solve_lat. }
      rewrite GPS_INV_eq monPred_at_later wp_eq.
      iIntros (V2 ?) "Post". iIntros (𝓥3 Ext3) "H𝓥3 s".
      have Ext3' : V1 ⊑ 𝓥3.(cur) by etrans.
      rewrite {1}(monPred_mono _ _ _ Ext3') GPS_SWReader_eq.
      iDestruct "R" as (V Va ?) "(#naV & R & SL)". iDestruct "SL" as %SL.

      iApply program_logic.weakestpre.wp_fupd.
      iApply (GPSRaw_Read_non_ex IP l _ R t s (v,V) o γ _ _ Vo Va _ _ E SL SUB OR
        with "[$s D $Inv $R $naV]").
      { iNext. iIntros (t' s' v' Ext'). iSpecialize ("D" $! t' s' v' Ext').
        rewrite monPred_at_objectively. by iApply "D". }

      iNext. iIntros (t' s' V' 𝓥4 v') "(s' & Ext4 & Inv & Fp & R)".
      iDestruct "R" as (Vt') "[#R %]".
      iDestruct "Ext4" as %(ExtV & Exts & Extt & SL' & LeV').
      iMod (own_lat_auth_update with "H𝓥3") as "[$ #H𝓥4]"; [done|]. iModIntro.
      assert (V2 ⊑ 𝓥4.(cur)) by (etrans; [eassumption|by f_equiv]).
      iApply ("Post" $! t' s'). iSplitL""; [done|]. iSplitL ""; [|iSplitL "Inv"].
      - iExists Vt', Va, _. rewrite (monPred_mono _ _ _ (cur_mono _ _ ExtV)).
        iFrame (SL') "R naV".
      - iIntros (V0 Ext0 ?). iNext.
        rewrite -(monPred_mono _ (Vo ⊔ 𝓥4.(cur)) V0); [done|].
        apply lat_join_lub; [|done].
        rewrite /Vo. apply lat_join_lub; [by rewrite -Ext0 -ExtV|done].
      - rewrite acq_mod_eq. destruct OR; subst o; simpl; [iExists _|]; by iFrame.
    Qed.

    (* SW WRITE *)
    Lemma GPS_SWRaw_SWWrite IW R
      o t s s' v v' γ tid Vb E (Exts: s ⊑ s'):
      ↑histN ⊆ E → o = Relaxed ∨ o = AcqRel →
      let WVS: vProp := (∀ t', ⌜(t < t')%positive⌝ ={E}=∗ F γ t' s' v')%I in
      {{{ GPS_SWWriter t s v γ ∗ (monPred_in Vb → ▷ GPS_INV IW γ)
          ∗ ▷ (if decide (AcqRel ⊑ o) then WVS else △{tid} WVS)
          ∗ ▷ <obj> (F γ t s v ={E}=∗ IW l γ t s v ∗ R)  }}}
        Write o #l v' in tid @ E
      {{{ t', RET #☠; ⌜(t < t')%positive⌝ ∗ GPS_SWWriter t' s' v' γ ∗
          (monPred_in Vb → ▷ GPS_INV IW γ) ∗ R }}}.
    Proof.
      iIntros (SUB OW WVS Φ). iStartProof iProp. iIntros (V1) "(W & Inv & F & IW)".
      set Vo := V1 ⊔ Vb.
      iAssert ((▷ GPS_INV IW γ) Vo)%I with "[Inv]" as "Inv".
      { iApply "Inv". iPureIntro; solve_lat. }
      rewrite GPS_INV_eq monPred_at_later wp_eq.
      iIntros (V2 ?) "Post". iIntros (𝓥3 Ext3) "H𝓥3 s".
      have Ext3' : V1 ⊑ 𝓥3.(cur) by etrans.
      rewrite !(monPred_mono _ _ _ Ext3') GPS_SWWriter_eq.
      iDestruct "W" as (ζ) "(Prt & (W & ox) & MAX)". iDestruct "MAX" as %MAX.
      iDestruct "Prt" as (Vx Va ?) "(#naV & #R & SL)". iDestruct "SL" as %[SL LE].

      iApply program_logic.weakestpre.wp_fupd.
      iApply (GPSRaw_WriteSW IP l _ t s v Vx _ _ s' ζ γ _ _ _ _ _ _
                SL Exts MAX SUB OW with "[$s $Inv $R $naV $ox $W]").
      iNext. iIntros (t' V' Vt' 𝓥4) "(s' & H & Inv & Fx & ox' & R' & W' & M')".
      iDestruct "M'" as %(? & MAX2).
      iDestruct "H" as %(Ext5 & ? & Let & SL' & ? & ? & LeV & ? & ?).
      assert (V2 ⊑ 𝓥4.(cur)) by (etrans; [eassumption|by f_equiv]).
      iAssert (⌜Va ⊑ 𝓥3.(cur)⌝)%I as %LVa3.
      { rewrite naViewSeen_eq. by iDestruct "naV" as "[_ [$ _]]". }
      have Ext6: Vx ⊔ Va ⊑ 𝓥4.(cur) by rewrite -Ext5; solve_lat.
      iAssert (|={E}=> gps_inv IP IW l γ γ true (Vo ⊔ 𝓥4.(cur)) ∗
                       own tid (● to_latT 𝓥3) ∗ R 𝓥4.(cur))%I
        with "[Inv H𝓥3 Fx F IW]" as ">(Inv & H𝓥3 & Rs)".
      { iMod ("IW" $! (Vx ⊔ Va) with "Fx") as "[I $]".
         destruct OW; subst o; simpl.
        - rewrite rel_mod_eq /rel_mod_def. iDestruct "F" as (𝓥F) "[H𝓥F F]".
          iDestruct (own_lat_auth_max with "[$H𝓥3 $H𝓥F]") as %H𝓥F.
          iClear "H𝓥F". rewrite (monPred_mono _ _ _ (frel_mono _ _ H𝓥F)).
          assert (Le3: 𝓥3.(frel) ⊑ V').
          { etrans; [|by apply LeV]. simpl. by f_equiv. }
          rewrite (monPred_mono _ _ _ Le3).
          iFrame. iMod ("F" $! t' Let) as "F". by iApply ("Inv" with "F I").
        - assert (Le3: 𝓥3.(cur) ⊑ V').
          { etrans; [|by apply LeV]. simpl. by f_equiv. }
          rewrite (monPred_mono _ _ _ Le3).
          iFrame. iMod ("F" $! t' Let) as "F". by iApply ("Inv" with "F I"). }
      iMod (own_lat_auth_update with "H𝓥3") as "[$ _]"; [done|].
      iApply "Post". iModIntro. iFrame (Let) "Rs". iSplitL "W' ox' R'".
      - iExists _. iFrame (MAX2) "W' ox'".
        iExists Vt', Va, _. rewrite (monPred_mono _ _ _ (cur_mono _ _ Ext5)).
        by iFrame "naV R'".
      - iIntros (V0 Ext0 ?). iNext.
        rewrite -(monPred_mono _ (Vo ⊔ 𝓥4.(cur)) V0); [done|].
        apply lat_join_lub; [|done]. apply lat_join_lub; [by rewrite Ext3' Ext5|done].
    Qed.

    (* SW WRITE RELEASE itself *)
    Lemma GPS_SWRaw_SWWrite_rel_view
      IW Q Q1 Q2 P t s s' v v' γ tid Vb E (Exts: s ⊑ s'):
      ↑histN ⊆ E →
      let WVS: vProp :=
        (∀ t', ⌜(t < t')%positive⌝ -∗ GPS_SWWriter t' s' v' γ -∗ Q2 -∗ P
                  ={E}=∗ (<obj> (Q1 ={E}=∗ IW l γ t s v)) ∗
                         |={E}=> (F γ t' s' v' ∗ Q t'))%I in
      {{{ GPS_SWWriter t s v γ ∗ (monPred_in Vb → ▷ GPS_INV IW γ)
          ∗ ▷ WVS ∗ ▷ P ∗ ▷ <obj> (F γ t s v ={E}=∗ Q1 ∗ Q2) }}}
        Write AcqRel #l v' in tid @ E
      {{{ t' V', RET #☠; ⌜(t < t')%positive⌝ ∗ GPS_SWReader t' s' v' γ ∗
            monPred_in V' ∗ ⎡ P V' ⎤ ∗
            (⎡ P V' ⎤ ={E}=∗ ⎡ (monPred_in Vb → ▷ GPS_INV IW γ) V' ⎤ ∗ Q t') }}}.
    Proof.
      iIntros (SUB WVS Φ). iStartProof iProp. iIntros (V1) "(W & Inv & F & P & IW)".
      set Vo := V1 ⊔ Vb.
      iAssert ((▷ GPS_INV IW γ) Vo)%I with "[Inv]" as "Inv".
      { iApply "Inv". iPureIntro; solve_lat. }
      rewrite GPS_INV_eq monPred_at_later wp_eq.
      iIntros (V2 ?) "Post". iIntros (𝓥3 Ext3) "H𝓥3 s".
      have Ext3' : V1 ⊑ 𝓥3.(cur) by etrans. rewrite !(monPred_mono _ _ _ Ext3').
      rewrite {1}GPS_SWWriter_eq.
      iDestruct "W" as (ζ) "(Prt & (W & ox) & MAX)". iDestruct "MAX" as %MAX.
      iDestruct "Prt" as (Vx Va ?) "(#naV & #R & SL)". iDestruct "SL" as %[SL LE].
      iApply program_logic.weakestpre.wp_fupd.
      iApply (GPSRaw_WriteSW IP l IW t s v Vx _ _ s' ζ γ _ _ _ _ _ _
                SL Exts MAX SUB with "[$s $Inv $R $naV $ox $W]"); [by right|].
      iNext. iIntros (t' V' Vt' 𝓥4) "(s' & H & Inv & Fx & ox' & R' & W' & M')".
      iDestruct "M'" as %(? & MAX2).
      iDestruct "H" as %(Ext5 & ? & Let & SL' & ? & ? & LeV & ? & ?).
      iSpecialize ("F" $! t' Let).
      iAssert (⌜Va ⊑ 𝓥3.(cur)⌝)%I as %LVa3.
      { rewrite naViewSeen_eq. by iDestruct "naV" as "[_ [$ _]]". }
      have LEV': 𝓥3.(cur) ⊑ V' by rewrite Ext5 -LeV /=.
      rewrite !(monPred_mono _ _ _  LEV').
      iAssert (GPS_SWWriter t' s' v' γ V')%I with "[ox' R' W']" as "SW'".
      { rewrite GPS_SWWriter_eq. iExists _. iFrame (MAX2) "ox' W'".
        iExists _,_,_. iFrame "naV R'". iPureIntro. split; [done|subst V'; solve_lat]. }
      iDestruct (GPS_SWWriter_Reader with "SW'") as "#SR'".
      iMod ("IW" $! (Vx ⊔ Va) with "Fx") as "[Q1 Q2]".
      iMod (own_lat_auth_update with "H𝓥3") as "[$ _]"; [done|].
      have ?: V2 ⊑ 𝓥4.(cur) by (etrans; [eassumption|by f_equiv]).
      have LEV4': V' ⊑ 𝓥4.(cur) by subst V'; apply lat_join_lub; [|rewrite LVa3 Ext5].
      iApply ("Post" $! _ V'). iFrame (Let LEV4') "SR' P". iModIntro.
      iIntros (V0 Ext0) "P".
      iMod ("F" with "SW' Q2 P") as "[IW >[F $]]".
      iMod ("IW" $! (Vx ⊔ Va) with "Q1") as "IW".
      iSpecialize ("Inv" with "F IW"). iIntros "!>" (V4 Ext4 ?).
      iNext. rewrite -(monPred_mono _ (Vo ⊔ 𝓥4.(cur)) V4); [done|]. simpl in LeV.
      rewrite LeV Ext4 /Vo. solve_lat.
    Qed.

    Lemma GPS_SWRaw_SWWrite_rel
      IW Q Q1 Q2 t s s' v v' γ tid Vb E (Exts: s ⊑ s'):
      ↑histN ⊆ E →
      let WVS: vProp :=
        (∀ t', ⌜(t < t')%positive⌝ -∗ GPS_SWWriter t' s' v' γ -∗ Q2
                  ={E}=∗ (<obj> (Q1 ={E}=∗ IW l γ t s v)) ∗
                         |={E}=> (F γ t' s' v' ∗ Q t'))%I in
      {{{ GPS_SWWriter t s v γ ∗ (monPred_in Vb → ▷ GPS_INV IW γ)
          ∗ ▷ WVS ∗ ▷ <obj> (F γ t s v ={E}=∗ Q1 ∗ Q2) }}}
        Write AcqRel #l v' in tid @ E
      {{{ t', RET #☠; ⌜(t < t')%positive⌝ ∗ GPS_SWReader t' s' v' γ ∗
            (monPred_in Vb → ▷ GPS_INV IW γ) ∗ Q t' }}}.
    Proof.
      iIntros (SUB WVS Φ). iIntros "(W & Inv & F & IW) Post".
      iApply wp_fupd.
      iApply (GPS_SWRaw_SWWrite_rel_view _ Q Q1 Q2 True%I
                with "[$W $Inv F $IW]"); [done..| |].
      { iSplit; [|done]. iIntros "!>" (t' Let) "W Q2 _".
        iApply ("F" with "[%//] W Q2"). }
      iIntros "!>" (t' V') "(Let & W & In & _ & VS)".
      iMod ("VS" with "[//]") as "[Inv Q]".
      iDestruct (monPred_in_elim with "In Inv") as "Inv".
      by iApply ("Post" with "[$Let $W $Inv $Q]").
    Qed.

    (* Single-writers can be shared temporarily.
       Writes are not exclusive in this phase *)
    Definition GPS_SWSharedWriter_def t s v γ : vProp :=
      (∃ ζ, PrtSeenView l t s v γ ∗ ⎡ Writer γ ζ ⎤ ∗
            ⌜∀ t', is_Some (ζ !! t') → t' ⊑ t⌝)%I.
    Definition GPS_SWSharedWriter_aux : seal (@GPS_SWSharedWriter_def).  by eexists. Qed.
    Definition GPS_SWSharedWriter := unseal (@GPS_SWSharedWriter_aux).
    Definition GPS_SWSharedWriter_eq : @GPS_SWSharedWriter = _ := seal_eq _.

    Global Instance GPS_SWSharedWriter_timeless : Timeless (GPS_SWSharedWriter t s v γ).
    Proof. rewrite GPS_SWSharedWriter_eq. by apply _. Qed.

    (* SharedReader holds a share of the exclusive write *)
    Definition GPS_SWSharedReader_def t s v q γ : vProp :=
      (∃ (tx: time), PrtSeen l t s v γ ∗ ⎡ exwr γ tx q ⎤ ∗ ⌜(tx ≤ t)%positive⌝)%I.
    Definition GPS_SWSharedReader_aux : seal (@GPS_SWSharedReader_def). by eexists. Qed.
    Definition GPS_SWSharedReader := unseal (@GPS_SWSharedReader_aux).
    Definition GPS_SWSharedReader_eq : @GPS_SWSharedReader = _ := seal_eq _.

    Global Instance GPS_SWSharedReader_timeless : Timeless (GPS_SWSharedReader t s v q γ).
    Proof. rewrite GPS_SWSharedReader_eq. by apply _. Qed.

    Global Instance GPS_SWSharedReader_fractional t s v γ :
      Fractional (λ q, GPS_SWSharedReader t s v q γ).
    Proof.
      rewrite /Fractional =>p q. rewrite GPS_SWSharedReader_eq. iSplit.
      - iDestruct 1 as (tx) "(#Pr & [Hp Hq] & #Le)".
        iSplitL "Hp"; iExists tx; by iFrame "Pr Le".
      - iDestruct 1 as "[Hp Hq]". iDestruct "Hp" as (tx) "(Pr & Hp & Le)".
        iDestruct "Hq" as (tx') "(_ & Hq & _)". iExists tx. iFrame "Pr Le".
        iDestruct (ExWrite_frac_agree with "Hp Hq") as %?. subst tx'.
        by iCombine "Hp Hq" as "Hpq".
    Qed.
    Global Instance GPS_SWSharedReader_asfractional t s v q γ :
      AsFractional (GPS_SWSharedReader t s v q γ) (λ q, GPS_SWSharedReader t s v q γ) q.
    Proof. split. done. apply _. Qed.

    Lemma GPS_SWWriter_shared t s v γ :
      GPS_SWWriter t s v γ -∗
      GPS_SWSharedWriter t s v γ ∗ GPS_SWSharedReader t s v 1%Qp γ.
    Proof.
      rewrite GPS_SWWriter_eq GPS_SWSharedWriter_eq GPS_SWSharedReader_eq.
      iDestruct 1 as (ζ) "(Prv & (W & ex) & MAX)".
      iDestruct (PrtSeenView_PrtSeen with "Prv") as "#Pr".
      iSplitR "ex".
      - iExists ζ. by iFrame.
      - iExists t. by iFrame "ex Pr".
    Qed.

    Lemma GPS_SWSharedReader_Reader t s v q γ:
      GPS_SWSharedReader t s v q γ -∗ GPS_SWReader t s v γ.
    Proof.
      rewrite GPS_SWReader_eq GPS_SWSharedReader_eq.
      iDestruct 1 as (tx) "[$ _]".
    Qed.

    Lemma GPS_SWSharedReader_Reader_join_subjectively t s v s' v' q γ:
      GPS_SWReader t s v γ -∗
      <subj> GPS_SWSharedReader t s' v' q γ -∗
      GPS_SWSharedReader t s v q γ.
    Proof.
      rewrite GPS_SWReader_eq GPS_SWSharedReader_eq.
      iIntros "Prt". rewrite monPred_subjectively_exist.
      iDestruct 1 as (tx) "R". rewrite 2!monPred_subjectively_sep.
      iDestruct "R" as "(_ & ex & Le)". rewrite monPred_subjectively_unfold /=.
      iDestruct "ex" as (?) "ex". iDestruct "Le" as (?) "%".
      rewrite monPred_at_embed. iExists tx. by iFrame "Prt ex".
    Qed.

    Lemma GPS_SWSharedReaders_join_subjectively t1 t2 s1 s2 v1 v2 q1 q2 γ:
      GPS_SWSharedReader t1 s1 v1 q1 γ -∗
      <subj> GPS_SWSharedReader t2 s2 v2 q2 γ -∗
      GPS_SWSharedReader t1 s1 v1 (q1 + q2) γ.
    Proof.
      rewrite GPS_SWSharedReader_eq.
      iDestruct 1 as (tx) "(Prt & ex1 & Le)". rewrite monPred_subjectively_exist.
      iDestruct 1 as (tx') "ex". rewrite 2!monPred_subjectively_sep.
      iDestruct "ex" as "(_&ex2&_)". rewrite monPred_subjectively_unfold /=.
      iDestruct "ex2" as (?) "ex2". rewrite monPred_at_embed.
      iExists tx. iFrame "Prt Le".
      iDestruct (ExWrite_frac_agree with "ex1 ex2") as %?. subst tx'.
      by iCombine "ex1" "ex2" as "$".
    Qed.

    Lemma GPS_SWSharedReaders_join t1 t2 s1 s2 v1 v2 q1 q2 γ:
      GPS_SWSharedReader t1 s1 v1 q1 γ -∗ GPS_SWSharedReader t2 s2 v2 q2 γ -∗
      GPS_SWSharedReader t1 s1 v1 (q1 + q2) γ.
    Proof.
      iIntros "R1 R2". iApply (GPS_SWSharedReaders_join_subjectively with "R1").
      by iModIntro.
    Qed.

    Lemma GPS_SWRaw_SharedReader_Reader_agree IW t1 t2 s1 s2 v1 v2 q2 γ Vb E:
      GPS_SWReader t1 s1 v1 γ -∗ GPS_SWSharedReader t2 s2 v2 q2 γ -∗
      (monPred_in Vb → ▷ GPS_INV IW γ) ={E}=∗
        GPS_SWSharedReader t2 s2 v2 q2 γ ∗ (monPred_in Vb → ▷ GPS_INV IW γ) ∗
       ⌜s1 ⊑ s2 ∧ t1 ⊑ t2 ∨ s2 ⊑ s1 ∧ t2 ⊑ t1⌝.
    Proof.
      iIntros "R1 R2".
      iDestruct (GPS_SWSharedReader_Reader with "R2") as "#RR2".
      iFrame "R2". by iApply (GPS_SWReaders_agree with "R1 RR2").
    Qed.

    Lemma GPS_SWRaw_SharedReaders_agree IW t1 t2 s1 s2 v1 v2 q1 q2 γ Vb E:
      GPS_SWSharedReader t1 s1 v1 q1 γ -∗ GPS_SWSharedReader t2 s2 v2 q2 γ -∗
      (monPred_in Vb → ▷ GPS_INV IW γ) ={E}=∗
        GPS_SWSharedReader t1 s1 v1 q1 γ ∗ GPS_SWSharedReader t2 s2 v2 q2 γ ∗
        (monPred_in Vb → ▷ GPS_INV IW γ) ∗ ⌜s1 ⊑ s2 ∧ t1 ⊑ t2 ∨ s2 ⊑ s1 ∧ t2 ⊑ t1⌝.
    Proof.
      iIntros "R1". iDestruct (GPS_SWSharedReader_Reader with "R1") as "#RR1".
      iFrame "R1". by iApply (GPS_SWRaw_SharedReader_Reader_agree with "RR1").
    Qed.

    Lemma GPS_SWWriter_SharedWriter_excluded t1 t2 s1 s2 v1 v2 γ :
      GPS_SWWriter t1 s1 v1 γ -∗ GPS_SWSharedWriter t2 s2 v2 γ -∗ False.
    Proof.
      rewrite GPS_SWWriter_eq GPS_SWSharedWriter_eq.
      iDestruct 1 as (?) "(_&(W1&_)&_)". iDestruct 1 as (?) "(_&W2&_)".
      by iDestruct (Writer_Exclusive with "W1 W2") as %?.
    Qed.

    Lemma GPS_SWWriter_SharedReader_excluded t1 t2 s1 s2 v1 v2 (q2: Qp) γ :
      GPS_SWWriter t1 s1 v1 γ -∗ GPS_SWSharedReader t2 s2 v2 q2 γ -∗ False.
    Proof.
      rewrite GPS_SWWriter_eq GPS_SWSharedReader_eq.
      iDestruct 1 as (?) "(_ & (_ & Ex1) & _)".
      iDestruct 1 as (?) "(_ & Ex2 & _)". iCombine "Ex1" "Ex2" as "Ex".
      iDestruct (own_valid with "Ex") as %[_ [[VAL _] _]].
      simpl in VAL. exfalso. rewrite -> frac_valid' in VAL.
      by apply Qp_not_plus_q_ge_1 in VAL.
    Qed.

    Lemma GPS_SWSharedWriter_latest_time_1 t t' s s' v v' γ:
      GPS_SWSharedWriter t s v γ -∗ GPS_SWReader t' s' v' γ -∗ ⌜t' ⊑ t⌝.
    Proof.
      rewrite GPS_SWSharedWriter_eq GPS_SWReader_eq. iStartProof iProp.
      iIntros (?). iDestruct 1 as (ζ) "(#Seen1 & W & MAX)".
      iDestruct "MAX" as %MAX. iIntros (? ->).
      iDestruct 1 as (V Va ?) "(#na & R & SL2)".
      iDestruct (Writer_latest with "[$W $R]") as %IN. iPureIntro.
      specialize (IN t'). rewrite lookup_singleton in IN.
      destruct (ζ !! t') as [svV|] eqn:Eqt'; last by inversion IN. simpl in IN.
      subst svV. by apply MAX; eexists.
    Qed.

    Lemma GPS_SWSharedWriter_latest_time t t' s s' v v' q γ:
      GPS_SWSharedWriter t s v γ -∗ GPS_SWSharedReader t' s' v' q γ -∗ ⌜t' ⊑ t⌝.
    Proof.
      iIntros "W R". iDestruct (GPS_SWSharedReader_Reader with "R") as "R".
      by iApply (GPS_SWSharedWriter_latest_time_1 with "W R").
    Qed.

    Lemma GPS_SWRaw_SharedWriter_latest_state IW t t' s s' v v' q γ Vb E:
      GPS_SWSharedWriter t s v γ -∗ GPS_SWSharedReader t' s' v' q γ -∗
      (monPred_in Vb → ▷ GPS_INV IW γ) ={E}=∗
        GPS_SWSharedWriter t s v γ ∗ GPS_SWSharedReader t' s' v' q γ ∗
        (monPred_in Vb → ▷ GPS_INV IW γ) ∗ ⌜s' ⊑ s ∧ t' ⊑ t⌝.
    Proof.
      rewrite GPS_SWSharedWriter_eq GPS_SWSharedReader_eq. iStartProof iProp.
      iIntros (?). iDestruct 1 as (ζ) "(Seen1 & W & MAX)".
      iIntros (? ->). iDestruct 1 as (tx') "(Seen2 & ex2 & Le2)".
      iDestruct "Seen2" as (???) "(na2 & R2 & SL2)".
      iIntros (V ->) "Inv".
      set Vo := V ⊔ Vb.
      iAssert ((▷ GPS_INV IW γ) Vo)%I with "[Inv]" as "Inv".
      { iApply "Inv". iPureIntro; solve_lat. }
      rewrite GPS_INV_eq monPred_at_later.
      iDestruct "Seen1" as (V1 Va1 ?) "(na1 & R1 & F1)".
      iMod (Writer_stsorted with "Inv W") as "(Inv & W & S)".
      iDestruct "S" as %SORTED. iDestruct "MAX" as %MAX.
      iDestruct "Inv" as (μ C tx Va ???) "(hist & at & aw & na & >oA & REST)".
      iDestruct (gps_ownA_Reader_singleton_sub with "oA R1") as %HR1.
      iDestruct (gps_ownA_Reader_singleton_sub with "oA R2") as %HR2.
      iDestruct (Writer_exact_2 with "oA W") as %?.
      iModIntro.
      iSplitL "na1 R1 F1 W"; last iSplitL "ex2 Le2 na2 R2 SL2"; last iSplitR "".
      - iExists ζ. iFrame (MAX) "W". iExists V1, Va1, _. by iFrame.
      - iExists tx'. iFrame "ex2 Le2". iExists _,_,_. by iFrame.
      - iIntros (V' ??). rewrite -(monPred_mono _ Vo V'); [|solve_lat].
        iExists μ, C, tx, Va, _, _,_. by iFrame.
      - iPureIntro. subst ζ. have ?: t' ⊑ t by apply MAX; eexists.
        split; [by apply (SORTED _ _ _ _ HR2 HR1)|done].
    Qed.

    Lemma GPS_SWSharedReader_Reader_update_max q t1 t2 s1 s2 v1 v2 γ:
      GPS_SWSharedReader t1 s1 v1 q γ -∗ GPS_SWReader t2 s2 v2 γ -∗
      let t := if (decide (t2 ≤ t1)%positive) then t1 else t2 in
      let s := if (decide (t2 ≤ t1)%positive) then s1 else s2 in
      let v := if (decide (t2 ≤ t1)%positive) then v1 else v2 in
      GPS_SWSharedReader t s v q γ.
    Proof.
      iIntros "SR R". case decide => [//|/Pos.lt_nle Le1].
      rewrite GPS_SWSharedReader_eq GPS_SWReader_eq.
      iDestruct "SR" as (tx) "(_ & ex & Le)".
      iExists tx. iFrame "R ex". iDestruct "Le" as %Le2.
      iPureIntro. by eapply Pos.lt_le_incl, Pos.le_lt_trans.
    Qed.

    Lemma GPS_SWSharedWriter_Reader_update t t' s s' v v' q γ :
      GPS_SWSharedWriter t s v γ -∗ GPS_SWSharedReader t' s' v' q γ -∗
      GPS_SWSharedWriter t s v γ ∗ GPS_SWSharedReader t s v q γ.
    Proof.
      rewrite GPS_SWSharedWriter_eq GPS_SWSharedReader_eq. iStartProof iProp.
      iIntros (?). iDestruct 1 as (ζ) "(#Seen1 & W & MAX)".
      iDestruct "MAX" as %MAX. iIntros (? ->).
      iDestruct 1 as (tx) "(Seen2 & ex & Le2)". iDestruct "Le2" as %Le2.
      iDestruct "Seen2" as (V Va ?) "(#na & R & _)".
      iDestruct (Writer_latest with "[$W $R]") as %IN.
      specialize (IN t'). rewrite lookup_singleton in IN.
      destruct (ζ !! t') as [svV|] eqn:Eqt'; last by inversion IN. simpl in IN.
      subst svV. have Let': t' ⊑ t by apply MAX; eexists.
      iSplitL "W".
      - iExists _. iFrame (MAX) "Seen1 W".
      - iExists tx. iDestruct (PrtSeenView_PrtSeen with "Seen1") as "$".
        iFrame " ex". iPureIntro. by etrans.
    Qed.

    Lemma SWConnect_cbends_le (C: cell) (tx t t': time) m m'
      (MAX: ∀ t', is_Some (C !! t') → t' ⊑ t) :
      SWConnect C tx → (tx ≤ t)%positive → (tx ≤ t')%positive → C !! t = Some m →
      cbends.cbends C !! t' = Some m' → (t ≤ t')%positive.
    Proof.
      move => SC Letx Letx' Eqt /map_filter_lookup_Some [Eqt' /= NADJ].
      case (decide (t ≤ t')%positive) => [//|/Pos.lt_nle NLe]. exfalso.
      apply (SC t); [by eexists|done|].
      exists (t' + 1)%positive. repeat split.
      - etrans; [apply Letx'|]. lia.
      - rewrite Pos.add_1_r. by apply Pos.le_succ_l.
      - destruct (C !! (t' + 1)%positive) as [m1|] eqn:Eq1; [|done].
        exfalso. apply (NADJ (t'+ 1)%positive m1 Eq1); [lia|done].
    Qed.

    Lemma GPS_SWRaw_SharedWriter_revert_subjectively IW t t' s s' v v' γ Vb E:
      GPS_SWSharedWriter t s v γ -∗ <subj> GPS_SWSharedReader t' s' v' 1%Qp γ
      -∗ (monPred_in Vb → ▷ GPS_INV IW γ)
      ={E}=∗ GPS_SWWriter t s v γ ∗ (monPred_in Vb → ▷ GPS_INV IW γ).
    Proof.
      rewrite GPS_SWSharedWriter_eq GPS_SWSharedReader_eq.
      iStartProof iProp. iIntros (?).
      iDestruct 1 as (ζ) "(#Seen1 & W & MAX)". iDestruct "MAX" as %MAX.
      iIntros (? ->). rewrite monPred_subjectively_exist.
      iDestruct 1 as (tx') "R". rewrite 2!monPred_subjectively_sep.
      iDestruct "R" as "(_ & ex2 & _)". iDestruct "ex2" as (?) "ex2".
      iIntros (V ->) "Inv".
      set Vo := V ⊔ Vb.
      iAssert ((▷ GPS_INV IW γ) Vo)%I with "[Inv]" as "Inv".
      { iApply "Inv". iPureIntro; solve_lat. }
      rewrite GPS_INV_eq monPred_at_later.
      iDestruct "Inv" as (μ C tx Va ???)
                          "(hist & at & aw & na & >oA & rB & rD & >FACT)".
      iDestruct "FACT" as % (FIN & SORTED & GH & JV & LeVo & NAL & IStx & ND & RC).
      iAssert (⌜∃ V, toState μ C !! t = Some (s, (v, V))⌝)%I with "[oA]" as %[Vt ISt].
      { iDestruct "Seen1" as (???) "(_&R&_)".
        iDestruct (gps_ownA_Reader_singleton_sub with "oA R") as %?.
        iPureIntro. by eexists. }
      iDestruct (gps_ownA_split with "oA") as "(oM & oE & oNa)".
      iDestruct (Writer_exact with "oM W") as %?. subst ζ.
      iDestruct (ExWrite_agree with "oE ex2") as %?. subst tx'.
      iMod (exwr_update _ _ t with "oE ex2") as "[oE ex2]". iModIntro.
      iSplitL "W ex2".
      { rewrite GPS_SWWriter_eq. iExists _. by iFrame (MAX) "Seen1 W ex2". }
      iIntros (V' ??). rewrite -(monPred_mono _ Vo V'); [|solve_lat].
      iExists μ, C, t, Va, _, _, _. rewrite gps_ownA_split.
      iFrame "hist at aw na oM oE oNa".
      have LEtx: (tx ≤ t)%positive by apply MAX.
      iSplitL "rB".
      { iNext. rewrite /resBE /gps_res. iApply (big_sepM_mono with "rB").
        move => t0 m0 Eq0 /=.
        have Let0: (tx ≤ t0)%positive → (t ≤ t0)%positive.
        { move => ?. apply toState_lookup_message in ISt as [? [??]].
          eapply (SWConnect_cbends_le C tx t t0 _ m0); eauto.
          move => t1 [m1 Eq1]. specialize (ND _ _ Eq1).
          destruct m1.(mval) eqn:Eqv1; [|done|]; apply MAX; eexists;
            (apply toState_lookup_state_inv; [|done]);
            rewrite Eqv1; constructor. }
        have Let0': ¬ (tx ≤ t0)%positive → ¬ (t ≤ t0)%positive
          by move => NL ?; apply NL; etrans.
        destruct m0.(mval) ; [|done|];
          (case (decide (tx ≤ t0)%positive) => [/Let0 Le|/Let0' NLe];
            by [rewrite (decide_True _ _ Le)|rewrite (decide_False _ _ NLe)]). }
      iSplitL "rD".
      { iNext. rewrite /resD /gps_res. iApply (big_sepM_mono with "rD").
        move => t0 m0 Eq0 /=.
        have Let0': ¬ (tx ≤ t0)%positive → ¬ (t ≤ t0)%positive
          by move => NL ?; apply NL; etrans.
        destruct m0.(mval) eqn:Eqv0; [|done|];
          (case (decide (tx ≤ t0)%positive) => [Le|/Let0' NLe];
            [case (decide (t ≤ t0)%positive) => ?; by iDestruct 1 as "[$|$]"|
            rewrite (decide_False _ _ NLe); by iDestruct 1 as "[$|[$|$]]"]). }
      iPureIntro. split; [by apply (FinalInv_mono _ _ _ _ LEtx)|].
      do 5 (split; [done|]). repeat split; [by eexists|done|].
      move => t0 [m0 Eqt0] Let0.
      have ?: t0 = t.
      { apply : anti_symm; [|done]. apply MAX. specialize (ND _ _ Eqt0).
        destruct m0.(mval) eqn:Eqv0; [|done|]; eexists;
          (apply toState_lookup_state_inv; [|done]); rewrite Eqv0; constructor. }
      subst t0. move => [td [? [? NIN]]].
      have ?: td = t by apply : anti_symm. subst td. by rewrite Eqt0 in NIN.
    Qed.

    Lemma GPS_SWRaw_SharedWriter_revert IW t t' s s' v v' γ Vb E:
      GPS_SWSharedWriter t s v γ -∗ GPS_SWSharedReader t' s' v' 1%Qp γ
      -∗ (monPred_in Vb → ▷ GPS_INV IW γ)
      ={E}=∗ GPS_SWWriter t s v γ ∗ (monPred_in Vb → ▷ GPS_INV IW γ).
    Proof.
      iIntros "W R". iApply (GPS_SWRaw_SharedWriter_revert_subjectively with "W").
      by iModIntro.
    Qed.

    Lemma GPS_SWRaw_SharedReader_Read IW R o t s v (q: frac) γ tid Vb E:
      ↑histN ⊆ E → o = Relaxed ∨ o = AcqRel →
      {{{ GPS_SWSharedReader t s v q γ ∗ (monPred_in Vb → ▷ GPS_INV IW γ)
          ∗ (▷ ∀ t' s' v', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
               <obj> ((F_read γ t' s' v' ={E}=∗ F_read γ t' s' v' ∗ R t' s' v') ∧
                      (F γ t' s' v' ={E}=∗ F γ t' s' v' ∗ R t' s' v'))) }}}
        Read o #l in tid @ E
      {{{ t' s' v', RET v'; ⌜s ⊑ s' ∧ t ⊑ t'⌝
          ∗ GPS_SWSharedReader t' s' v' q γ ∗ (monPred_in Vb → ▷ GPS_INV IW γ)
          ∗ if (decide (AcqRel ⊑ o)) then R t' s' v' else ▽{tid} R t' s' v' }}}.
    Proof.
      iIntros (SUB OR Φ). iStartProof iProp. iIntros (V1) "(R & Inv & D)".
      set Vo := V1 ⊔ Vb.
      iAssert ((▷ GPS_INV IW γ) Vo)%I with "[Inv]" as "Inv".
      { iApply "Inv". iPureIntro; solve_lat. }
      rewrite GPS_INV_eq monPred_at_later wp_eq.
      iIntros (V2 ?) "Post". iIntros (𝓥3 Ext3) "H𝓥3 s".
      have Ext3' : V1 ⊑ 𝓥3.(cur) by etrans.
      rewrite {1}(monPred_mono _ _ _ Ext3') GPS_SWSharedReader_eq.
      iDestruct "R" as (tx) "(R & ex & Le)". iDestruct "Le" as %Le.
      iDestruct "R" as (V Va ?) "(#naV & R & SL)". iDestruct "SL" as %SL.

      iApply program_logic.weakestpre.wp_fupd.
      iApply (GPSRaw_Read_ex IP l IW R tx t q s (v,V) o γ _ _ Vo Va _ _ E SL
                Le SUB OR with "[$s D $Inv $ex $R $naV]").
      {  iNext. iIntros (t' s' v' Ext'). iSpecialize ("D" $! t' s' v' Ext').
        rewrite monPred_at_objectively. by iApply "D". }

      iNext. iIntros (t' s' V' 𝓥4 v') "(s' & Ext4 & Inv & ex & Rs & R)".
      iDestruct "R" as (Vt') "[#R %]".
      iDestruct "Ext4" as %(ExtV & Exts & Extt & SL' & LeV').
      iMod (own_lat_auth_update with "H𝓥3") as "[$ #H𝓥4]"; [done|]. iModIntro.
      assert (V2 ⊑ 𝓥4.(cur)) by (etrans; [eassumption|by f_equiv]).
      iApply ("Post" $! t' s'). iSplitL""; [done|]. iSplitL "ex"; [|iSplitL "Inv"].
      - iExists tx. iFrame "ex". iSplit; [|iPureIntro; by etrans].
        iExists Vt', Va, _. rewrite (monPred_mono _ _ _ (cur_mono _ _ ExtV)).
        iFrame (SL') "R naV".
      - iIntros (V0 Ext0 ?). iNext.
        rewrite -(monPred_mono _ (Vo ⊔ 𝓥4.(cur)) V0); [done|].
        apply lat_join_lub; [|done].
        apply lat_join_lub; [by rewrite -Ext0 -ExtV|done].
      - rewrite acq_mod_eq. destruct OR; subst o; simpl; [iExists _|]; by iFrame.
    Qed.

    Lemma GPS_SWRaw_SharedReader_Read_val_dealloc
      IW P R R' vd t s v (q: frac) γ tid Vb E:
      ↑histN ⊆ E →
      {{{ GPS_SWSharedReader t s v q γ ∗ (monPred_in Vb → ▷ GPS_INV IW γ)
          ∗ P
          ∗ (▷ ∀ t' s', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
               (P -∗ GPS_SWSharedReader t s v q γ -∗
                (F γ t' s' vd ={E}=∗
                  monPred_in Vb ∗ GPS_SWSharedWriter t' s' vd γ ∗ R t' s') ∧
                (F_read γ t' s' vd ={E}=∗ False)) ∧
               <obj> (∀ v',
                      (F_read γ t' s' v' ={E}=∗ F_read γ t' s' v' ∗ R' t' s' v') ∧
                      (F γ t' s' v' ={E}=∗ F γ t' s' v' ∗ R' t' s' v'))) }}}
        Read AcqRel #l in tid @ E
      {{{ t' s' v', RET v'; ⌜s ⊑ s' ∧ t ⊑ t'⌝ ∗
          if (decide (v' = vd)) then l ↦ v' ∗ R t' s'
          else P ∗ R' t' s' v' ∗
            GPS_SWSharedReader t' s' v' q γ ∗ (monPred_in Vb → ▷ GPS_INV IW γ) }}}.
    Proof.
      iIntros (SUB Φ). iStartProof iProp. iIntros (V1) "(R & Inv & P & VS)".
      set Vo := V1 ⊔ Vb.
      iAssert ((▷ GPS_INV IW γ) Vo)%I with "[Inv]" as "Inv".
      { iApply "Inv". iPureIntro; solve_lat. }
      rewrite GPS_INV_eq monPred_at_later.
      iIntros (V2 Ext1) "Post". iApply wp_hist_inv; [done|].
      iIntros (V3 Ext2) "#HInv". rewrite wp_eq. iIntros (𝓥3 Ext3) "H𝓥3 s".
      have Ext3' : V1 ⊑ 𝓥3.(cur) by solve_lat.
      rewrite {1}(monPred_mono _ _ _ Ext3') GPS_SWSharedReader_eq.
      iDestruct "R" as (tx) "(R & ex & Le)". iDestruct "Le" as %Le.
      iDestruct "R" as (V Va ?) "(#naV & #R & SL)". iDestruct "SL" as %SL.

      iApply program_logic.weakestpre.wp_fupd.
      iApply (GPSRaw_Read_val_dealloc_ex IP l IW P R R' vd q tx t s (v,V) γ
              γ _ Vo Va _ _ E SL Le SUB
              with "[$HInv $s $P VS $Inv $ex $R $naV]").
      { iIntros "!>" (t' s' Ext'). iSpecialize ("VS" $! t' s' Ext'). iSplit.
        - rewrite bi.and_elim_l. rewrite (monPred_mono _ _ _ Ext3').
          iIntros (? Ext4) "P".
          iIntros (? Ext5) "ex". iSpecialize ("VS" with "P [ex]").
          { iDestruct "ex" as (tx') "ex". iExists tx'. iFrame "ex". iExists _,_,_.
            iFrame "naV R". iPureIntro. by eapply seen_local_mono. }
          iSplit; [rewrite bi.and_elim_l|by rewrite bi.and_elim_r]. rewrite Ext5.
          iIntros (? ?) "F". iMod ("VS" with "F") as "(% & W & $)".
          iModIntro. iSplit. { iPureIntro. rewrite /Vo. solve_lat. }
          rewrite GPS_SWSharedWriter_eq. iDestruct "W" as (?) "(Prt & W & MAX)".
          iExists _. iFrame "Prt W MAX".
        - rewrite bi.and_elim_r. iIntros (v'). by iApply ("VS" $! ∅ v'). }

      iIntros "!>" (t' s' V' 𝓥' v') "(s' & FACTs & R')".
      iDestruct "FACTs" as %(ExtV & Exts & Extt & SL' & LeV').
      iMod (own_lat_auth_update with "H𝓥3") as "[$ #H𝓥4]"; [done|]. iModIntro.
      assert (V2 ⊑ 𝓥'.(cur)). { by rewrite Ext2 Ext3 ExtV. }
      iApply ("Post" $! t' s'). iSplitL""; [done|].
      case (decide (v' = vd)) => ?.
      - iDestruct "R'" as "[$ $]".
      - iDestruct "R'" as "($ & $ & Inv & ex & Vt)". iSplitL "ex Vt".
        + iExists tx. iFrame "ex". iSplit; [|iPureIntro; by etrans].
          iDestruct "Vt" as (Vt') "[#R' %]".
          iExists Vt', Va, _. rewrite (monPred_mono _ _ _ (cur_mono _ _ ExtV)).
          iFrame (SL') "R' naV".
        + iIntros (V0 Ext0 ?). iNext.
          rewrite -(monPred_mono _ (Vo ⊔ 𝓥'.(cur)) V0); [done|].
          apply lat_join_lub; [|done].
          apply lat_join_lub; [by rewrite -Ext0 -ExtV|done].
    Qed.

    Lemma GPS_SWRaw_Reader_Read_val_dealloc IW P R R' vd t s v γ tid Vb E:
      ↑histN ⊆ E →
      {{{ GPS_SWReader t s v γ ∗ (monPred_in Vb → ▷ GPS_INV IW γ) ∗ P
          ∗ (▷ ∀ t' s', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
                  (P -∗ (F γ t' s' vd ={E}=∗
                    monPred_in Vb ∗ GPS_SWSharedWriter t' s' vd γ ∗ R t' s') ∧
                  (F_read γ t' s' vd ={E}=∗ False) ∧
                  (IW l γ t' s' vd ={E}=∗ False)) ∧
                  <obj> (∀ v',
                    (F_read γ t' s' v' ={E}=∗ F_read γ t' s' v' ∗ R' t' s' v') ∧
                    (F γ t' s' v' ={E}=∗ F γ t' s' v' ∗ R' t' s' v') ∧
                    (IW l γ t' s' v' ={E}=∗ IW l γ t' s' v' ∗ R' t' s' v'))) }}}
        Read AcqRel #l in tid @ E
      {{{ t' s' v', RET v'; ⌜s ⊑ s' ∧ t ⊑ t'⌝ ∗
          if (decide (v' = vd)) then
            l ↦ v' ∗ R t' s'
          else
            P ∗ R' t' s' v' ∗
            GPS_SWReader t' s' v' γ ∗ (monPred_in Vb → ▷ GPS_INV IW γ) }}}.
    Proof.
      iIntros (SUB Φ). iStartProof iProp. iIntros (V1) "(R & Inv & P & VS)".
      set Vo := V1 ⊔ Vb.
      iAssert ((▷ GPS_INV IW γ) Vo)%I with "[Inv]" as "Inv".
      { iApply "Inv". iPureIntro; solve_lat. }
      rewrite GPS_INV_eq monPred_at_later.
      iIntros (V2 Ext1) "Post". iApply wp_hist_inv; [done|].
      iIntros (V3 Ext2) "#HInv". rewrite wp_eq. iIntros (𝓥3 Ext3) "H𝓥3 s".
      have Ext3' : V1 ⊑ 𝓥3.(cur) by solve_lat.
      rewrite {1}(monPred_mono _ _ _ Ext3') GPS_SWReader_eq.
      iDestruct "R" as (V Va ?) "(#naV & #R & SL)". iDestruct "SL" as %SL.

      iApply program_logic.weakestpre.wp_fupd.
      iApply (GPSRaw_Read_val_dealloc_non_ex IP l IW P R R' vd t s (v,V) γ γ _
                Vo Va _ _ E SL SUB with "[$HInv $s $P VS $Inv $R $naV]").
      { iIntros "!>" (t' s' Ext'). iSpecialize ("VS" $! t' s' Ext'). iSplit.
        - rewrite bi.and_elim_l. rewrite (monPred_mono _ _ _ Ext3').
          iIntros (? Ext4) "P". iSpecialize ("VS" with "P").
          iSplit; [rewrite bi.and_elim_l|by rewrite bi.and_elim_r].
          iIntros (? ?) "F". iMod ("VS" with "F") as "(% & W & $)".
          iModIntro. iSplit. { iPureIntro. rewrite /Vo. solve_lat. }
          rewrite GPS_SWSharedWriter_eq. iDestruct "W" as (?) "(Prt & W & MAX)".
          iExists _. by iFrame "Prt W MAX".
        - rewrite bi.and_elim_r. iIntros (v'). by iApply ("VS" $! ∅ v'). }

      iIntros "!>" (t' s' V' 𝓥' v') "(s' & FACTs & R')".
      iDestruct "FACTs" as %(ExtV & Exts & Extt & SL' & LeV').
      iMod (own_lat_auth_update with "H𝓥3") as "[$ #H𝓥4]"; [done|]. iModIntro.
      assert (V2 ⊑ 𝓥'.(cur)). { by rewrite Ext2 Ext3 ExtV. }
      iApply ("Post" $! t' s'). iSplitL""; [done|].
      case (decide (v' = vd)) => ?.
      - iDestruct "R'" as "[$ $]".
      - iDestruct "R'" as "($ & $ & Inv & Vt)". iSplitL "Vt".
        + iDestruct "Vt" as (Vt') "[#R' %]".
          iExists Vt', Va, _. rewrite (monPred_mono _ _ _ (cur_mono _ _ ExtV)).
          iFrame (SL') "R' naV".
        + iIntros (V0 Ext0 ?). iNext.
          rewrite -(monPred_mono _ (Vo ⊔ 𝓥'.(cur)) V0); [done|].
          apply lat_join_lub; [|done]. apply lat_join_lub; [by rewrite -Ext0 -ExtV|done].
    Qed.

    Lemma GPS_SWRaw_SharedWriter_CAS_int_view_strong
      IW (dropR: bool) orf or (vr: Z) (vw: val) t s v q γ P P' Q R tid Vb E
      E' (SUBE: E' ⊆ E):
      ↑histN ⊆ E → orf = Relaxed ∨ orf = AcqRel → or = Relaxed ∨ or = AcqRel →
      let VSC : vProp :=
        (<obj> ∀ t' s' v', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
                      (F γ t' s' v' ∨ F_read γ t' s' v') -∗
                      ⌜∃ vl, v' = #vl ∧ lit_comparable vr vl⌝)%I in
      let VS : vProp :=
        (▷ ∀ t' s', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
            (P -∗ F γ t' s' #vr ={E}=∗ GPS_SWSharedWriter t' s' #vr γ ∗
                ∃ s'', ⌜s' ⊑ s''⌝ ∗
                  (∀ t'' , ⌜(t' < t'')%positive⌝ -∗ GPS_SWSharedWriter t'' s'' vw γ -∗
                    (if dropR then GPS_SWSharedReader t'' s'' vw q γ else True) ={E}=∗
                    (<obj> (|={E}=> F_read γ t' s' #vr)) ∗
                    |={E}=> Q t'' s'' ∗ F γ t'' s'' vw) ))%I in
      let VSF: vProp :=
        (∀ t' s' (v: lit), ⌜s ⊑ s' ∧ t ⊑ t' ∧ lit_neq vr v⌝ -∗
          ▷ <obj> ((F_read γ t' s' #v ={E}=∗ F_read γ t' s' #v ∗ R t' s' v) ∧
                 (F γ t' s' #v ={E}=∗ F γ t' s' #v ∗ R t' s' v)))%I in
      {{{ GPS_SWSharedReader t s v q γ
          ∗ (monPred_in Vb → ▷ GPS_INV IW γ)
          ∗ ▷ VSC ∗ VS ∗ VSF ∗ P ∗ P'}}}
        CAS #l #vr vw orf or AcqRel in tid @ E
      {{{ (b: bool) t' s' (v': lit), RET #b;
          ⌜s ⊑ s'⌝ ∗
          ( (⌜b = true  ∧ v' = LitInt vr ∧ (t < t')%positive⌝
              ∗ ∃ V', monPred_in V' ∗ ⎡ P V' ⎤ ∗ ⎡ P' V' ⎤ ∗
                (⎡ P V' ⎤ ={E}=∗ ⎡ (monPred_in Vb → ▷ GPS_INV IW γ) V' ⎤
                  ∗ ∃ s'', ⌜s' ⊑ s''⌝
                  ∗ (if dropR then True else GPS_SWSharedReader t' s'' vw q γ)
                  ∗ if decide (AcqRel ⊑ or) then Q t' s'' else ▽{tid} Q t' s''))
          ∨ (⌜b = false ∧ lit_neq vr v' ∧ (t ≤ t')%positive⌝
              ∗ (monPred_in Vb → ▷ GPS_INV IW γ)
              ∗ GPS_SWSharedReader t' s' #v' q γ
              ∗ (if decide (AcqRel ⊑ orf) then R t' s' v' else ▽{tid} (R t' s' v'))
              ∗ P ∗ P') ) }}}.
    Proof.
      iIntros (SUB ORF OR VSC VS VSF Φ). iStartProof iProp.
      iIntros (?) "(R & Inv & Comp & VS & VSF & P & P')". iIntros (V1 Ext1) "Post".
      rewrite -> Ext1.

      set Vo := V1 ⊔ Vb.
      iAssert ((▷ GPS_INV IW γ) Vo)%I with "[Inv]" as "Inv".
      { iApply "Inv". iPureIntro; solve_lat. }
      rewrite GPS_INV_eq monPred_at_later wp_eq.
      iIntros (𝓥3 Ext3) "H𝓥3 s". rewrite {1}Ext3 GPS_SWSharedReader_eq.
      iDestruct "R" as (tx) "(Prt & ex & Le)". iDestruct "Le" as %Let.
      iDestruct "Prt" as (Vr Va ?) "(#naV & R & SL)". iDestruct "SL" as %SL.
      (* end bookkeeping proof *)

      iApply program_logic.weakestpre.wp_fupd.
      iApply (GPSRaw_CAS_int_write_view IP l IW true t tx s (v, Vr) orf or AcqRel
                vr vw γ γ q Vo Va _ _ E E' True%I R SL SUBE SUB ORF OR
              with "[$s $Inv $R $naV $ex Comp VSF]"); [by right|..].
      { iFrame (Let). iSplitL "Comp"; [rewrite monPred_at_objectively; by iNext|].
        iSplitR ""; [|done].
        iIntros (t' s' ? -> [Ext' Ext'']). iSplit; [done|].
        iIntros (v' ? -> Eqv'). iSpecialize ("VSF" $! t' s' v' with "[%//]").
        iIntros "!>" (V). iSpecialize ("VSF" $! V). iSplit; last iSplit; last done;
        by [rewrite bi.and_elim_l|rewrite bi.and_elim_r]. }

      iNext. iIntros (b v' s' t' V' Vt' 𝓥4) "(s' & H & ex & _ & CASE)".
      iDestruct "H" as %(Ext4 & Exts & SL' & ?).
      iMod (own_lat_auth_update with "H𝓥3") as "[$ #H𝓥4]"; [done|]. iModIntro.
      assert (V1 ⊑ 𝓥4.(cur)) by (etrans; [eassumption|by f_equiv]).

      iApply ("Post" $! _ t' s' v').

      iDestruct "CASE" as "[CASE|(H & Inv & #R & QR)]".
      - iDestruct "CASE" as (tr Vtr) "(FACT & VSW & F & #R)".
        iDestruct "FACT" as %(?&?&?&Le1&Le2&Le3&Le4&Le5&Le6). simpl in Le5.
        subst b v'. iSplit; [done|].
        iLeft. iSplit. { iPureIntro. repeat split. by eapply Pos.le_lt_trans. }
        iExists 𝓥4.(cur). iSplit; [done|]. iFrame "P P'".
        iIntros (? Ext4') "P".
        iSpecialize ("VS" $! tr s' with "[%//] P").
        rewrite (monPred_mono _ _ _ (lat_join_sqsubseteq_l 𝓥4.(cur) (Vtr ⊔ Va))).
        have Le7: 𝓥4.(cur) ⊔ (Vtr ⊔ Va) ⊑ V' by (subst V'; solve_lat).
        iMod ("VS" with "F") as "[W VS]".
        iDestruct "VS" as (s'' Ext'') "VS".
        iSpecialize ("VSW" with "[W]").
        { rewrite GPS_SWSharedWriter_eq. iDestruct "W" as (ζ) "(Prt & W & MAX)".
          iExists ζ. iFrame "Prt W MAX". }
        iMod ("VSW" $! s'' Ext'') as "(W & #R' & Inv)". rewrite GPS_SWSharedWriter_eq.
        iSpecialize  ("VS" $! t' Le1 with "W").

        iAssert (GPS_SWSharedReader_def t' s'' vw q γ 𝓥4.(cur))%I
          with "[ex]" as "OR".
        { iExists tx. iDestruct "ex" as "[$ %]". iSplit; last first.
          - iPureIntro. etrans; [eauto|]. by eapply Pos.lt_le_incl, Pos.le_lt_trans.
          - iExists _, _, _. rewrite (monPred_mono _ _ _ (cur_mono _ _ Ext4)).
            iFrame (SL') "naV R'". }

        iAssert (((if dropR then True else GPS_SWSharedReader_def t' s'' vw q γ)
                ∗ (if dropR then GPS_SWSharedReader_def t' s'' vw q γ else True))
                𝓥4.(cur))%I with "[OR]" as "[OR1 OR2]".
        { destruct dropR; by iFrame. }

        iMod ("VS" with "OR2") as "[Fr >[Q F]]".
        rewrite monPred_at_objectively. iMod ("Fr" $! (Vtr ⊔ Va)) as "Fr".
        iMod ("Inv" with "F Fr") as "Inv". iModIntro.
        iSplitL "Inv".
        { iIntros (V5 LeV5 Leb). iNext.
          rewrite -(monPred_mono _ (Vo ⊔ 𝓥4.(cur)) V5); [done|].
          apply lat_join_lub; [|done].
          apply lat_join_lub; [by rewrite -LeV5|done]. }

        iExists s''. iFrame (Ext'') "OR1". case decide => ROR.
        + rewrite (decide_True _ _ ROR) in Le6. iFrame "Q".
        + rewrite (decide_False _ _ ROR) in Le6.
          rewrite acq_mod_eq. iExists _. iFrame "H𝓥4 Q".

      - iClear "VS".
        iFrame (Exts). iRight. iDestruct "H" as %(?&?&Le1&Le2).
        iSplit; [done|]. iFrame "P P'".
        iSplitL "Inv".
        { iIntros (V5 LeV5 Leb). iNext.
          rewrite -(monPred_mono _ (Vo ⊔ 𝓥4.(cur)) V5); [done|].
          apply lat_join_lub; [|done].
          apply lat_join_lub; [by rewrite -LeV5|done]. }
        iSplitL "ex".
        { iExists _. iDestruct "ex" as "[$ %]".
          iSplit; last by (iPureIntro; etrans).
          iExists _, _, _. rewrite (monPred_mono _ _ _ (cur_mono _ _ Ext4)).
          iFrame (SL') "naV R". }
          rewrite acq_mod_eq.
          destruct ORF; subst orf; simpl; [iExists _; by iFrame|done].
    Qed.

    Lemma GPS_SWRaw_SharedWriter_CAS_int_view_weak
      IW orf or (vr: Z) (vw: val) t s v γ P P' Q R tid Vb E
      E' (SUBE: E' ⊆ E):
      ↑histN ⊆ E → orf = Relaxed ∨ orf = AcqRel → or = Relaxed ∨ or = AcqRel →
      let VSC : vProp :=
        (<obj> ∀ t' s' v', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
                      (F γ t' s' v' ∨ F_read γ t' s' v' ∨ IW l γ t' s' v') -∗
                      ⌜∃ vl, v' = #vl ∧ lit_comparable vr vl⌝)%I in
      let VS : vProp :=
        (▷ ∀ t' s', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
            (P -∗ F γ t' s' #vr ={E}=∗ GPS_SWSharedWriter t' s' #vr γ ∗
                ∃ s'', ⌜s' ⊑ s''⌝ ∗
                  (∀ t'' , ⌜(t' < t'')%positive⌝ -∗ GPS_SWSharedWriter t'' s'' vw γ ={E}=∗
                    (<obj> (|={E}=> F_read γ t' s' #vr)) ∗
                    |={E}=> Q t'' s'' ∗ F γ t'' s'' vw) ))%I in
      let VSF: vProp :=
        (∀ t' s', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
            (▷ (P -∗ IW l γ t' s' #vr ={E}=∗ False)) ∧
          ∀ (v: lit), ⌜lit_neq vr v⌝ -∗
            ▷ <obj> ((F_read γ t' s' #v ={E}=∗ F_read γ t' s' #v ∗ R t' s' v) ∧
                 (F γ t' s' #v ={E}=∗ F γ t' s' #v ∗ R t' s' v) ∧
                 (IW l γ t' s' #v ={E}=∗ IW l γ t' s' #v ∗ R t' s' v)))%I in
      {{{ GPS_SWReader t s v γ
          ∗ (monPred_in Vb → ▷ GPS_INV IW γ)
          ∗ ▷ VSC ∗ VS ∗ VSF ∗ P ∗ P' }}}
        CAS #l #vr vw orf or AcqRel in tid @ E
      {{{ (b: bool) t' s' (v': lit), RET #b;
          ⌜s ⊑ s'⌝ ∗
          ( (⌜b = true  ∧ v' = LitInt vr ∧ (t < t')%positive⌝
              ∗ ∃ V', monPred_in V' ∗ ⎡ P V' ⎤ ∗ ⎡ P' V' ⎤ ∗
                (⎡ P V' ⎤ ={E}=∗ ⎡ (monPred_in Vb → ▷ GPS_INV IW γ) V' ⎤
                  ∗ ∃ s'', ⌜s' ⊑ s''⌝
                  ∗ GPS_SWReader t' s'' vw γ
                  ∗ if decide (AcqRel ⊑ or) then Q t' s'' else ▽{tid} Q t' s''))
          ∨ (⌜b = false ∧ lit_neq vr v' ∧ (t ≤ t')%positive⌝
              ∗ (monPred_in Vb → ▷ GPS_INV IW γ)
              ∗ GPS_SWReader t' s' #v' γ
              ∗ (if decide (AcqRel ⊑ orf) then R t' s' v' else ▽{tid} (R t' s' v'))
              ∗ P ∗ P') ) }}}.
    Proof.
      iIntros (SUB ORF OR VSC VS VSF Φ). iStartProof iProp.
      iIntros (?) "(R & Inv & Comp & VS & VSF & P & P')". iIntros (V1 Ext1) "Post".
      rewrite !(monPred_mono _ _ _ Ext1).

      set Vo := V1 ⊔ Vb.
      iAssert ((▷ GPS_INV IW γ) Vo)%I with "[Inv]" as "Inv".
      { iApply "Inv". iPureIntro; solve_lat. }
      rewrite GPS_INV_eq monPred_at_later wp_eq.
      iIntros (𝓥3 Ext3) "H𝓥3 s". rewrite {1}Ext3 GPS_SWReader_eq.
      iDestruct "R" as (Vr Va ?) "(#naV & R & SL)". iDestruct "SL" as %SL.
      (* end bookkeeping proof *)

      iApply program_logic.weakestpre.wp_fupd.
      iApply (GPSRaw_CAS_int_write_view IP l IW false t 1 s (v, Vr) orf or AcqRel
                vr vw γ γ 1%Qp Vo Va _ _ E E' P R SL SUBE SUB ORF OR
              with "[$s $Inv $R $naV Comp VSF $P]"); [by right|..].
      { iSplitL "Comp"; [rewrite monPred_at_objectively; by iNext|].
        simpl. iIntros (t' s'). iSpecialize ("VSF"$! t' s'). iFrame "VSF". }

      iNext. iIntros (b v' s' t' V' Vt' 𝓥4) "(s' & H & _ & P & CASE)".
      iDestruct "H" as %(Ext4 & Exts & SL' & ?).
      iMod (own_lat_auth_update with "H𝓥3") as "[$ #H𝓥4]"; [done|]. iModIntro.
      assert (V1 ⊑ 𝓥4.(cur)) by (etrans; [eassumption|by f_equiv]).

      iApply ("Post" $! _ t' s' v').

      iDestruct "CASE" as "[CASE|(H & Inv & #R & QR)]".
      - iDestruct "CASE" as (tr Vtr) "(FACT & VSW & F & #R)".
        iDestruct "FACT" as %(?&?&?&Le1&Le2&Le3&Le4&Le5&Le6). simpl in Le5.
        subst b v'. iSplit; [done|].
        iLeft. iSplit. { iPureIntro. repeat split. by eapply Pos.le_lt_trans. }
        iExists 𝓥4.(cur). iSplit; [done|]. iFrame "P P'".
        iIntros (? Ext4') "P".
        iSpecialize ("VS" $! tr s' with "[%//] P").
        rewrite (monPred_mono _ _ _ (lat_join_sqsubseteq_l 𝓥4.(cur) (Vtr ⊔ Va))).
        have Le7: 𝓥4.(cur) ⊔ (Vtr ⊔ Va) ⊑ V' by (subst V'; solve_lat).
        iMod ("VS" with "F") as "[W VS]".
        iDestruct "VS" as (s'' Ext'') "VS".
        iSpecialize ("VSW" with "[W]").
        { rewrite GPS_SWSharedWriter_eq. iDestruct "W" as (ζ) "(Prt & W & MAX)".
          iExists ζ. iFrame "Prt W MAX". }
        iMod ("VSW" $! s'' Ext'') as "(W & #R' & Inv)". rewrite GPS_SWSharedWriter_eq.
        iMod  ("VS" $! t' Le1 with "W") as "[Fr >[Q F]]".

        rewrite monPred_at_objectively. iMod ("Fr" $! (Vtr ⊔ Va)) as "Fr".
        iMod ("Inv" with "F Fr") as "Inv". iModIntro.
        iSplitL "Inv".
        { iIntros (V5 LeV5 Leb). iNext.
          rewrite -(monPred_mono _ (Vo ⊔ 𝓥4.(cur)) V5); [done|].
          apply lat_join_lub; [|done].
          apply lat_join_lub; [by rewrite -LeV5|done]. }

        iAssert (GPS_SWReader_def t' s'' vw γ 𝓥4.(cur))%I as "#RR'".
        { iExists _, _, _. rewrite (monPred_mono _ _ _ (cur_mono _ _ Ext4)).
          iFrame (SL') "naV R'". }

        iExists s''. iFrame (Ext'') "RR'". case decide => ROR.
        + rewrite (decide_True _ _ ROR) in Le6. iFrame "Q".
        + rewrite (decide_False _ _ ROR) in Le6.
          rewrite acq_mod_eq. iExists _. iFrame "H𝓥4 Q".

      - iClear "VS".
        iFrame (Exts). iRight. iDestruct "H" as %(?&?&Le1&Le2).
        iSplit; [done|]. iFrame "P P'".
        iSplitL "Inv".
        { iIntros (V5 LeV5 Leb). iNext.
          rewrite -(monPred_mono _ (Vo ⊔ 𝓥4.(cur)) V5); [done|].
          apply lat_join_lub; [|done].
          apply lat_join_lub; [by rewrite -LeV5|done]. }
        iSplitL "".
        + iExists _, _, _. rewrite (monPred_mono _ _ _ (cur_mono _ _ Ext4)).
          iFrame (SL') "naV R".
        + rewrite acq_mod_eq.
          destruct ORF; subst orf; simpl; [iExists _; by iFrame|done].
    Qed.

    Lemma GPS_SWRaw_SharedWriter_CAS_int_strong
      IW (dropR: bool) orf or ow (vr: Z) (vw: val) t s v q γ P Q Q1 Q2 R tid Vb E
      E' (SUBE: E' ⊆ E):
      ↑histN ⊆ E →
      orf = Relaxed ∨ orf = AcqRel →
      or = Relaxed ∨ or = AcqRel → ow = Relaxed ∨ ow = AcqRel →
      let VSC : vProp :=
        (<obj> ∀ t' s' v', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
                      (F γ t' s' v' ∨ F_read γ t' s' v') -∗
                      ⌜∃ vl, v' = #vl ∧ lit_comparable vr vl⌝)%I in
      let VS : vProp :=
        (∀ t' s', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
            ((<obj> (▷ F γ t' s' #vr ={E}=∗ ▷ Q1 t' s' ∗ ▷ Q2 t' s')) ∗
             (P -∗ ▷ Q2 t' s' ={E}=∗ ▷ GPS_SWSharedWriter t' s' #vr γ ∗
                ∃ s'', ⌜s' ⊑ s''⌝ ∗
                  (∀ t'' , ⌜(t' < t'')%positive⌝ -∗ ▷ GPS_SWSharedWriter t'' s'' vw γ -∗
                    (if dropR then ▷ GPS_SWSharedReader t'' s'' vw q γ else True) ={E}=∗
                    (<obj> (▷ Q1 t' s' ={E}=∗ ▷ F_read γ t' s' #vr)) ∗
                    |={E,E'}▷=> Q t'' s'' ∗ F γ t'' s'' vw)) ) ∧
            ▷ (∀ (v: lit), ⌜lit_neq vr v⌝ -∗
              <obj> ((F_read γ t' s' #v ={E}=∗ F_read γ t' s' #v ∗ R t' s' v) ∧
                     (F γ t' s' #v ={E}=∗ F γ t' s' #v ∗ R t' s' v))))%I in
      {{{ GPS_SWSharedReader t s v q γ
          ∗ (monPred_in Vb → ▷ GPS_INV IW γ)
          ∗ (▷ VSC)
          ∗ (if decide (AcqRel ⊑ ow) then VS else △{tid} VS)
          ∗ (if decide (AcqRel ⊑ ow) then P else △{tid} P) }}}
        CAS #l #vr vw orf or ow in tid @ E
      {{{ (b: bool) t' s' (v': lit), RET #b;
          (monPred_in Vb → ▷ GPS_INV IW γ) ∗ ⌜s ⊑ s'⌝
          ∗ ( (⌜b = true  ∧ v' = LitInt vr ∧ (t < t')%positive⌝
                ∗ (if dropR then True else GPS_SWSharedReader t' s' vw q γ)
                ∗ if decide (AcqRel ⊑ or) then Q t' s' else ▽{tid} Q t' s')
            ∨ (⌜b = false ∧ lit_neq vr v' ∧ (t ≤ t')%positive⌝
                ∗ GPS_SWSharedReader t' s' #v' q γ
                ∗ (if decide (AcqRel ⊑ orf) then R t' s' v' else ▽{tid} (R t' s' v'))
                ∗ (if decide (AcqRel ⊑ ow) then P else △{tid} P) ) ) }}}.
    Proof.
      iIntros (SUB ORF OR OW VSC VS Φ). iStartProof iProp.
      iIntros (?) "(R & Inv & Comp & VS & P)". iIntros (V1 ->) "Post".
      set Vo := V1 ⊔ Vb.
      iAssert ((▷ GPS_INV IW γ) Vo)%I with "[Inv]" as "Inv".
      { iApply "Inv". iPureIntro; solve_lat. }
      rewrite GPS_INV_eq monPred_at_later wp_eq.
      iIntros (𝓥3 Ext3) "H𝓥3 s". rewrite {1}Ext3 GPS_SWSharedReader_eq.
      iDestruct "R" as (tx) "(Prt & ex & Le)". iDestruct "Le" as %Let.
      iDestruct "Prt" as (Vr Va ?) "(#naV & R & SL)". iDestruct "SL" as %SL.
      (* end bookkeeping proof *)

      set Vw := (if decide (AcqRel ⊑ ow) then cur 𝓥3 else frel 𝓥3).
      iAssert (P Vw ∗ own tid (● to_latT 𝓥3))%I
        with "[P H𝓥3]" as "[P H𝓥3]".
      { destruct OW; subst ow; simpl; [|by iFrame].
        rewrite rel_mod_eq /rel_mod_def. iDestruct "P" as (𝓥) "[#H𝓥 P]".
        iDestruct (own_lat_auth_max with "[$H𝓥 $H𝓥3]") as %?.
        iFrame. by assert (𝓥.(frel) ⊑ 𝓥3.(frel)); [f_equiv|done]. }
      iAssert (VS Vw ∗ own tid (● to_latT 𝓥3))%I with "[H𝓥3 VS]" as "[VS H𝓥3]".
      { destruct OW; subst ow; simpl; [|by iFrame].
        rewrite monPred_at_embed rel_mod_eq /rel_mod_def.
        iDestruct "VS" as (𝓥) "[#H𝓥 VS]".
        iDestruct (own_lat_auth_max with "[$H𝓥 $H𝓥3]") as %?.
        iFrame. by assert (𝓥.(frel) ⊑ 𝓥3.(frel)); [f_equiv|done]. }

      iApply program_logic.weakestpre.wp_fupd.
      iApply (GPSRaw_CAS_int_strong_write IP l _ t tx s (v, Vr) orf or ow vr vw
              γ γ q Vo Va _ _ E E' P Q Q1 Q2 R dropR SL Let SUBE SUB ORF OR OW
              with "[$s $P $Inv $R $naV $ex Comp VS]").
      { rewrite monPred_at_objectively. iSplitL "Comp"; [by iNext..|].
        iIntros (t' s' ? Ext' Exts). iSpecialize ("VS" $! t' s' Exts).
        iSplit; [rewrite bi.and_elim_l|rewrite bi.and_elim_r]; last first.
        { iNext. iIntros (t0). by iApply "VS". }
        iDestruct "VS" as "[$ VS]". rewrite /Vw Ext'.
        iIntros (? ->) "P". iIntros (? ->) "Q".
        iMod ("VS" with "P [$Q]") as "[W VS]". iSplitL "W".
        { rewrite GPS_SWSharedWriter_eq.
          iDestruct "W" as (?) "(?&?&?)". iExists _. by iFrame. }
        iDestruct "VS" as (s'') "[Le' VS]". iDestruct "Le'" as %Le'.
        iExists s''. iFrame (Le'). iIntros "!>" (t''). iSpecialize ("VS" $! t'').
        iIntros (? -> Ext'' ? ->) "W". iSpecialize ("VS" with "[//] [W]").
        { rewrite GPS_SWSharedWriter_eq.
          iDestruct "W" as (?) "(?&?&?)". iExists _. by iFrame. }
        iIntros (? ->) "RD". rewrite GPS_SWSharedReader_eq.
        by iMod ("VS" with "RD") as "[$ $]". }

      iNext. iIntros (b v' s' t' V' Vt' 𝓥4) "(s' & H & Inv & CASE)".
      iDestruct "H" as %(Ext4 & Exts & SL' & ?).
      iMod (own_lat_auth_update with "H𝓥3") as "[$ #H𝓥4]"; [done|]. iModIntro.
      assert (V1 ⊑ 𝓥4.(cur)) by (etrans; [eassumption|by f_equiv]).
      iApply ("Post" $! _ t' s' v').
      iDestruct "CASE" as "[(H & #R & RS & QR)|(H & ex & #R & QR)]";
        iDestruct "H" as %(?&?&?&Le); iFrame (Exts);
          [iSplitR "RS QR"|iSplitR "QR ex"].
      - iIntros (V0 Ext0 ?). iNext.
        rewrite -(monPred_mono _ (Vo ⊔ 𝓥4.(cur)) V0); [done|].
        apply lat_join_lub; [|done].
        apply lat_join_lub; [by rewrite -Ext0|done].
      - iLeft. iSplit; [done|]. iSplitL "RS".
        + destruct dropR; [done|]. iExists tx. iFrame "RS".
          iSplit; [|iPureIntro; by eapply Pos.lt_le_incl, Pos.le_lt_trans].
          iExists Vt', Va, _. rewrite (monPred_mono _ _ _ (cur_mono _ _ Ext4)).
          by iFrame "naV R".
        + destruct OR; subst or; simpl; [|done].
          rewrite acq_mod_eq /acq_mod_def. iExists _. by iFrame.
      - iIntros (V0 Ext0 ?). iNext.
        rewrite -(monPred_mono _ (Vo ⊔ 𝓥4.(cur)) V0); [done|].
        apply lat_join_lub; [|done].
        apply lat_join_lub; by [rewrite -Ext0|].
      - iRight. iSplit; [done|]. iDestruct "QR" as "[QR P]".
        iSplitR "QR P"; [|iSplitL "QR"].
        + iExists tx. iFrame "ex". iSplit; [|iPureIntro; by etrans].
          iExists Vt', Va, _. rewrite (monPred_mono _ _ _ (cur_mono _ _ Ext4)).
          by iFrame "naV R".
        + rewrite acq_mod_eq.
          destruct ORF; subst orf; simpl; [iExists _; by iFrame|done].
        + rewrite rel_mod_eq.
          destruct OW; subst ow; simpl; [iExists _; by iFrame|done].
    Qed.

    Lemma GPS_SWRaw_SharedWriter_CAS_int_weak
      IW orf or ow (vr: Z) (vw: val) t s v γ P Q Q1 Q2 R tid Vb E E'
      (SUBE: E' ⊆ E):
      ↑histN ⊆ E →
      orf = Relaxed ∨ orf = AcqRel →
      or = Relaxed ∨ or = AcqRel → ow = Relaxed ∨ ow = AcqRel →
      let VSC : vProp :=
        (<obj> ∀ t' s' v', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
                      (F γ t' s' v' ∨ F_read γ t' s' v' ∨ IW l γ t' s' v') -∗
                      ⌜∃ vl, v' = #vl ∧ lit_comparable vr vl⌝)%I in
      let VS : vProp :=
        (∀ t' s', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
            ((▷ (P -∗ IW l γ t' s' #vr ={E}=∗ False)) ∧
             ((<obj> (▷ F γ t' s' #vr ={E}=∗ ▷ Q1 t' s' ∗ ▷ Q2 t' s')) ∗
             (P -∗ ▷ Q2 t' s' ={E}=∗ ▷ GPS_SWSharedWriter t' s' #vr γ ∗
                ∃ s'', ⌜s' ⊑ s''⌝ ∗
                  (∀ t'' , ⌜(t' < t'')%positive⌝ -∗ ▷ GPS_SWSharedWriter t'' s'' vw γ ={E}=∗
                      (<obj> (▷ Q1 t' s' ={E}=∗ ▷ F_read γ t' s' #vr)) ∗
                      |={E,E'}▷=> Q t'' s'' ∗ F γ t'' s'' vw))) ) ∧
            ▷ (∀ (v: lit), ⌜lit_neq vr v⌝ -∗
            <obj> ((F_read γ t' s' #v ={E}=∗ F_read γ t' s' #v ∗ R t' s' v) ∧
                   (F γ t' s' #v ={E}=∗ F γ t' s' #v ∗ R t' s' v) ∧
                   (IW l γ t' s' #v ={E}=∗ IW l γ t' s' #v ∗ R t' s' v))))%I in
      {{{ GPS_SWReader t s v γ
          ∗ (monPred_in Vb → ▷ GPS_INV IW γ)
          ∗ (▷ VSC)
          ∗ (if decide (AcqRel ⊑ ow) then VS else △{tid} VS)
          ∗ (if decide (AcqRel ⊑ ow) then P else △{tid} P) }}}
        CAS #l #vr vw orf or ow in tid @ E
      {{{ (b: bool) t' s' (v': lit), RET #b;
          (monPred_in Vb → ▷ GPS_INV IW γ) ∗ ⌜s ⊑ s'⌝
          ∗ ( (⌜b = true  ∧ v' = LitInt vr ∧ (t < t')%positive⌝
                ∗ GPS_SWReader t' s' vw γ
                ∗ if decide (AcqRel ⊑ or) then Q t' s' else ▽{tid} Q t' s')
            ∨ (⌜b = false ∧ lit_neq vr v' ∧ (t ≤ t')%positive⌝
                ∗ GPS_SWReader t' s' #v' γ
                ∗ (if decide (AcqRel ⊑ orf) then R t' s' v' else ▽{tid} (R t' s' v'))
                ∗ (if decide (AcqRel ⊑ ow) then P else △{tid} P) ) ) }}}.
    Proof.
      iIntros (SUB ORF OR OW VSC VSW Φ). iStartProof iProp.
      iIntros (?) "(R & Inv & Comp & VS & P)". iIntros (V1 ->) "Post".
      set Vo := V1 ⊔ Vb.
      iAssert ((▷ GPS_INV IW γ) Vo)%I with "[Inv]" as "Inv".
      { iApply "Inv". iPureIntro; solve_lat. }
      rewrite GPS_INV_eq monPred_at_later wp_eq.
      iIntros (𝓥3 Ext3) "H𝓥3 s". rewrite {1}Ext3 GPS_SWReader_eq.
      iDestruct "R" as (Vr Va ?) "(#naV & R & SL)". iDestruct "SL" as %SL.
      (* end bookkeeping proof *)

      set Vw := (if decide (AcqRel ⊑ ow) then cur 𝓥3 else frel 𝓥3).
      iAssert (P Vw ∗ own tid (● to_latT 𝓥3))%I
        with "[P H𝓥3]" as "[P H𝓥3]".
      { destruct OW; subst ow; simpl; [|by iFrame].
        rewrite rel_mod_eq /rel_mod_def. iDestruct "P" as (𝓥) "[#H𝓥 P]".
        iDestruct (own_lat_auth_max with "[$H𝓥 $H𝓥3]") as %?.
        iFrame. by assert (𝓥.(frel) ⊑ 𝓥3.(frel)); [f_equiv|done]. }
      iAssert (VSW Vw ∗ own tid (● to_latT 𝓥3))%I with "[H𝓥3 VS]" as "[VS H𝓥3]".
      { destruct OW; subst ow; simpl; [|by iFrame].
        rewrite monPred_at_embed rel_mod_eq /rel_mod_def.
        iDestruct "VS" as (𝓥) "[#H𝓥 VS]".
        iDestruct (own_lat_auth_max with "[$H𝓥 $H𝓥3]") as %?.
        iFrame. by assert (𝓥.(frel) ⊑ 𝓥3.(frel)); [f_equiv|done]. }

      iApply program_logic.weakestpre.wp_fupd.
      iApply (GPSRaw_CAS_int_weak_write IP l _ t s (v, Vr) orf or ow vr vw
                    γ γ Vo Va _ _ E E' P Q Q1 Q2 R SL SUBE SUB ORF OR OW
              with "[$s $P $Inv $R $naV Comp VS]").
      { rewrite monPred_at_objectively. iSplitL "Comp"; [by iNext..|].
        iIntros (t' s' ? Ext' Exts). iSpecialize ("VS" $! t' s' Exts).
        iSplit; [rewrite bi.and_elim_l|rewrite bi.and_elim_r]; last first.
        { iNext. iIntros (t0). by iApply "VS". } rewrite /Vw Ext'.
        iSplit; [by rewrite bi.and_elim_l|]. rewrite bi.and_elim_r.
        iDestruct "VS" as "[$ VS]".
        iIntros (? ->) "P". iIntros (? ->) "Q".
        iMod ("VS" with "P [$Q]") as "[W VS]". iSplitL "W".
        { rewrite GPS_SWSharedWriter_eq.
          iDestruct "W" as (?) "(?&?&?)". iExists _. by iFrame. }
        iDestruct "VS" as (s'') "[Le' VS]". iDestruct "Le'" as %Le'.
        iExists s''. iFrame (Le'). iIntros "!>" (t''). iSpecialize ("VS" $! t'').
        iIntros (? -> Ext'' ? ->) "W". iApply ("VS" with "[//] [W]").
        rewrite GPS_SWSharedWriter_eq.
        iDestruct "W" as (?) "(?&?&?)". iExists _. by iFrame. }

      iNext. iIntros (b v' s' t' V' Vt' 𝓥4) "(s' & H & Inv & CASE)".
      iDestruct "H" as %(Ext4 & Exts & SL' & ?).
      iMod (own_lat_auth_update with "H𝓥3") as "[$ #H𝓥4]"; [done|]. iModIntro.
      assert (V1 ⊑ 𝓥4.(cur)) by (etrans; [eassumption|by f_equiv]).
      iApply ("Post" $! _ t' s' v').
      iDestruct "CASE" as "[(H & R & QR)|(H & R & QR)]";
        iDestruct "H" as %(?&?&?&Le); iFrame (Exts); iSplitR "QR R".
      - iIntros (V0 Ext0 ?). iNext.
        rewrite -(monPred_mono _ (Vo ⊔ 𝓥4.(cur)) V0); [done|].
        apply lat_join_lub; [|done].
        apply lat_join_lub; [by rewrite -Ext0|done].
      - iLeft. iSplit; [done|]. iSplitR "QR".
        + iExists Vt', Va, _. rewrite (monPred_mono _ _ _ (cur_mono _ _ Ext4)).
          by iFrame "naV R".
        + destruct OR; subst or; simpl; [|done].
          rewrite acq_mod_eq /acq_mod_def. iExists _. by iFrame.
      - iIntros (V0 Ext0 ?). iNext.
        rewrite -(monPred_mono _ (Vo ⊔ 𝓥4.(cur)) V0); [done|].
        apply lat_join_lub; [|done].
        apply lat_join_lub; by [rewrite -Ext0|].
      - iRight. iSplit; [done|]. iDestruct "QR" as "[QR P]".
        iSplitR "QR P"; [|iSplitL "QR"].
        + iExists Vt', Va, _. rewrite (monPred_mono _ _ _ (cur_mono _ _ Ext4)).
          by iFrame "naV R".
        + rewrite acq_mod_eq.
          destruct ORF; subst orf; simpl; [iExists _; by iFrame|done].
        + rewrite rel_mod_eq.
          destruct OW; subst ow; simpl; [iExists _; by iFrame|done].
    Qed.
  End SingleWriter.

  Notation exwr_ag γ t := (own γ (to_agree t)).
  Section Plain.
    Context `{!gpsExAgG Σ}.
    Definition GPS_PPInv_def γn : vProp :=
      (∃ γ γx, ⌜γn = encode (γ,γx)⌝ ∗ gps_inv_vProp (λ _ _ _ _ _, True) γ γn false
          ∗ ⎡ (∃ ζ, Writer γ ζ) ∗ (∃ tx, exwr γ tx 1%Qp ∗ exwr_ag γx tx) ⎤)%I.
    Definition GPS_PPInv_aux : seal (@GPS_PPInv_def). by eexists. Qed.
    Definition GPS_PPInv := unseal (@GPS_PPInv_aux).
    Definition GPS_PPInv_eq : @GPS_PPInv = _ := seal_eq _.

    Definition GPS_PP_Local_def t s v γn : vProp :=
      (∃ γ γx tx, ⌜γn = encode (γ,γx) ∧ tx ⊑ t⌝ ∗ PrtSeen l t s v γ
          ∗ ⎡ exwr_ag γx tx ⎤)%I.
    Definition GPS_PP_Local_aux : seal (@GPS_PP_Local_def). by eexists. Qed.
    Definition GPS_PP_Local := unseal (@GPS_PP_Local_aux).
    Definition GPS_PP_Local_eq : @GPS_PP_Local = _ := seal_eq _.

    Global Instance GPS_PP_Local_persistent : Persistent (GPS_PP_Local t s v γn).
    Proof. rewrite GPS_PP_Local_eq. by apply _. Qed.

    Global Instance GPS_PP_Local_timeless : Timeless (GPS_PP_Local t s v γn).
    Proof. rewrite GPS_PP_Local_eq. by apply _. Qed.

    Lemma GPS_PP_Local_agree t1 t2 s1 s2 v1 v2 γn Vb E:
      GPS_PP_Local t1 s1 v1 γn -∗ GPS_PP_Local t2 s2 v2 γn -∗
      (monPred_in Vb → ▷ GPS_PPInv γn) ={E}=∗
      (monPred_in Vb → ▷ GPS_PPInv γn) ∗ ⌜s1 ⊑ s2 ∧ t1 ⊑ t2 ∨ s2 ⊑ s1 ∧ t2 ⊑ t1⌝.
    Proof.
      rewrite GPS_PP_Local_eq. iStartProof iProp. iIntros (?).
      (* start bookkeeping proof *)
      iDestruct 1 as (γ γx ? (Eq1 & _)) "[R1 _]".
      iDestruct "R1" as (???) "(_ & R1 & _)". iIntros (??).
      iDestruct 1 as (γ2 γx2 ? (Eq2 & _)) "[R2 _]".
      iDestruct "R2" as (???) "(_ & R2 & _)".
      iIntros (V ?) "Inv".
      set Vo := V ⊔ Vb.
      iAssert ((▷ GPS_PPInv γn) Vo)%I with "[Inv]" as "Inv".
      { iApply "Inv". iPureIntro; solve_lat. }
      rewrite GPS_PPInv_eq monPred_at_later.
      iDestruct "Inv" as (γ3 γx3) "(>Eq3 & Inv & Inv2)".
      iDestruct "Eq3" as %Eq3.
      rewrite Eq1 in Eq2. apply encode_inj in Eq2. inversion Eq2.
      rewrite Eq1 in Eq3. apply encode_inj in Eq3. inversion Eq3.
      subst γ2 γx2 γ3 γx3. clear Eq2 Eq3.
      (* end bookkeeping proof *)
      iMod (Reader_singleton_agree with "Inv R1 R2") as "[Inv $]".
      iModIntro. iIntros (V' ??). iExists γ, γx. iNext. iFrame (Eq1) "Inv2".
      rewrite -(monPred_mono _ Vo V'); [done|solve_lat].
    Qed.

    Lemma GPS_PPInv_dealloc b γ E (SUB: ↑histN ⊆ E):
      ⎡ hist_inv ⎤ -∗ ▷?b GPS_PPInv γ
      ={E}=∗ ∃ t s v, l ↦ v ∗ ▷?b F γ t s v.
    Proof.
      constructor. iIntros (?) "HInv". iIntros (? _).
      rewrite GPS_PPInv_eq /=. iDestruct 1 as (??) "(?&INV&?)".
      by iApply (GPSRaw_dealloc with "HInv INV").
    Qed.

    Lemma GPS_PPRaw_Init_vs b v s:
      l ↦ v -∗ (∀ t γ, ▷?b F γ t s v)
      ==∗ ∃ γ t, GPS_PP_Local t s v γ ∗ ▷?b GPS_PPInv γ.
    Proof.
      iStartProof iProp. iIntros (?) "own".
      iIntros (V->) "F".
      rewrite own_loc_na_eq GPS_PP_Local_eq.
      iDestruct "own" as (t m) "[own [% LE]]".
      iDestruct "own" as (???) "[[ALL [% AT]] [own [at [aw na]]]]".
      iDestruct "AT" as %(AT & AW & NA).
      iDestruct "ALL" as %ALL. iDestruct "LE" as %LE.
      iMod (own_alloc (to_agree t)) as (γx) "#exAg"; [done|].
      iSpecialize ("F" $! t).
      iMod (GPSRaw_Init_vs IP _ _ _ t v s m with "own at aw na F")
        as (γ) "(Inv & ox & W & R & na)"; [done..|].
      iModIntro. iExists (encode (γ,γx)), t. iSplitR "Inv W ox".
      - iExists γ, γx, t. iFrame "exAg". iSplit; [done|].
        iExists _, _, _. iFrame "R na". iPureIntro.
         by eapply alloc_local_singleton_seen_local.
      - iNext. rewrite GPS_PPInv_eq.
        iExists γ, γx. iSplit; [done|]. iSplitR "W ox".
        + have Ext2: default ∅ (mrel m) ⊔ {[l := [{t,∅,∅,∅}]]} ⊔ V ⊑ V.
          { apply lat_join_lub; [apply lat_join_lub|done]; [by rewrite LE|].
            destruct ALL as [t' [m' [[??]%lookup_singleton_Some [ND SL]]]].
            subst t' m'. by eapply seen_local_view_sqsubseteq. }
          iApply (monPred_mono _ _ _ Ext2). by iFrame.
        + iSplitL "W"; iExists _; by iFrame.
    Qed.

    Lemma GPS_PPRaw_Init b v s E tid :
      ↑histN ⊆ E →
      {{{ l ↦ ? ∗ (∀ t γ, ▷?b F γ t s v) }}}
        #l <- v in tid @ E
      {{{ γ t, RET #☠; GPS_PP_Local t s v γ ∗ ▷?b GPS_PPInv γ }}}.
    Proof.
      iIntros (? Φ) "(own & F) Post". iApply wp_fupd. wp_write.
      iMod (GPS_PPRaw_Init_vs with "own F") as (γ t) "?". by iApply "Post".
    Qed.

    Lemma GPS_PPRaw_Local_Read R o t s v γn tid Vb E:
      ↑histN ⊆ E → o = Relaxed ∨ o = AcqRel →
      {{{ GPS_PP_Local t s v γn ∗ (monPred_in Vb → ▷ GPS_PPInv γn) ∗
          (▷ ∀ t' s' v', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
             <obj> ((F_read γn t' s' v' ={E}=∗ F_read γn t' s' v' ∗ R t' s' v') ∧
                    (F γn t' s' v' ={E}=∗ F γn t' s' v' ∗ R t' s' v'))) }}}
        Read o #l in tid @ E
      {{{ t' s' v', RET v'; ⌜s ⊑ s' ∧ t ⊑ t'⌝
          ∗ GPS_PP_Local t' s' v' γn ∗ (monPred_in Vb → ▷ GPS_PPInv γn)
          ∗ if (decide (AcqRel ⊑ o)) then R t' s' v'
                                     else ▽{tid} R t' s' v' }}}.
    Proof.
      iIntros (SUB OR Φ). iStartProof iProp. iIntros (V1) "(R & Inv & D)".
      set Vo := V1 ⊔ Vb.
      iAssert ((▷ GPS_PPInv γn) Vo)%I with "[Inv]" as "Inv".
      { iApply "Inv". iPureIntro; solve_lat. }
      rewrite GPS_PPInv_eq monPred_at_later wp_eq.
      iIntros (V2 ?) "Post". iIntros (𝓥3 Ext3) "H𝓥3 s".
      have Ext3' : V1 ⊑ 𝓥3.(cur) by etrans.
      rewrite {1}(monPred_mono _ _ _ Ext3') GPS_PP_Local_eq.
      iDestruct "R" as (γ γx tx (Eq1 & Ext1)) "[PS Ex]".
      iDestruct "PS" as (Vr Va ?) "(#naV & R & SL)". iDestruct "SL" as %SL.
      iDestruct "Inv" as (γ1 γx1) "(>Eq2 & Inv & Inv2 & ex)".
      iDestruct "Eq2" as %Eq2. rewrite Eq1 in Eq2. apply encode_inj in Eq2.
      inversion Eq2. subst γ1 γx1. clear Eq2.
      iDestruct "ex" as (tx') "[>ex >exAg]".
      iDestruct (own_valid_2 with "Ex exAg") as %?%agree_op_invL'. subst tx'.
      (* end bookkeeping proof *)

      iApply program_logic.weakestpre.wp_fupd.
      iApply (GPSRaw_Read_ex IP l _ R tx t 1%Qp s _ o γ _ false _ _ _ _ E SL Ext1
                  SUB OR with "[$s D $Inv $ex $R $naV]").
      { iNext. iIntros (t' s' v' Ext). iSpecialize ("D" $! t' s' v' Ext).
        rewrite monPred_at_objectively. by iApply "D". }

      iNext. iIntros (t' s' V' 𝓥4 v') "(s' & Ext4 & Inv & ex & Rs & R)".
      iDestruct "Ext4" as %(ExtV & Exts & Extt & SL' & LeV').
      iDestruct "R" as (Vt') "[R ?]".
      iMod (own_lat_auth_update with "H𝓥3") as "[$ #H𝓥4]"; [done|]. iModIntro.

      assert (V2 ⊑ 𝓥4.(cur)) by (etrans; [eassumption|by f_equiv]).
      iApply ("Post" $! t' s' v'). iSplit; [done|].
      iSplitL "R Ex"; [|iSplitR "Rs"].
      - iExists γ, γx, tx. iFrame "Ex".
        iSplit; [iPureIntro; split|]; [done|by etrans|].
        iExists Vt', Va, _. rewrite (monPred_mono _ _ _ (cur_mono _ _ ExtV)).
        by iFrame (SL') "naV R".
      - iIntros (V0 Ext0 ?). iNext. iExists γ, γx. iFrame (Eq1) "Inv2".
        iSplitR "ex exAg".
        + rewrite -(monPred_mono _ (Vo ⊔ 𝓥4.(cur)) V0); [done|].
          apply lat_join_lub; [|done].
          apply lat_join_lub; [by rewrite -Ext0 -ExtV|done].
        + iExists tx. by iFrame.
      - rewrite acq_mod_eq. destruct OR; subst o; simpl; [iExists _|]; by iFrame.
    Qed.

    Lemma GPS_PPRaw_Local_Write o t v v' s s' γn Vb tid E (FIN: final_st s'):
      ↑histN ⊆ E → o = Relaxed ∨ o = AcqRel →
      let VS : vProp := (∀ t' : time, ⌜(t < t')%positive⌝ -∗
                          GPS_PP_Local t' s' v' γn ={E}=∗ F γn t' s' v')%I in
      {{{ GPS_PP_Local t s v γn ∗ (monPred_in Vb → ▷ GPS_PPInv γn)
          ∗ ▷ if decide (AcqRel ⊑ o) then VS else △{tid} VS }}}
        Write o #l v' in tid @ E
      {{{ t', RET #☠; GPS_PP_Local t' s' v' γn ∗ (monPred_in Vb → ▷ GPS_PPInv γn) }}}.
    Proof.
      iIntros (SUB OW VS Φ). iStartProof iProp. iIntros (V1) "(R & Inv & VS)".
      set Vo := V1 ⊔ Vb.
      iAssert ((▷ GPS_PPInv γn) Vo)%I with "[Inv]" as "Inv".
      { iApply "Inv". iPureIntro; solve_lat. }
      rewrite GPS_PPInv_eq monPred_at_later wp_eq.
      iIntros (V2 ?) "Post". iIntros (𝓥3 Ext3) "H𝓥3 s".
      have Ext3' : V1 ⊑ 𝓥3.(cur) by etrans.
      rewrite {1}(monPred_mono _ _ _ Ext3') GPS_PP_Local_eq.
      iDestruct "R" as (γ γx tx (Eq1 & Ext1)) "[PS Ex]".
      iDestruct "PS" as (Vr Va ?) "(#naV & R & SL)". iDestruct "SL" as %SL.
      iDestruct "Inv" as (γ1 γx1) "(>Eq2 & Inv & W & ox)".
      iDestruct "Eq2" as %Eq2. rewrite Eq1 in Eq2.
      apply encode_inj in Eq2. inversion Eq2. subst γ1 γx1. clear Eq2.
      iDestruct "W" as (ζ) ">W". iDestruct "ox" as (tx') ">[ex #exAg]".
      iDestruct (own_valid_2 with "Ex exAg") as %?%agree_op_invL'. subst tx'.
      (* end bookkeeping proof *)

      iApply program_logic.weakestpre.wp_fupd.
      iApply (GPSRaw_Write IP _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ SL FIN Ext1 SUB OW
              with "[$s $Inv $R $naV $ex $W]").
      iNext. iIntros (t' V' 𝓥4) "(s' & H & Inv & #R' & ex & W')".
      iDestruct "H" as %(Ext4 & ? & Lt & SL' & SL1 & ? & ? & LeV).

      have Ext4': V1 ⊑ 𝓥4.(cur) by etrans; [eassumption|by f_equiv].
      have ?: tx ⊑ t'. { etrans; [apply Ext1|by apply Pos.lt_le_incl]. }

      iAssert (GPS_PP_Local t' s' v' γn (V' ⊔ Va))%I as "PP'".
      { rewrite GPS_PP_Local_eq. iExists γ, γx, tx. iFrame "exAg". iSplit; [done|].
        iExists V', Va, _. iFrame "R'". iSplit.
        - rewrite naViewSeen_eq. iDestruct "naV" as "[$ [? %]]".
          iPureIntro. split; [solve_lat|done].
        - iPureIntro. eapply (seen_local_mono); [|apply SL1]. solve_lat. }

      iAssert (gps_inv IP _ l γ γn false (Vo ⊔ 𝓥4.(cur)) ∗ own tid (● to_latT 𝓥3))%I
        with "[> Inv H𝓥3 VS PP']" as "[Inv H𝓥3]".
      { destruct OW; subst o; simpl.
        - rewrite rel_mod_eq /rel_mod_def. iDestruct "VS" as (𝓥F) "[H𝓥F VS]".
          iDestruct (own_lat_auth_max with "[$H𝓥3 $H𝓥F]") as %H𝓥F.
          iClear "H𝓥F". rewrite (monPred_mono _ _ _ (frel_mono _ _ H𝓥F)).
          assert (Le3: 𝓥3.(frel) ⊑ V').
          { etrans; [|by apply LeV]. simpl. by f_equiv. }
          rewrite (monPred_mono _ _ _ Le3).
          iFrame. iMod ("VS" $! t' Lt with "PP'") as "F". by iApply "Inv".
        - iFrame. iMod ("VS" $! t' Lt with "PP'") as "F". by iApply "Inv". }

      iMod (own_lat_auth_update with "H𝓥3") as "[$ _]"; [done|]. iModIntro.
      assert (V2 ⊑ 𝓥4.(cur)) by (etrans; [eassumption|by f_equiv]).
      iApply "Post". iSplitL "".
      { iExists γ, γx, tx. iFrame "exAg". iSplit; [done|].
        iExists V', Va, _. rewrite (monPred_mono _ _ _ (cur_mono _ _ Ext4)).
        iFrame (SL') "naV R'". }

      iIntros (V0 Ext0 ?). iNext. iExists γ, γx. iFrame (Eq1).
      iSplitL "Inv".
      - rewrite -(monPred_mono _ (Vo ⊔ 𝓥4.(cur)) V0); [done|].
        apply lat_join_lub; [|done].
        apply lat_join_lub; [by rewrite Ext3' Ext4|done].
      - iSplitL "W'"; [by iExists _|iExists tx; by iFrame "ex exAg"].
    Qed.

    Lemma GPS_PPRaw_Local_CAS_int_simple
      orf or ow v (vr: Z) (vw: val) t s P Q Q1 Q2 R γn tid Vb E :
      ↑histN ⊆ E →
      orf = Relaxed ∨ orf = AcqRel →
      or = Relaxed ∨ or = AcqRel → ow = Relaxed ∨ ow = AcqRel →
      let VSC : vProp :=
        (<obj> ∀ t' s' v', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
                      (F γn t' s' v' ∨ F_read γn t' s' v') -∗
                      ⌜∃ vl, v' = #vl ∧ lit_comparable vr vl⌝)%I in
      let VS : vProp :=
        (∀ t' s', ⌜s ⊑ s' ∧ t ⊑ t'⌝ -∗
          ((<obj> (▷ F γn t' s' #vr ={E}=∗ ▷ Q1 t' s' ∗ ▷ Q2 t' s')) ∗
           (P -∗ ▷ Q2 t' s' ={E}=∗ ∃ s'', ⌜s' ⊑ s''⌝ ∗
              (∀ t , ⌜(t' < t)%positive⌝ -∗ ▷ (GPS_PP_Local t s'' vw γn) ={E}=∗
                  (<obj> (▷ Q1 t' s' ={E}=∗ ▷ F_read γn t' s' #vr)) ∗
                  |={E}▷=> Q t s'' ∗ F γn t s'' vw)) ) ∧
          (▷ ∀ (v: lit), ⌜lit_neq vr v⌝ -∗
              <obj> ((F_read γn t' s' #v ={E}=∗ F_read γn t' s' #v ∗ R t' s' v) ∧
                     (F γn t' s' #v ={E}=∗ F γn t' s' #v ∗ R t' s' v))))%I in
      {{{ GPS_PP_Local t s v γn ∗ (monPred_in Vb → ▷ GPS_PPInv γn)
          ∗ ▷ VSC
          ∗ (if decide (AcqRel ⊑ ow) then VS else △{tid} VS)
          ∗ (if decide (AcqRel ⊑ ow) then P  else △{tid} P )  }}}
        CAS #l #vr vw orf or ow in tid @ E
      {{{ (b: bool) t' s' (v': lit), RET #b;
          (monPred_in Vb → ▷ GPS_PPInv γn) ∗ ⌜s ⊑ s'⌝
          ∗ ( (⌜b = true  ∧ v' = LitInt vr ∧ (t < t')%positive⌝
                ∗ GPS_PP_Local t' s' vw γn
                ∗ if decide (AcqRel ⊑ or) then Q t' s' else ▽{tid} Q t' s')
            ∨ (⌜b = false ∧ lit_neq vr v' ∧ (t ≤ t')%positive⌝
                ∗ GPS_PP_Local t' s' #v' γn
                ∗ (if decide (AcqRel ⊑ orf) then R t' s' v' else ▽{tid} (R t' s' v'))
                ∗ (if decide (AcqRel ⊑ ow) then P  else △{tid} P ))) }}}.
    Proof.
      iIntros (SUB ORF OR OW VSC VSW Φ). iStartProof iProp.
      iIntros (V1) "(R & Inv & Comp & VS & P)".
      set Vo := V1 ⊔ Vb.
      iAssert ((▷ GPS_PPInv γn) Vo)%I with "[Inv]" as "Inv".
      { iApply "Inv". iPureIntro; solve_lat. }
      rewrite GPS_PPInv_eq monPred_at_later wp_eq.
      iIntros (V2 ?) "Post". iIntros (𝓥3 Ext3) "H𝓥3 s".
      have Ext3' : V1 ⊑ 𝓥3.(cur) by etrans.
      rewrite {1}(monPred_mono _ _ _ Ext3') GPS_PP_Local_eq.
      iDestruct "R" as (γ γx tx (Eq1 & Ext1)) "[PS Ex]".
      iDestruct "PS" as (Vs Va ?) "(#naV & R & SL)". iDestruct "SL" as %SL.
      iDestruct "Inv" as (γ1 γx1) "(>Eq2 & Inv & >W & ox)".
      iDestruct "Eq2" as %Eq2. rewrite Eq1 in Eq2.
      apply encode_inj in Eq2. inversion Eq2. subst γ1 γx1. clear Eq2.
      iDestruct "ox" as (tx') ">[ex #exAg]".
      iDestruct (own_valid_2 with "Ex exAg") as %?%agree_op_invL'. subst tx'.
      (* end bookkeeping proof *)

      set Vw := if decide (AcqRel ⊑ ow) then 𝓥3.(cur) else 𝓥3.(frel).
      iAssert (P Vw ∗ own tid (● to_latT 𝓥3))%I with "[P H𝓥3]" as "[P H𝓥3]".
      { destruct OW; subst ow; simpl; [|by iFrame].
        rewrite rel_mod_eq /rel_mod_def. iDestruct "P" as (𝓥) "[#H𝓥 P]".
        iDestruct (own_lat_auth_max with "[$H𝓥 $H𝓥3]") as %?.
        iFrame. by assert (𝓥.(frel) ⊑ 𝓥3.(frel)); [f_equiv|done]. }
      iAssert (VSW Vw ∗ own tid (● to_latT 𝓥3))%I with "[H𝓥3 VS]" as "[VS H𝓥3]".
      { destruct OW; subst ow; simpl; [|by iFrame].
        rewrite monPred_at_embed rel_mod_eq /rel_mod_def.
        iDestruct "VS" as (𝓥) "[#H𝓥 VS]".
        iDestruct (own_lat_auth_max with "[$H𝓥 $H𝓥3]") as %?.
        iFrame. by assert (𝓥.(frel) ⊑ 𝓥3.(frel)); [f_equiv|done]. }

      iApply program_logic.weakestpre.wp_fupd.
      (* TODO: make faster by sealing *)
      iApply (GPSRaw_CAS_int_simple IP l (λ _ _ _ _ _, True)%I t tx s (v, Vs)
              orf or ow vr vw γ γn _ Vo Va _ _ E P Q Q1 Q2 R SL Ext1 SUB ORF
              OR OW with "[$s $P $Inv $R $naV $ex $W Comp VS]").
      { iSplitL "Comp". rewrite monPred_at_objectively. by iNext.
        rewrite /VSW /Vw. iIntros (t' s' ? -> Ext').
        iSpecialize ("VS" $! t' s' Ext').
        iSplit; [rewrite bi.and_elim_l|rewrite bi.and_elim_r //].
        iDestruct "VS" as "[$ VS]". iIntros (? ->) "P". iIntros (? ->) "Q".
        iMod ("VS" with "P [$Q]") as (s'' Ext'') "VS". iIntros "!>".
        iExists s''. iFrame (Ext''). iIntros (t''). iSpecialize ("VS" $! t'').
        iIntros (? -> Ext2 ? ->) "Prt". iApply ("VS" $! Ext2 with "[Prt]").
        iNext. rewrite GPS_PP_Local_eq. iExists γ, γx, tx. iFrame "Prt exAg".
        iPureIntro. split; [done|]. destruct Ext' as [? Ext'].
        apply Pos.lt_le_incl. eapply Pos.le_lt_trans; [|exact Ext2].
        etrans; [exact Ext1|exact Ext']. }

      iNext. iIntros (b v' s' t' V' Vt' 𝓥4) "(s' & H & Inv & ex & W' & CASE)".
      iDestruct "H" as %(Ext4 & Exts & SL' & ?).
      iMod (own_lat_auth_update with "H𝓥3") as "[$ #H𝓥4]"; [done|]. iModIntro.
      assert (V2 ⊑ 𝓥4.(cur)) by (etrans; [eassumption|by f_equiv]).
      iApply ("Post" $! _ t' s' v').
      iDestruct "CASE" as "[(H & R & QR)|(H & R & QR)]";
        iDestruct "H" as %(?&?&?&Le); iFrame (Exts); iSplitR "QR R".
      - iIntros (V0 Ext0 ?). iNext. iExists γ, γx. iFrame (Eq1) "W'".
        iSplitR "ex"; [|iExists tx; by iFrame].
        rewrite -(monPred_mono _ (Vo ⊔ 𝓥4.(cur)) V0); [done|].
        apply lat_join_lub; [|done].
        apply lat_join_lub; [by rewrite Ext3' Ext4|done].
      - iLeft. iSplit; [done|]. iSplitR "QR".
        + iExists γ, γx, tx. iFrame "exAg".
          iSplit; [iPureIntro; split|]; [done|..].
          * destruct Le as [Le _]. by etrans; [apply Ext1|apply Pos.lt_le_incl].
          * iExists Vt', Va, _. rewrite (monPred_mono _ _ _ (cur_mono _ _ Ext4)).
            by iFrame (SL') "naV R".
        + destruct OR; subst or; simpl; [|done].
          rewrite acq_mod_eq /acq_mod_def. iExists _. by iFrame.
      - iIntros (V0 Ext0 ?). iNext. iExists γ, γx. iFrame (Eq1) "W'".
        iSplitR "ex"; [|iExists tx; by iFrame].
        rewrite -(monPred_mono _ (Vo ⊔ 𝓥4.(cur)) V0); [done|].
        apply lat_join_lub; [|done].
        apply lat_join_lub; [by rewrite Ext3' Ext4|done].
      - iRight. iSplit; [done|]. iDestruct "QR" as "[QR P]".
        iSplitR "QR P"; [|iSplitL "QR"].
        + iExists γ, γx, tx. iFrame "exAg".
          iSplit; [iPureIntro; split|]; [done|by etrans|].
          iExists Vt', Va, _. rewrite (monPred_mono _ _ _ (cur_mono _ _ Ext4)).
          by iFrame (SL') "naV R".
        + rewrite acq_mod_eq.
          destruct ORF; subst orf; simpl; [iExists _; by iFrame|done].
        + rewrite rel_mod_eq.
          destruct OW; subst ow; simpl; [iExists _; by iFrame|done].
    Qed.
  End Plain.

End SimpleRules.

Section GName.
  Context `{!noprolG Σ, !gpsG Σ Prtcl} (l : loc)
          (IP : gname → interpC Σ Prtcl).
  Local Notation iProp := (iProp Σ).
  Local Notation vProp := (vProp Σ).

  Lemma GPS_SWRaw_Init_vs_gname b v s (Q: time → gname → vProp) E :
    l ↦ v -∗ (∀ t γ, GPS_SWWriter l t s v γ ={E}=∗ ▷?b (IP γ) true l γ t s v ∗ Q t γ)
    ={E}=∗ ∃ γ t, ∀ IW, ▷?b GPS_INV (IP γ) l IW γ ∗ Q t γ.
  Proof.
    iStartProof iProp. iIntros (?) "own1".
    iIntros (V->) "F". rewrite own_loc_na_eq.
    iDestruct "own1" as (t m) "[own1 [% LE]]".
    iDestruct "own1" as (???) "[[ALL [% [% [% %]]]] [own1 [at1 [aw1 na1]]]]".
    iDestruct "ALL" as %ALL1. iDestruct "LE" as %LE1.
    have SL1:= alloc_local_singleton_seen_local _ _ _ _ ALL1.
    iMod (GPSRaw_Init_strong_vs_gname l IP b t v s m _ _ _
            true V _ id (Q t) with "own1 at1 aw1 na1 [F]")
            as (γ) "Inv"; [done..| |].
    { iIntros (γ) "ex1 W1 R1 na1". rewrite /= GPS_SWWriter_eq.
      iMod ("F" $! t γ with "[ex1 W1 na1 R1]") as "($ & $)"; [..|done].
      iExists _. iFrame "ex1 W1". iSplit.
      - iExists _,_,_. iFrame "na1 R1". iPureIntro. split; [done|by rewrite LE1].
      - iPureIntro. move => ? [? /lookup_singleton_Some [-> //]]. }
    iModIntro. iExists γ, t. iIntros (IW). rewrite GPS_INV_eq.
    by iApply "Inv".
  Qed.
End GName.

Section Dual.
  Context `{!noprolG Σ, !gpsG Σ Prtcl1, !gpsG Σ Prtcl2} (l1 l2 : loc)
          (IP1 : gname → interpC Σ Prtcl1)
          (IP2 : gname → interpC Σ Prtcl2).
  Local Notation iProp := (iProp Σ).
  Local Notation vProp := (vProp Σ).

  Lemma GPS_SWRaw_Init_vs_strong_2 b v1 v2 (s1 : Prtcl1) (s2: Prtcl2)
    (Q: time → time → gname → gname → vProp) E :
    l1 ↦ v1 -∗ l2 ↦ v2 -∗
    (∀ t1 t2 γ1 γ2, GPS_SWWriter l1 t1 s1 v1 γ1 -∗
          GPS_SWWriter l2 t2 s2 v2 γ2 ={E}=∗ ▷?b (IP1 γ2) true l1 γ1 t1 s1 v1
          ∗ ▷?b (IP2 γ1) true l2 γ2 t2 s2 v2 ∗ Q t1 t2 γ1 γ2)
    ={E}=∗ ∃ γ1 γ2 t1 t2, ∀ IW1 IW2,
          ▷?b GPS_INV (IP1 γ2) l1 IW1 γ1 ∗
          ▷?b GPS_INV (IP2 γ1) l2 IW2 γ2 ∗ Q t1 t2 γ1 γ2.
  Proof.
    iStartProof iProp. iIntros (?) "own1". iIntros (? ->) "own2".
    iIntros (V->) "F". rewrite own_loc_na_eq.
    iDestruct "own1" as (t1 m1) "[own1 [% LE]]".
    iDestruct "own1" as (???) "[[ALL [% [% [% %]]]] [own1 [at1 [aw1 na1]]]]".
    iDestruct "ALL" as %ALL1. iDestruct "LE" as %LE1.
    have SL1:= alloc_local_singleton_seen_local _ _ _ _ ALL1.
    iDestruct "own2" as (t2 m2) "[own2 [% LE]]".
    iDestruct "own2" as (???) "[[ALL [% [% [% %]]]] [own2 [at2 [aw2 na2]]]]".
    iDestruct "ALL" as %ALL2. iDestruct "LE" as %LE2.
    have SL2:= alloc_local_singleton_seen_local _ _ _ _ ALL2.
    iMod (GPSRaw_Init_strong_vs_2 l1 l2 IP1 IP2 b t1 t2 v1 v2 s1 s2 m1 m2
            _ _ _ _ _ _ true true V _ id id (Q t1 t2)
      with "own1 own2 at1 at2 aw1 aw2 na1 na2 [F]") as (γ1 γ2) "Inv"; [done..| |].
    { iIntros (γ1 γ2) "ex1 W1 R1 na1 ex2 W2 R2 na2". rewrite /= 2!GPS_SWWriter_eq.
      iMod ("F" $! t1 t2 γ1 γ2 with "[ex1 W1 na1 R1] [ex2 W2 na2 R2]")
        as "($ & $ & $)"; [..|done].
      - iExists _. iFrame "ex1 W1". iSplit.
        + iExists _,_,_. iFrame "na1 R1". iPureIntro. split; [done|by rewrite LE1].
        + iPureIntro. move => ? [? /lookup_singleton_Some [-> //]].
      - iExists _. iFrame "ex2 W2". iSplit.
        + iExists _,_,_. iFrame "na2 R2". iPureIntro. split; [done|by rewrite LE2].
        + iPureIntro. move => ? [? /lookup_singleton_Some [-> //]]. }
    iModIntro. iExists γ1, γ2, t1, t2. iIntros (IW1 IW2). rewrite 2!GPS_INV_eq.
    by iApply "Inv".
  Qed.
End Dual.

Section Extra.
  Context `{!noprolG Σ, !gpsG Σ Prtcl}.
  Implicit Types (IP : interpC Σ Prtcl).

  Global Instance gps_inv_vProp_ne :
    NonExpansive (λ IP, gps_inv_vProp IP l IW γ γF b).
  Proof.
    move => ????? n f1 f2 EQ. constructor => ?.
    repeat apply bi.exist_ne => ?. repeat (apply bi.sep_ne; [done|]).
    rewrite /resBE /resD /gps_res.
    repeat apply bi.sep_ne; [..|done];
      (apply big_opM_ne; [|done] => ? m; destruct m.(mval); [|done|];
        case decide => ?//; try apply EQ);
      rewrite !monPred_at_or; repeat (apply bi.or_ne; apply EQ);
      (apply bi.or_ne; [apply EQ|apply bi.or_ne; [apply EQ|done]]).
  Qed.

  Global Instance GPS_INV_ne :
    NonExpansive (λ IP, GPS_INV IP l IW γ).
  Proof. move => ??????. rewrite 2!GPS_INV_eq. by apply gps_inv_vProp_ne. Qed.

  Global Instance GPS_PPInv_ne `{inG Σ (agreeR (leibnizC time))} :
    NonExpansive (λ IP, GPS_PPInv IP l γ).
  Proof.
    move => ?? n f1 f2 EQ. rewrite 2!GPS_PPInv_eq.
    repeat apply bi.exist_ne => ?. apply bi.sep_ne; [done|].
    apply bi.sep_ne; by [apply gps_inv_vProp_ne|].
  Qed.

  Lemma gps_inv_vProp_proper :
    Proper ((≡) ==> (=) ==> (≡) ==> (eq) ==> (eq) ==> (eq) ==> (≡)) gps_inv_vProp.
  Proof.
    move => ?? Eq1 ?? ? ?? Eq2 ?????????. subst. constructor => ?. simpl.
    do 7 apply bi.exist_proper => ?. do 5 (apply bi.sep_proper; [done|]).
    apply bi.sep_proper; last (apply bi.sep_proper; [|done]).
    - apply big_sepL_proper => ?[? m]? /=.
      destruct m.(mval); [|done|]; (case decide => ?; [apply Eq1|apply Eq2]).
    - apply big_sepL_proper => ?[? m]? /=.
      destruct m.(mval); [|done|]; case decide => ? /=; rewrite !monPred_at_or;
        apply bi.or_proper.
      apply Eq1. apply Eq1. apply Eq1. apply bi.or_proper. apply Eq1. apply Eq2.
      apply Eq1. apply Eq1. apply Eq1. apply bi.or_proper. apply Eq1. apply Eq2.
  Qed.

  Lemma GPS_INV_proper:
    Proper ((≡) ==> (=) ==> (≡) ==> (eq) ==> (≡)) GPS_INV.
  Proof.
    move => ????????????. subst. rewrite 2!GPS_INV_eq.
    by apply gps_inv_vProp_proper.
  Qed.

  Lemma GPS_PPInv_proper `{inG Σ (agreeR (leibnizC time))} :
    Proper ((≡) ==> (=) ==> (=) ==> (≡)) (λ IP l, GPS_PPInv IP l).
  Proof.
    move => ?? Eq ??????. subst. rewrite 2!GPS_PPInv_eq.
    do 2 apply bi.exist_proper => ?. apply bi.sep_proper; [done|].
    apply bi.sep_proper; [|done]. by apply gps_inv_vProp_proper.
  Qed.
  Global Instance InterpType_cofe : Cofe (interpC Σ Prtcl) := _.
  Global Instance interp_inhabited : Inhabited (interpC Σ Prtcl).
  Proof. constructor; repeat intro; exact True%I. Qed.
  Global Instance InterpCasType_cofe : Cofe (interpCasC Σ Prtcl) := _.
  Global Instance interpCas_inhabited : Inhabited (interpCasC Σ Prtcl).
  Proof. constructor; repeat intro; exact True%I. Qed.

  (* Recursive protocol *)
  (* From gpfsl.gps Require Import protocols.
   Program Example int : interpC Σ unitProtocol :=
    @fixpoint _ _ _ (λ (F : interpC Σ unitProtocol) b l s v, ∃ c, ⌜c = #1⌝ ∗ ▷ F b l s v)%I _.
  Next Obligation.
    repeat intro.
    apply bi.exist_ne => ? . apply bi.sep_ne; [done|]. apply bi.later_contractive.
    destruct n; [done|]. apply H.
  Qed. *)

End Extra.
