From iris.program_logic Require Import adequacy.
From iris.algebra Require Import auth.
From gpfsl.lang Require Import weakestpre.

Class noprolPreG (Σ : gFunctors) : Set := NoprolG {
  noprol_preG_invg :> invPreG Σ;
  noprol_inG :> inG Σ (authR histUR);
  hist_readG :> inG Σ (authR hist_readUR);
  noprol_freeable_inG :> inG Σ (authR hist_freeableUR);
  noprol_na_view_inG :> inG Σ viewR;
  hist_at_write_inG :> inG Σ (authR hist_writesUR);
  noprol_tview_inG :> inG Σ (authR (latUR tview_Lat));
  noprol_phys_inG :> inG Σ (authR (optionUR (exclR stateC)));
}.

Definition noprolΣ : gFunctors :=
  #[invΣ;
    GFunctor (authR histUR); GFunctor (authR hist_freeableUR);
    GFunctor (authR hist_readUR);
    GFunctor viewR; GFunctor (authR hist_writesUR);
    GFunctor (authR (latUR tview_Lat));
    GFunctor (authR (optionUR (exclR stateC)))].
Instance subG_noprolPreG {Σ} : subG noprolΣ Σ → noprolPreG Σ.
Proof. solve_inG. Qed.

Lemma noprol_adequacy Σ `{noprolPreG Σ} e (σ: state) 𝓥 φ:
  Wf σ →
  𝓥 ∈ σ.(mem) →
  (∀ `{noprolG Σ} tid , WP e in tid {{ v, ⌜φ v⌝ }}%I) →
  adequate NotStuck (e at 𝓥) σ (λ v _, φ v).
Proof.
  intros WFσ H𝓥 Hwp; eapply (wp_adequacy _ _); iIntros (Hinv ?).
  set (M := mem_cut σ.(mem) σ.(na)).
  iMod (own_alloc (● to_hist M)) as (hist_name) "hist".
  { split; eauto using ucmra_unit_leastN, mem_cut_valid. }
  iMod (own_alloc (● (∅ : hist_freeableUR))) as (hist_freeable_name) "hf"; [done|].
  iMod (own_alloc (● to_atr σ.(na))) as (hist_atread_name) "har".
  { split; eauto using ucmra_unit_leastN, to_atr_valid. }
  iMod (own_alloc (● to_atw σ.(na))) as (hist_atwrite_name) "haw".
  { split; eauto using ucmra_unit_leastN, to_atw_valid. }
  iMod (own_alloc (● to_nar σ.(na))) as (hist_naread_name) "hn".
  { split; eauto using ucmra_unit_leastN, to_nar_valid. }
  iMod (own_alloc (● to_latT 𝓥.(acq) ⋅ ◯ to_latT 𝓥.(acq)))
    as (hist_gtime_name) "[hV1 hV2]".
  { split; [|done]. move=> n /=. by rewrite left_id. }
  iMod (own_alloc (● Excl' σ ⋅ ◯ Excl' σ))
    as (hist_phys_name) "[HAσ HNσ]"; [done|].
  set (HISTG := HistG Σ _ _ _ _ _ _ _ hist_name hist_freeable_name
                      hist_atwrite_name
                      hist_atread_name hist_naread_name
                      hist_gtime_name hist_phys_name).
  iAssert (hist_ctx σ) with "[- hV2 HAσ HNσ]" as "CTX".
  { iExists _, _,_. iFrame. iPureIntro.
    split; [done|split; [done|split;[apply H𝓥|done]]]. }
  set (NOPROLG := NoProLG _ _).
  iExists (λ σ _, hist_interp σ). iSplitR "hV2".
  - rewrite /hist_interp /history_cmra.hist_phys_name /=. iFrame "HNσ".
    iApply (invariants.inv_alloc histN). iNext. iExists _. iFrame.
  - iMod (own_alloc (● to_latT 𝓥)) as (tid) "tid".
    { split; [|done]. intros. apply ucmra_unit_leastN. }
    iDestruct (Hwp _ tid) as "H".
    iSpecialize ("H" $! 𝓥.(cur)). rewrite wp_eq /wp_def /=.
    iModIntro. iApply (iris.program_logic.weakestpre.wp_wand with "[-]").
    + iApply ("H" with "[//] tid"). rewrite seen_eq /seen_def. by iFrame.
    + by iIntros ([v 𝓥']) "/= [_ %]".
Qed.

Lemma noprol_adequacy' Σ `{noprolPreG Σ} e φ:
  (∀ `{noprolG Σ} tid , WP e in tid {{ v, ⌜φ v⌝ }}%I) →
  adequate NotStuck (e at init_tview) (mkGB ∅ ∅) (λ v _, φ v).
Proof.
  intros. eapply noprol_adequacy=>//. split=>//. split=> l /=.
  - destruct (memory_loc_not_elem_of_dom (VAL := val) l ∅) as [EQ _].
    rewrite EQ //. intros t m. naive_solver.
  - destruct (memory_loc_not_elem_of_dom (VAL := val) l ∅) as [EQ _].
    rewrite EQ //.
Qed.
