From promising Require Import monotone.
From gpfsl Require Import lang notation.
From iris.program_logic Require Import adequacy.

(** Monotonicity of the reductions w.r.t. views *)

Lemma tview_mono_head_progress σ σ' e e' efs 𝓥1 𝓥2 𝓥1'
  (Ext: 𝓥1 ⊑ 𝓥2) (In: 𝓥2 ∈ σ.(mem)) (WF: Wf σ):
  nopro_lang.head_step (e at 𝓥1) σ (e' at 𝓥1') σ' efs →
  ∃ 𝓥2' σ2' efs' e2, nopro_lang.head_step (e at 𝓥2) σ (e2 at 𝓥2') σ2' efs'.
Proof.
  inversion_clear 1; [|inversion ExprStep; clear ExprStep; subst; simpl in *].
  - inversion BaseStep; subst; clear BaseStep; (destruct efs as [|? []]=>//; []);
    try by eexists _, _, [], _; repeat constructor.
    destruct (fork_head_step e0 σ' 𝓥2) as [?[??]]. eauto.
  - destruct (alloc_fresh_head_step n σ 𝓥2 In) as [?[??]]; [done|].
    by do 4 eexists.
  - have Eqn: Pos.to_nat (Z.to_pos n) = Z.to_nat n.
      { rewrite -Z2Nat.inj_pos. f_equal. by apply Z2Pos.id. }
    inversion PStep; clear PStep; subst. simpl in *.
    destruct (dealloc_head_step (Z.to_nat n) l σ 𝓥2) as [σ2' [𝓥2' STEP]];
      [| | | |apply WF|done|lia|].
    + have REMOVE := dealloc_step_remove _ _ _ _ _ _ _ DEALLOC.
      rewrite Eqn in REMOVE.
      move => n' /REMOVE /elem_of_intersection [/elem_of_difference [? _] _ ].
      by apply memory_loc_elem_of_dom.
    + move => n' Lt. simpl. etrans; last apply Ext.
      inversion DRF.
      assert (∃ 𝑚, 𝑚s !! n' = Some 𝑚) as [𝑚 In𝑚].
      { apply lookup_lt_is_Some_2.
        by rewrite (dealloc_step_length _ _ _ _ _ _ _ DEALLOC) Eqn. }
      destruct (LOCAL 𝑚) as [otl DRFL].
      { by eapply elem_of_list_lookup_2. }
      inversion DRFL.
      rewrite -OTL (dealloc_step_loc_eq _ _ _ _ _ _ _ DEALLOC _ _ In𝑚) in LAST.
      by apply LAST.
    + inversion DEALLOC; clear DEALLOC; subst.
      inversion_clear MEMALL. inversion DEALLOC. by rewrite Eqn in SIZE.
    + inversion DEALLOC; clear DEALLOC; subst.
      inversion_clear MEMALL. inversion DEALLOC. rewrite Eqn in ALLOC.
      move => ? /ALLOC /elem_of_difference [In1 NIn] DE.
      by apply NIn, elem_of_filter.
    + do 4 eexists. rewrite Z2Nat.id in STEP=>//. lia.
  - inversion PStep; clear PStep; subst. simpl in *.
    destruct (read_head_step (mloc 𝑚) o σ 𝓥2 In) as [?[?[??]]];
      [apply WF|apply WF|by inversion READ|..].
    + inversion READ=>EQ.
      change (σ.(mem)!!(𝑚.(mloc),𝑚.(mto)) = Some 𝑚.(mbase)) in IN.
      by rewrite memory_lookup_cell EQ in IN.
    + etrans; last by apply Ext.
      inversion DRF. inversion LOCAL. by rewrite OTL.
    + inversion READ. subst. inversion DRF. inversion LOCAL.
      move => ????. etrans; last by apply Ext. rewrite OTL. by apply MAX.
    + by do 4 eexists.
  - inversion PStep; clear PStep; subst. simpl in *.
    destruct (write_head_step (mloc 𝑚) e0 v o σ 𝓥2 In) as [?[??]];
      [apply WF|..|done| |].
    + inversion WRITE. subst. inversion WRITE0. by inversion PROMISE.
    + inversion WRITE. subst. inversion WRITE0. inversion PROMISE.
      destruct LALL as [? [? _]]. by eexists.
    + etrans; last apply Ext. inversion DRF. inversion LOCAL. by rewrite OTL.
    + by do 4 eexists.
  - inversion PStep; clear PStep; subst. simpl in *.
    destruct (update_head_step (mloc 𝑚) e1 e2 lit1 v2 orf or ow σ 𝓥2) as [?[?[??]]];
      [apply WF|done|apply WF|by inversion READ| |done..| | | |by do 4 eexists].
    + inversion READ=>EQ.
      change (σ.(mem)!!(𝑚.(mloc),𝑚.(mto)) = Some 𝑚.(mbase)) in IN.
      by rewrite memory_lookup_cell EQ in IN.
    + etrans; last by apply Ext.
      inversion DRF. inversion LOCAL. by rewrite OTL.
    + inversion READ. subst. exists 𝑚.(mto), 𝑚.(mbase). auto with congruence.
    + inversion READ. subst. inversion DRF. inversion LOCAL => t ???.
      assert (𝓥1.(cur).(pln) !! 𝑚.(mloc) ⊑ Some t); [|by eauto].
      etrans; [|done]. apply Ext.
  - inversion PStep; clear PStep; subst.
    destruct (update_head_step (mloc 𝑚1) e1 e2 lit1 v2 orf or ow σ 𝓥2) as [?[?[??]]];
      [apply WF|done|apply WF|by inversion READ| |done..| | | |by do 4 eexists].
    + inversion READ=>EQ.
      change (σ.(mem)!!(𝑚1.(mloc),𝑚1.(mto)) = Some 𝑚1.(mbase)) in IN.
      by rewrite memory_lookup_cell EQ in IN.
    + inversion READ. simpl. etrans; last by apply Ext.
      inversion DRFR. inversion LOCAL. by rewrite OTL.
    + inversion READ. subst. exists 𝑚1.(mto), 𝑚1.(mbase). done.
    + inversion READ. subst. inversion DRFR. inversion LOCAL => t ???.
      assert (𝓥1.(cur).(pln) !! 𝑚1.(mloc) ⊑ Some t); [|by eauto].
      etrans; [|done]. apply Ext.
  - destruct (acq_fence_head_step σ 𝓥2) as [?[??]]. by do 4 eexists.
  - destruct (rel_fence_head_step σ 𝓥2) as [?[??]]. by do 4 eexists.
  - destruct (sc_fence_head_step σ 𝓥2) as [?[??]]. by do 4 eexists.
Qed.

Lemma tview_mono_progress σ e 𝓥1 𝓥2 (Ext: 𝓥1 ⊑ 𝓥2) (In: 𝓥2 ∈ σ.(mem)) (WF: Wf σ) :
  reducible (e at 𝓥1) σ → reducible (e at 𝓥2) σ.
Proof.
  move => [[e' 𝓥'] [σ' [efs STEP]]].
  inversion STEP as [K [e0 𝓥0] [e0' 𝓥0'] DC1 DC2 HSTEP]. simpl in *.
  rewrite -fill_base_nopro in DC1. inversion DC1; subst; clear DC1.
  rewrite -fill_base_nopro in DC2. inversion DC2; subst; clear DC2.
  destruct (tview_mono_head_progress _ _ _ _ _ _ _ _ Ext In WF HSTEP)
    as [𝓥2' [σ2' [efs' [e2 HSTEP']]]].
  exists (fill K e2 at 𝓥2')%E, σ2', efs'. rewrite fill_base_nopro.
  econstructor; [done|apply fill_base_nopro|exact HSTEP'].
Qed.

Lemma tview_mono_head_preservation σ σ' e e' 𝓥1 𝓥2 𝓥2' efs2
  (Ext: 𝓥1 ⊑ 𝓥2) (WFs: Wf (mkCFG (mkLC 𝓥2 ∅) σ)) (IN1: 𝓥1 ∈ σ.(mem))
  (RED: reducible (e at 𝓥1) σ)
  (STEP: nopro_lang.head_step (e at 𝓥2) σ (e' at 𝓥2') σ' efs2):
  ∃ σ1' 𝓥1' efs1, 𝓥1' ⊑ 𝓥2'
    ∧ σ1' ⊑ σ'
    ∧ Forall2 (λ '(e1 at 𝓥1) '(e2 at 𝓥2), e1 = e2 ∧ 𝓥1 ⊑ 𝓥2)%E efs1 efs2
    ∧ nopro_lang.head_step (e at 𝓥1) σ (e' at 𝓥1') σ1' efs1.
Proof.
  inversion STEP; clear STEP; subst.
  { (* STATELESS REDUCTIONS *)
    inversion BaseStep; subst; clear BaseStep; (destruct efs2 as [|?[]]=>//; []);
      try (exists σ', 𝓥1, []; repeat (split; [done|]); by do 2 constructor).
    eexists σ', 𝓥1, [(_ at _)%E]; repeat (split; [done|]).
    split; [|by repeat constructor]. repeat constructor. destruct e; simpl in *.
    split; [congruence|]. do 2 inversion ForkViews; subst.
    constructor=>//=; apply Ext. }
  have VEX := program_step_tview_sqsubseteq _ _ _ _ PStep.
  inversion ExprStep; clear ExprStep; subst;
    (destruct efs2, c' as [[]σ']=>//; []); simpl in *;
    assert (prm = ∅) as -> by by apply (program_step_no_promise _ _ _ _ PStep).
  - (* ALLOC *)
    destruct (vmono_alloc_preservation _ _ _ _ _ _ _ Ext PStep)
      as [𝓥1' [Ext' STEP]].
    eexists σ', 𝓥1', []. do 3 (split; first done).
    eapply nopro_lang.impure_step in STEP; [done|by constructor..].
  - (* DEALLOC *)
    destruct (vmono_dealloc_preservation _ _ _ _ _ _ _ Ext PStep)
      as [𝓥1' [Ext' STEP]].
    { have HRED: head_reducible (Free (Lit n) (Lit l) at 𝓥1)%E σ.
      { apply prim_head_reducible; first done.
        apply ectxi_language_sub_redexes_are_values => Ki [e' V'] Hfill.
        rewrite /ectxi_language.fill_item /= /fill_item in Hfill.
        destruct Ki; repeat (simplify_eq/=; case_match=>//);
          rewrite /fill_item /= in Hfill; inversion Hfill; by eexists. }
      destruct RED as [?[?[? PRIM]]]. clear PStep.
      have HSTEP := head_reducible_prim_step _ _ _ _ _ HRED PRIM.
      clear PRIM HRED.
      inversion HSTEP; clear HSTEP; subst.
      - inversion BaseStep.
      - inversion ExprStep; clear ExprStep; subst. by eexists. }
    exists σ', 𝓥1', []. do 3 (split; first done).
    eapply nopro_lang.impure_step in STEP; [done|by constructor..].
  - (* READ *)
    destruct (vmono_read_preservation _ _ _ _ _ _ _ _ Ext PStep)
      as [𝓥1' [Ext' STEP]].
    { have HRED: head_reducible (Read o (Lit l) at 𝓥1)%E σ.
      { apply prim_head_reducible; first done.
        apply ectxi_language_sub_redexes_are_values => Ki [e' V'] Hfill.
        rewrite /ectxi_language.fill_item /= /fill_item in Hfill.
        destruct Ki; repeat (simplify_eq/=; case_match=>//);
          rewrite /fill_item /= in Hfill; inversion Hfill; by eexists. }
      destruct RED as [?[?[? PRIM]]].
      have HSTEP := head_reducible_prim_step _ _ _ _ _ HRED PRIM.
      clear PRIM HRED.
      inversion HSTEP; clear HSTEP; subst.
      - inversion BaseStep.
      - inversion ExprStep; clear ExprStep; subst. by eexists _, _. }
    exists σ', 𝓥1', []. do 3 (split; first done).
    eapply nopro_lang.impure_step in STEP; [done|by constructor..].
  - (* WRITE *)
    destruct (vmono_write_preservation _ _ _ _ _ _ _ _ Ext IN1
                                      (config_wf_global _ WFs) PStep)
      as [𝓥1' [σ2' [Ext' [Exts STEP]]]].
    { have HRED: head_reducible (Write o (Lit l) e0 at 𝓥1)%E σ.
      { apply prim_head_reducible; first done.
        apply ectxi_language_sub_redexes_are_values => Ki [e' V'] Hfill.
        rewrite /ectxi_language.fill_item /= /fill_item in Hfill.
        destruct Ki; repeat (simplify_eq/=; case_match=>//);
          rewrite /fill_item /= in Hfill; inversion Hfill; subst.
          by eexists. eexists.
            by rewrite /ectxi_language.to_val /= /nopro_lang.to_val /= H3. }
      destruct RED as [?[?[? PRIM]]]. clear PStep.
      have HSTEP := head_reducible_prim_step _ _ _ _ _ HRED PRIM.
      clear PRIM HRED.
      inversion HSTEP; clear HSTEP; subst.
      - inversion BaseStep.
      - inversion ExprStep; clear ExprStep; subst.
        assert (v = v0) as -> by congruence. by eexists. }
    exists σ2', 𝓥1', []. do 3 (split; first done).
    eapply nopro_lang.impure_step in STEP; [done|by constructor..].
  - (* FAILED CAS ~ READ *)
    have HRED: head_reducible (CAS (Lit l) e1 e2 orf or ow at 𝓥1)%E%E σ.
    { apply prim_head_reducible; first done.
      apply ectxi_language_sub_redexes_are_values => Ki [e' V'] Hfill.
      rewrite /ectxi_language.fill_item /= /fill_item in Hfill.
      destruct Ki; repeat (simplify_eq/=; case_match=>//);
        rewrite /fill_item /= in Hfill; inversion Hfill; subst.
      - by eexists.
      - eexists. by rewrite /ectxi_language.to_val /= /nopro_lang.to_val /= H4.
      - eexists. by rewrite /ectxi_language.to_val /= /nopro_lang.to_val /= H5. }
    destruct RED as [?[?[? PRIM]]].
    have HSTEP := head_reducible_prim_step _ _ _ _ _ HRED PRIM.
    clear PRIM HRED.
    inversion HSTEP; clear HSTEP; subst.
    { inversion BaseStep. }
    destruct (vmono_read_preservation _ _ _ _ _ _ _ _ Ext PStep)
      as [𝓥1' [Ext' STEP]].
    { inversion ExprStep; clear ExprStep; subst.
      - by eexists _, _.
      - inversion PStep0; clear PStep; subst; simpl in *.
        have HRLX: Relaxed ⊑ or by etrans; [exact H15|].
        destruct (read_step_relaxed _ _ _ _ _ orf HRLX READ) as [L2' READ'].
        eexists _, _. constructor; [done|by apply (drf_read_relaxed _ _ _ _ or)]. }
    exists σ', 𝓥1', []. do 3 (split; first done).
    eapply nopro_lang.impure_step in STEP; [done|econstructor=>//|by econstructor].
    inversion ExprStep; by assert (lit0 = lit1) as -> by congruence.
  - (* SUCCESSFUL CAS *)
    have HRED: head_reducible (CAS (Lit l) e1 e2 orf or ow at 𝓥1)%E σ.
    { apply prim_head_reducible; first done.
      apply ectxi_language_sub_redexes_are_values => Ki [e' V'] Hfill.
      rewrite /ectxi_language.fill_item /= /fill_item in Hfill.
      destruct Ki; repeat (simplify_eq/=; case_match=>//);
        rewrite /fill_item /= in Hfill; inversion Hfill; subst.
      - by eexists.
      - eexists. by rewrite /ectxi_language.to_val /= /nopro_lang.to_val /= H4.
      - eexists. by rewrite /ectxi_language.to_val /= /nopro_lang.to_val /= H5. }
    destruct RED as [?[?[? PRIM]]].
    have HSTEP := head_reducible_prim_step _ _ _ _ _ HRED PRIM.
    clear PRIM HRED.
    inversion HSTEP; clear HSTEP; subst.
    { inversion BaseStep. }
    destruct (vmono_update_preservation  _ _ _ _ _ _ _ _ _ _ Ext IN1
                                      (config_wf_global _ WFs) PStep)
      as [𝓥1' [σ2' [Ext' [Exts STEP]]]].
    { inversion ExprStep; clear ExprStep; subst.
      - have HRLX: Relaxed ⊑ or by etrans; [exact H15|].
        inversion PStep0; clear PStep; subst; simpl in *.
        destruct (read_step_relaxed _ _ _ _ _ or H15 READ) as [L2' READ'].
        eexists _, _. constructor; [done|by apply (drf_read_relaxed _ _ _ _ orf)].
      - inversion PStep0; clear PStep; subst; simpl in *.
        eexists _, _. by constructor. }
    exists σ2', 𝓥1', []. do 3 (split; first done).
    eapply nopro_lang.impure_step in STEP; [done|econstructor=>//|by econstructor].
    inversion ExprStep; by assert (lit0 = lit1) as -> by congruence.
  - (* ACQ Fence *)
    destruct (vmono_acq_fence_preservation _ _ _ _ _ Ext PStep)
      as [𝓥1' [Ext' STEP]].
    exists σ', 𝓥1', []. do 3 (split; first done).
    inversion PStep; clear PStep; subst.
    eapply nopro_lang.impure_step in STEP; [done|by econstructor..].
  - (* Rel Fence *)
    destruct (vmono_rel_fence_preservation _ _ _ _ _ Ext PStep)
      as [𝓥1' [Ext' STEP]].
    exists σ', 𝓥1', []. do 3 (split; first done).
    inversion PStep; clear PStep; subst.
    eapply nopro_lang.impure_step in STEP; [done|by econstructor..].
  - (* SC Fence *)
    destruct (vmono_sc_fence_preservation _ _ _ _ _ Ext PStep)
      as [𝓥1' [σ1' [Ext' [Exts STEP]]]].
    exists σ1', 𝓥1', []. do 3 (split; first done).
    eapply nopro_lang.impure_step in STEP; [done|by econstructor..].
Qed.
