From iris.bi Require Import monpred fractional.
From iris.base_logic Require Export lib.own.

From gpfsl Require Export history.
From gpfsl Require Import vprop.

Set Default Proof Using "Type".
Import uPred.

Section na_defs.
  Context `{noprolG Σ}.

  Program Definition own_loc_prim l q C : vProp Σ :=
    MonPred (λ V, ∃ rsa rsn ws,
              ⌜alloc_local l C V ∧ good_hist C ∧
              atr_local l rsa V ∧ atw_local l ws V ∧ na_local l rsn V⌝ ∗
              hist l q C ∗ atread l q rsa ∗ atwrite l q ws ∗ naread l q rsn)%I _.
  Next Obligation. solve_proper. Qed.

  Program Definition own_loc_na_def l q v : vProp Σ :=
    MonPred (λ V, ∃ t m, own_loc_prim l q {[t := m]} V
                         ∗ ⌜memval_val_rel m.(mval) v ∧ m.(mrel) ⊑ Some V⌝)%I _.
  Next Obligation. solve_proper. Qed.
  Definition own_loc_na_aux : seal (@own_loc_na_def). by eexists. Qed.
  Definition own_loc_na := unseal (@own_loc_na_aux).
  Definition own_loc_na_eq : @own_loc_na = _ := seal_eq _.

  Definition own_loc_def l q : vProp Σ :=
    ( ∃ t m, own_loc_prim l q {[t := m]})%I.
  Definition own_loc_aux : seal (@own_loc_def). by eexists. Qed.
  Definition own_loc := unseal (@own_loc_aux).
  Definition own_loc_eq : @own_loc = _ := seal_eq _.

  Definition own_loc_na_vec (l : loc) (q : Qp) (vl : list val) : vProp Σ :=
    ([∗ list] i ↦ v ∈ vl, own_loc_na (l >> i) q v)%I.
  Definition own_loc_vec (l : loc) (q : Qp) (n : nat) : vProp Σ :=
    ([∗ list] i ∈ seq 0 n, own_loc (l >> Z.of_nat i) q)%I.
End na_defs.

Notation "l ↦ v" := (own_loc_na l 1 v)
  (at level 20, format "l  ↦  v") : bi_scope.
Notation "l ↦ ?" := (own_loc l 1) (at level 20, format "l  ↦  ?")  : bi_scope.
Notation "l ↦∗ vl" := (own_loc_na_vec l 1 vl) (at level 20) : bi_scope.
Notation "l ↦∗: P " := (∃ vl, l ↦∗ vl ∗ P vl)%I
  (at level 20, format "l  ↦∗:  P") : bi_scope.

Notation "l ↦{ q } v" := (own_loc_na l q v)
  (at level 20, format "l  ↦{ q }  v") : bi_scope.
Notation "l ↦{ q } ?" := (own_loc l q) (at level 20, format "l  ↦{ q }  ?").
Notation "l ↦∗{ q } vl" := (own_loc_na_vec l q vl)
  (at level 20, q at level 50, format "l  ↦∗{ q }  vl") : bi_scope.
Notation "l ↦∗{ q }: P" := (∃ vl, l ↦∗{ q } vl ∗ P vl)%I
  (at level 20, q at level 50, format "l  ↦∗{ q }:  P") : bi_scope.

Typeclasses Opaque own_loc_na_vec own_loc_vec.

Section na_props.
  Context `{noprolG Σ}.

  (* Global Instance own_loc_prim_into_sep l q C V :
    IntoSep (own_loc_prim l q C V) ⌜alloc_local l C V ∧ Wf C⌝ (hist l q C) := _.
  Global Instance own_loc_prim_from_sep l q C V :
    FromSep (own_loc_prim l q C V) ⌜alloc_local l C V ∧ Wf C⌝ (hist l q C) := _. *)

  Global Instance own_loc_na_into_exist l q v V :
    IntoExist (own_loc_na l q v V)
              (λ t, ∃ m, own_loc_prim l q {[t := m]} V
                         ∗ ⌜memval_val_rel m.(mval) v ∧ m.(mrel) ⊑ Some V⌝)%I.
  Proof. rewrite own_loc_na_eq //. Qed.
  Global Instance own_loc_na_from_exist l q v V :
    FromExist (own_loc_na l q v V)
              (λ t, ∃ m, own_loc_prim l q {[t := m]} V
                         ∗ ⌜memval_val_rel m.(mval) v ∧ m.(mrel) ⊑ Some V⌝)%I.
  Proof. rewrite own_loc_na_eq //. Qed.

  Global Instance own_loc_into_exist l q :
    IntoExist (own_loc l q) (λ t, ∃ m, own_loc_prim l q {[t := m]})%I.
  Proof. rewrite own_loc_eq /own_loc_def //. Qed.
  Global Instance own_loc_from_exist l :
    FromExist (own_loc l q) (λ t, ∃ m, own_loc_prim l q {[t := m]})%I.
  Proof. rewrite own_loc_eq /own_loc_def //. Qed.

  Global Instance own_loc_prim_fractional l C : Fractional (λ q, own_loc_prim l q C).
  Proof.
    rewrite /Fractional =>p q. split=>V. iSplit.
    - iDestruct 1 as (rsa rsn ws)
        "(% & [Hp Hq] & [Ap Aq] & [Wp Wq] & [Np Nq])".
      iSplitL "Hp Ap Wp Np"; iExists rsa, rsn, ws; by iFrame.
    - iDestruct 1 as "[Hp Hq]".
      iDestruct "Hp" as (rsa1 rsn1 ws1) "[(%&%&%&%&%) (Hp&Ap&Wp&Np)]".
      iDestruct "Hq" as (rsa2 rsn2 ws2) "[(%&%&%&%&%) (Hq&Aq&Wq&Nq)]".
      iExists (rsa1 ∪ rsa2), (rsn1 ∪ rsn2), (ws1 ∪ ws2).
      iFrame "Hp Hq". iSplit.
      + iPureIntro. do 2 (split; [done|]). split; last split.
        by apply atr_local_join. by apply atw_local_join. by apply na_local_join.
      + iDestruct (atread_combine with "Ap Aq") as "$".
        iDestruct (naread_combine with "Np Nq") as "$".
        iDestruct (atwrite_agree with "[$Wp $Wq]") as %?. subst ws2.
        rewrite union_idemp_L.
        by iDestruct (atwrite_combine with "[$Wp $Wq]") as "$".
  Qed.

  Global Instance own_loc_prim_asfractional l q C :
    AsFractional (own_loc_prim l q C) (λ q, own_loc_prim l q C) q.
  Proof. split. done. apply _. Qed.

  Global Instance own_loc_prim_timeless l q C : Timeless (own_loc_prim l q C).
  Proof. rewrite /Timeless. iStartProof (iProp _). iIntros (V). iApply timeless. Qed.

  Lemma own_loc_prim_agree l q1 q2 C1 C2 :
    own_loc_prim l q1 C1 -∗ own_loc_prim l q2 C2 -∗ ⌜C1 = C2⌝.
  Proof.
    iStartProof (iProp _). iIntros (V) "Hp %% Hq".
    iDestruct "Hp" as (???) "[% [Hp _]]". iDestruct "Hq" as (???) "[% [Hq _]]".
    iApply hist_agree. iFrame.
  Qed.
  Lemma own_loc_prim_combine l q1 q2 C1 C2 :
    own_loc_prim l q1 C1 -∗ own_loc_prim l q2 C2 -∗ own_loc_prim l (q1+q2) C1.
  Proof.
    iIntros "H1 H2". iDestruct (own_loc_prim_agree with "H1 H2") as %<-. iFrame.
  Qed.

  Instance own_loc_na_fractional' l v V: Fractional (λ q, own_loc_na l q v V).
  Proof.
    unfold Fractional=>p q. iSplit.
    - iDestruct 1 as (??) "[[Hp Hq]%]". by iSplitL "Hp"; iExists _, _; iFrame.
    - iDestruct 1 as "[Hp Hq]".
      iDestruct "Hp" as (??) "[Hp %]". iDestruct "Hq" as (??) "[Hq %]".
      iExists _, _. iSplit; [iApply (own_loc_prim_combine with "Hp Hq")|done].
  Qed.
  Global Instance own_loc_na_fractional l v: Fractional (λ q, own_loc_na l q v).
  Proof. constructor => ?. by rewrite own_loc_na_fractional' monPred_at_sep. Qed.
  Global Instance own_loc_na_asfractional l q v :
    AsFractional (own_loc_na l q v) (λ q, own_loc_na l q v) q.
  Proof. split. done. apply _. Qed.

  Instance own_loc_na_timeless' l q v V: Timeless (own_loc_na l q v V).
  Proof. rewrite own_loc_na_eq. apply _. Qed.
  Global Instance own_loc_na_timeless l q v: Timeless (own_loc_na l q v).
  Proof. rewrite /Timeless. constructor => ?.
         by rewrite monPred_at_later monPred_at_except_0 own_loc_na_timeless'. Qed.

  Lemma own_loc_na_agree l q1 q2 v1 v2 :
    own_loc_na l q1 v1 -∗ own_loc_na l q2 v2 -∗ ⌜v1 = v2⌝.
  Proof.
    rewrite own_loc_na_eq. constructor. iIntros (?) "H1". iIntros (V->) "H2".
    iDestruct "H1" as (x [v1' ?]) "[H1 Hrel1]". iDestruct "Hrel1" as %[Hrel1 _].
    iDestruct "H2" as (x' [v2' ?]) "[H2 Hrel2]". iDestruct "Hrel2" as %[Hrel2 _].
    iDestruct (own_loc_prim_agree with "H1 H2") as %Eq.
    apply (f_equal (lookup x)) in Eq. rewrite lookup_singleton in Eq.
    destruct (decide (x = x')) as [<-|]; last by rewrite lookup_singleton_ne in Eq.
    rewrite lookup_singleton in Eq. iPureIntro. inversion Eq. subst. simpl in *.
    by destruct Hrel1; inversion Hrel2; subst.
  Qed.

  Lemma own_loc_na_combine l q1 q2 v1 v2 :
    own_loc_na l q1 v1 -∗ own_loc_na l q2 v2 -∗ own_loc_na l (q1+q2) v1.
  Proof.
    iIntros "H1 H2". iDestruct (own_loc_na_agree with "H1 H2") as %<-. iFrame.
  Qed.

  Lemma own_loc_na_own_loc l v :
    l ↦ v -∗ l ↦ ?.
  Proof.
    iStartProof (iProp _). iIntros (?). rewrite own_loc_na_eq own_loc_eq.
    iDestruct 1 as (t m) "[own ?]". iExists _,_. by iFrame.
  Qed.

  Global Instance own_loc_fractional l : Fractional (own_loc l).
  Proof.
    unfold Fractional=>p q. iSplit.
    - iDestruct 1 as (??) "[Hp Hq]". iSplitL "Hp"; iExists _,_; iFrame.
    - iDestruct 1 as "[Hp Hq]". iDestruct "Hp" as (??) "Hp". iDestruct "Hq" as (??) "Hq".
      iExists _,_. iApply (own_loc_prim_combine with "Hp Hq").
  Qed.
  Global Instance own_loc_asfractional l q :
    AsFractional (own_loc l q) (own_loc l) q.
  Proof. split. done. apply _. Qed.

  Global Instance own_loc_timeless l q : Timeless (l ↦{q} ?).
  Proof. rewrite own_loc_eq. apply _. Qed.

  Global Instance own_loc_na_vec_fractional l vl: Fractional (λ q, l ↦∗{q} vl)%I.
  Proof.
    intros p q. rewrite /own_loc_na_vec -big_sepL_sepL. do 3 f_equiv.
    by setoid_rewrite (fractional (Φ := λ q, own_loc_na _ q _)%I).
  Qed.
  Global Instance own_loc_na_vec_as_fractional l q vl:
    AsFractional (l ↦∗{q} vl) (λ q, l ↦∗{q} vl)%I q.
  Proof. split. done. apply _. Qed.

  Global Instance own_loc_na_vec_timeless l q vl : Timeless (l ↦∗{q} vl).
  Proof. rewrite /own_loc_na_vec. apply _. Qed.

  Global Instance own_loc_vec_timeless l q n : Timeless (own_loc_vec l q n).
  Proof. rewrite /own_loc_vec. apply _. Qed.

  Global Instance own_loc_vec_fractional l n: Fractional (λ q, own_loc_vec l q n)%I.
  Proof.
    intros p q. rewrite /own_loc_vec -big_sepL_sepL. do 3 f_equiv. apply fractional.
  Qed.
  Global Instance own_loc_vec_as_fractional l q n:
    AsFractional (own_loc_vec l q n) (λ q, own_loc_vec l q n)%I q.
  Proof. split. done. apply _. Qed.

  Lemma own_loc_na_vec_nil l q : l ↦∗{q} [] ⊣⊢ True.
  Proof. rewrite /own_loc_na_vec True_emp //. Qed.

  Lemma own_loc_na_vec_app l q vl1 vl2 :
    l ↦∗{q} (vl1 ++ vl2) ⊣⊢ l ↦∗{q} vl1 ∗ (l >> length vl1) ↦∗{q} vl2.
  Proof.
    rewrite /own_loc_na_vec big_sepL_app.
    do 2 f_equiv. intros k v. by rewrite shift_nat_assoc.
  Qed.

  Lemma own_loc_na_vec_singleton l q v : l ↦∗{q} [v] ⊣⊢ own_loc_na l q v.
  Proof. rewrite /own_loc_na_vec  /= shift_lblock_0 right_id //. Qed.

  Lemma own_loc_na_vec_cons l q v vl:
    l ↦∗{q} (v :: vl) ⊣⊢ own_loc_na l q v ∗ (l >> 1) ↦∗{q} vl.
  Proof.
    rewrite (own_loc_na_vec_app l q [v] vl) own_loc_na_vec_singleton //.
  Qed.

  Lemma own_loc_na_vec_agree l q1 q2 vl1 vl2 :
    length vl1 = length vl2 →
    l ↦∗{q1} vl1 -∗ l ↦∗{q2} vl2 -∗ ⌜vl1 = vl2⌝.
  Proof.
    revert l vl2. induction vl1 as [|v1 vl1 IH]; intros l [|v2 vl2] [=]; [by auto|].
    rewrite !own_loc_na_vec_cons. iIntros "[Hv1 Hvl1] [Hv2 Hvl2]".
    iDestruct (IH with "Hvl1 Hvl2") as %->; [done|].
      by iDestruct (own_loc_na_agree with "Hv1 Hv2") as %[=->].
  Qed.

  Lemma own_loc_na_vec_combine l q1 q2 vl1 vl2 :
    length vl1 = length vl2 →
    l ↦∗{q1} vl1 -∗ l ↦∗{q2} vl2 -∗ l ↦∗{q1+q2} vl1.
  Proof.
    iIntros (Hlen) "H1 H2".
    iDestruct (own_loc_na_vec_agree with "H1 H2") as %->; [done|]. iFrame.
  Qed.

  Global Instance own_loc_na_pred_fractional l (P : list val → vProp Σ):
    (∀ vl, Persistent (P vl)) → Fractional (λ q, l ↦∗{q}: P)%I.
  Proof.
    intros p q q'. iSplit.
    - iIntros "H". iDestruct "H" as (vl) "[[Hown1 Hown2] #HP]".
      iSplitL "Hown1"; eauto.
    - iIntros "[H1 H2]". iDestruct "H1" as (vl1) "[Hown1 #HP1]".
      iDestruct "H2" as (vl2) "[Hown2 #HP2]".
      set (ll := min (length vl1) (length vl2)).
      rewrite -[vl1](firstn_skipn ll) -[vl2](firstn_skipn ll) 2!own_loc_na_vec_app.
      iDestruct "Hown1" as "[Hown1 _]". iDestruct "Hown2" as "[Hown2 _]".
      assert (length (take ll vl1) = length (take ll vl2)).
      { rewrite !firstn_length /ll. lia. }
      iDestruct (own_loc_na_vec_agree with "Hown1 Hown2") as %Hl; [done|].
      rewrite -Hl. iCombine "Hown1" "Hown2" as "Hown".
      iExists (take ll vl1). iFrame.
      destruct (Min.min_spec (length vl1) (length vl2)) as [[Hne Heq]|[Hne Heq]].
      + iClear "HP2". rewrite take_ge; last by rewrite -Heq.
        rewrite drop_ge; first by rewrite app_nil_r. by rewrite -Heq.
      + iClear "HP1". rewrite Hl take_ge; last by rewrite -Heq.
        rewrite drop_ge; first by rewrite app_nil_r. by rewrite -Heq.
  Qed.
  Global Instance own_loc_na_pred_as_fractional l q (P : list val → vProp Σ):
    (∀ vl, Persistent (P vl)) → AsFractional (l ↦∗{q}: P) (λ q, l ↦∗{q}: P)%I q.
  Proof. split. done. apply _. Qed.

  Lemma own_loc_na_pred_combine l q1 q2 n (Φ : list val → vProp Σ) :
    (∀ vl, Φ vl -∗ ⌜length vl = n⌝) →
    l ↦∗{q1}: Φ ∗ l ↦∗{q2}: (λ vl, ⌜length vl = n⌝) ⊣⊢ l ↦∗{q1+q2}: Φ.
  Proof.
    intros Hlen. iSplit.
    - iIntros "[Hmt1 Hmt2]".
      iDestruct "Hmt1" as (vl) "[Hmt1 Hown]". iDestruct "Hmt2" as (vl') "[Hmt2 %]".
      iDestruct (Hlen with "Hown") as %?. subst. iExists _. iFrame "Hown".
      by iApply (own_loc_na_vec_combine with "Hmt1 Hmt2").
    - iIntros "Hmt". iDestruct "Hmt" as (vl) "[[Hmt1 Hmt2] Hown]".
      iDestruct (Hlen with "Hown") as %?.
      iSplitL "Hmt1 Hown"; iExists _; by iFrame.
  Qed.

  Lemma own_loc_na_pred_wand l q Φ1 Φ2 :
    l ↦∗{q}: Φ1 -∗ (∀ vl, Φ1 vl -∗ Φ2 vl) -∗ l ↦∗{q}: Φ2.
  Proof.
    iIntros "Hm Hwand". iDestruct "Hm" as (vl) "[Hm HΦ]". iExists vl.
    iFrame "Hm". by iApply "Hwand".
  Qed.

  Lemma own_loc_na_pred_iff_proper l q Φ1 Φ2 :
    □ (∀ vl, Φ1 vl ↔ Φ2 vl) -∗ □ (l ↦∗{q}: Φ1 ↔ l ↦∗{q}: Φ2).
  Proof.
    iIntros "#HΦ !#".
    iSplit; iIntros; iApply (own_loc_na_pred_wand with "[-]"); try done; [|];
    iIntros; by iApply "HΦ".
  Qed.

  Lemma own_loc_vec_nil l q : own_loc_vec l q 0 ⊣⊢ True.
  Proof. rewrite /own_loc_vec True_emp //. Qed.

  Lemma own_loc_vec_app l q n1 n2 :
    own_loc_vec l q (n1 + n2) ⊣⊢ own_loc_vec l q n1 ∗ own_loc_vec (l >> n1) q n2.
  Proof.
    rewrite /own_loc_vec -!(big_sepL_fmap (λ i:nat, _ >> i) (λ _ l, l ↦{q} ?))
            -big_sepL_app.
    f_equiv. apply reflexive_eq, list_eq=>i.
    destruct (decide (i < n1)%nat).
    - rewrite lookup_app_l ?fmap_length ?seq_length // !list_lookup_fmap
              !lookup_seq //. lia.
    - rewrite lookup_app_r ?fmap_length ?seq_length ?Nat.le_ngt // !list_lookup_fmap.
      destruct (decide (i < n1 + n2)%nat).
      + rewrite !lookup_seq /= ?shift_nat_assoc; repeat f_equal; lia.
      + rewrite !lookup_seq_ge //; lia.
  Qed.

  Lemma own_loc_vec_singleton l q : own_loc_vec l q 1 ⊣⊢ l ↦{q} ?.
  Proof. rewrite /own_loc_vec /= shift_lblock_0 right_id //. Qed.

  Lemma own_loc_vec_S l q n :
    own_loc_vec l q (S n) ⊣⊢ l ↦{q} ? ∗ own_loc_vec (l >> 1) q n.
  Proof. rewrite (own_loc_vec_app l q 1 n) own_loc_vec_singleton //. Qed.

End na_props.
