From gpfsl Require Export lang.

Coercion App : expr >-> Funclass.
Coercion nopro_lang.of_val : nopro_lang.val >-> nopro_lang.expr.
Coercion of_val : val >-> expr.

Coercion Var : string >-> expr.

Coercion BNamed : string >-> binder.
Notation "<>" := BAnon : binder_scope.

Notation "[ ]" := (@nil expr) : expr_scope.
Notation "[ x ]" := (@cons expr x%E (@nil expr)) : expr_scope.
Notation "[ x1 ; x2 ; .. ; xn ]" :=
  (@cons expr x1%E (@cons expr x2%E
        (..(@cons expr xn%E (@nil expr))..))) : expr_scope.

(* No scope for the values, does not conflict and scope is often not inferred
properly. *)
Notation "# l" := (LitV l%Z%V%L) (at level 8, format "# l").
Notation "# l" := (Lit l%Z%V%L) (at level 8, format "# l") : expr_scope.


(** Syntax inspired by Coq/Ocaml. Constructions with higher precedence come
    first. *)
Notation "'case:' e0 'of' el" := (Case e0%E el%E)
  (at level 102, e0, el at level 150) : expr_scope.
Notation "'if:' e1 'then' e2 'else' e3" := (If e1%E e2%E e3%E)
  (only parsing, at level 102, e1, e2, e3 at level 150) : expr_scope.
Notation "☠" := LitPoison : val_scope.

Notation "! e" := (Read NonAtomic e%E) (at level 9, format "! e") : expr_scope.
Notation "!ʳˡˣ e" := (Read Relaxed e%E) (at level 9, format "!ʳˡˣ e") : expr_scope.
Notation "!ᵃᶜ e" := (Read AcqRel e%E) (at level 9, format "!ᵃᶜ e") : expr_scope.

(* The unicode ← is already part of the notation "_ ← _; _" for bind. *)
Notation "e1 <- e2" := (Write NonAtomic e1%E e2%E)
  (at level 80) : expr_scope.
Notation "e1 <-ʳˡˣ e2" := (Write Relaxed e1%E e2%E)
  (at level 80) : expr_scope.
Notation "e1 <-ʳᵉˡ e2" := (Write AcqRel e1%E e2%E)
  (at level 80) : expr_scope.

Notation "e1 + e2" := (BinOp PlusOp e1%E e2%E)
  (at level 50, left associativity) : expr_scope.
Notation "e1 - e2" := (BinOp MinusOp e1%E e2%E)
  (at level 50, left associativity) : expr_scope.
Notation "e1 ≤ e2" := (BinOp LeOp e1%E e2%E)
  (at level 70) : expr_scope.
Notation "e1 = e2" := (BinOp EqOp e1%E e2%E)
  (at level 70) : expr_scope.
Notation "e1 +ₗ e2" := (BinOp OffsetOp e1%E e2%E)
  (at level 50, left associativity) : expr_scope.

Notation "'rec:' f xl := e" := (Rec f%bind xl%bind e%E)
  (at level 102, f, xl at level 1, e at level 200) : expr_scope.
Notation "'rec:' f xl := e" := (locked (RecV f%bind xl%bind e%E))
  (at level 102, f, xl at level 1, e at level 200) : val_scope.

Notation "λ: xl , e" := (Lam xl%bind e%E)
  (at level 102, xl at level 1, e at level 200) : expr_scope.
Notation "λ: xl , e" := (locked (LamV xl%bind e%E))
  (at level 102, xl at level 1, e at level 200) : val_scope.

Notation "e 'at' 𝓥" := (mkExpr e 𝓥) (at level 180) : expr_scope.
Notation "v 'at' 𝓥" := (mkVal v 𝓥) (at level 180) : val_scope.
