From iris.algebra Require Import auth.
From iris.bi Require Import monpred.

From gpfsl Require Import vprop.
From gpfsl Require Import history.

Implicit Types tid : thread_id.

Section RelAcq.
  Context `{histG Σ}.

  Definition acq_mod_def tid (P : vProp Σ) : iProp Σ :=
    (∃ 𝓥, own tid (◯ to_latT 𝓥) ∗ P 𝓥.(acq))%I.
  Definition acq_mod_aux : seal (@acq_mod_def). by eexists. Qed.
  Definition acq_mod := unseal (@acq_mod_aux).
  Definition acq_mod_eq : @acq_mod = _ := seal_eq _.

  Definition rel_mod_def tid (P : vProp Σ) : iProp Σ :=
    (∃ 𝓥, own tid (◯ to_latT 𝓥) ∗ P 𝓥.(frel))%I.
  Definition rel_mod_aux : seal (@rel_mod_def). by eexists. Qed.
  Definition rel_mod := unseal (@rel_mod_aux).
  Definition rel_mod_eq : @rel_mod = _ := seal_eq _.
End RelAcq.

Notation "△{ tid } P" := ⎡rel_mod tid P⎤%I
  (at level 20, right associativity, format "△{ tid }  P"): bi_scope.

Notation "▽{ tid } P" := ⎡acq_mod tid P⎤%I
  (at level 20, right associativity, format "▽{ tid }  P"): bi_scope.

Section RelAcqProp.
  Context `{histG Σ}.
  Local Notation vProp := (vProp Σ).
  Local Notation iProp := (iProp Σ).
  Implicit Types (P Q: vProp).

  Lemma rel_True_intro P tid : P ==∗ △{tid} True.
  Proof.
    iStartProof iProp. rewrite rel_mod_eq.
    iIntros (?) "_". iExists ∅. iSplitL""; [|done].
    rewrite (_: (◯ to_latT (∅: threadView) : authR _) = ε);
      [by iApply own_unit|done].
  Qed.

  Lemma rel_sep_objectively P Q tid :
    △{tid} P ∗ <obj> Q ⊢ △{tid} (P ∗ Q).
  Proof.
    iStartProof iProp. iIntros (?) "[rel obj]".
    rewrite rel_mod_eq monPred_at_objectively.
    iDestruct "rel" as (𝓥) "[own P]".
    iExists 𝓥. iFrame "own P". by iApply "obj".
  Qed.

  Lemma rel_mono_objectively P Q tid :
    △{tid} P ∗ <obj> (P -∗ Q) ⊢ △{tid} Q.
  Proof.
    iStartProof iProp. iIntros (?) "[rel obj]".
    rewrite rel_mod_eq monPred_at_objectively.
    iDestruct "rel" as (𝓥) "[own P]".
    iExists 𝓥. iFrame "own". by iApply "obj".
  Qed.

  Lemma rel_mono P Q tid :
    (P ⊢ Q : vProp) → (△{tid} P ⊢ △{tid} Q : vProp).
  Proof.
    iIntros (PQ). iStartProof iProp. iIntros (?) "rel".
    rewrite rel_mod_eq. iDestruct "rel" as (𝓥) "[own P]".
    iExists 𝓥. iFrame "own". by iApply PQ.
  Qed.

  Lemma rel_exist {A: Type} (Ψ: A → vProp) tid :
    ((△{tid} ∃ x, Ψ x) : vProp) ⊣⊢ ∃ x, △{tid} Ψ x.
  Proof.
    iStartProof iProp. iIntros (?). rewrite rel_mod_eq. iSplit.
    - iDestruct 1 as (?) "[own Ψ]". iDestruct "Ψ" as (?) "Ψ".
      iExists _,_. by iFrame.
    - iDestruct 1 as (??) "[own Ψ]". iExists _. iFrame "own". by iExists _.
  Qed.

  Lemma rel_forall_elim {A: Type} (Ψ: A → vProp) tid :
    ((△{tid} ∀ x, Ψ x) : vProp) ⊢ ∀ x, △{tid} Ψ x.
  Proof.
    iStartProof iProp. iIntros (?). rewrite rel_mod_eq.
    iDestruct 1 as (?) "[own Ψ]". iIntros (x).
    iExists _. iFrame "own". by iApply ("Ψ" $! x).
  Qed.

  Lemma rel_later_elim P tid :
    ((△{tid} ▷ P) : vProp) ⊢ ▷ △{tid} P.
  Proof.
    iStartProof iProp. iIntros (?). rewrite rel_mod_eq.
    iDestruct 1 as (?) "[own P]". iExists _. by iFrame "own".
  Qed.

  Lemma rel_objectively_elim P tid :
    ((△{tid} <obj> P) : vProp) ⊢ <obj> P.
  Proof.
    iStartProof iProp. iIntros (?). rewrite rel_mod_eq.
    iDestruct 1 as (?) "[_ P]". by rewrite 2!monPred_at_objectively.
  Qed.

  Lemma rel_embed_elim (P: iProp) tid :
    (△{tid} ⎡ P ⎤ : vProp) ⊢ ⎡ P ⎤.
  Proof.
    iStartProof iProp. iIntros (?). rewrite rel_mod_eq.
    by iDestruct (1) as (?) "[??]".
  Qed.

  Lemma rel_pure_elim (P: Prop) tid :
    (△{tid} ⌜P⌝ : vProp) ⊢ ⌜P⌝.
  Proof.
    iStartProof iProp. iIntros (?). rewrite rel_mod_eq.
    by iDestruct (1) as (?) "[??]".
  Qed.

  Lemma rel_sep_elim P Q tid :
    △{tid} (P ∗ Q) ⊢ △{tid} P ∗ (△{tid} Q : vProp).
  Proof.
    iStartProof iProp. iIntros (?). rewrite rel_mod_eq.
    iDestruct 1 as (𝓥) "[#own [P Q]]". iSplitL "P"; iExists 𝓥; by iFrame "own".
  Qed.

  Lemma rel_subjective P tid :
    △{tid} P ⊢ <subj> P.
  Proof.
    iStartProof iProp. iIntros (?). rewrite rel_mod_eq monPred_subjectively_eq.
    iDestruct 1 as (𝓥) "[#own P]". iExists _. iFrame "P".
  Qed.

  Lemma acq_exist {A: Type} (Ψ: A → vProp) tid :
    ((▽{tid} ∃ x, Ψ x) : vProp) ⊣⊢ ∃ x, ▽{tid} Ψ x.
  Proof.
    iStartProof iProp. iIntros (?). rewrite acq_mod_eq. iSplit.
    - iDestruct 1 as (?) "[own Ψ]". iDestruct "Ψ" as (?) "Ψ".
      iExists _,_. by iFrame.
    - iDestruct 1 as (??) "[own Ψ]". iExists _. iFrame "own". by iExists _.
  Qed.

  Lemma acq_forall_elim {A: Type} (Ψ: A → vProp) tid :
    ((▽{tid} ∀ x, Ψ x) : vProp) ⊢ ∀ x, ▽{tid} Ψ x.
  Proof.
    iStartProof iProp. iIntros (?). rewrite acq_mod_eq.
    iDestruct 1 as (?) "[own Ψ]". iIntros (x).
    iExists _. iFrame "own". by iApply ("Ψ" $! x).
  Qed.

  Lemma acq_later_elim P tid :
    ((▽{tid} ▷ P) : vProp) ⊢ ▷ ▽{tid} P.
  Proof.
    iStartProof iProp. iIntros (?). rewrite acq_mod_eq.
    iDestruct 1 as (?) "[own P]". iExists _. by iFrame "own".
  Qed.

  Lemma acq_objectively_elim P tid :
    ((▽{tid} <obj> P) : vProp) ⊢ <obj> P.
  Proof.
    iStartProof iProp. iIntros (?). rewrite acq_mod_eq.
    iDestruct 1 as (?) "[_ P]". by rewrite 2!monPred_at_objectively.
  Qed.

  Lemma acq_embed_elim (P: iProp) tid :
    (▽{tid} ⎡ P ⎤ : vProp) ⊢ ⎡ P ⎤.
  Proof.
    iStartProof iProp. iIntros (?). rewrite acq_mod_eq.
    by iDestruct (1) as (?) "[??]".
  Qed.

  Lemma acq_pure_elim (P: Prop) tid :
    (▽{tid} ⌜P⌝ : vProp) ⊢ ⌜P⌝.
  Proof.
    iStartProof iProp. iIntros (?). rewrite acq_mod_eq.
    by iDestruct (1) as (?) "[??]".
  Qed.

  Lemma acq_sep_elim P Q tid :
    ▽{tid} (P ∗ Q) ⊢ ▽{tid} P ∗ (▽{tid} Q : vProp).
  Proof.
    iStartProof iProp. iIntros (?). rewrite acq_mod_eq.
    iDestruct 1 as (𝓥) "[#own [P Q]]". iSplitL "P"; iExists 𝓥; by iFrame "own".
  Qed.

  Lemma acq_subjective P tid :
    ▽{tid} P ⊢ <subj> P.
  Proof.
    iStartProof iProp. iIntros (?). rewrite acq_mod_eq monPred_subjectively_eq.
    iDestruct 1 as (𝓥) "[#own P]". iExists _. iFrame "P".
  Qed.

  Lemma acq_mono_objectively P Q tid :
    ▽{tid} P ∗ <obj> (P -∗ Q) ⊢ ▽{tid} Q.
  Proof.
    iStartProof iProp. iIntros (?) "[acq obj]".
    rewrite acq_mod_eq monPred_at_objectively.
    iDestruct "acq" as (𝓥) "[own P]".
    iExists 𝓥. iFrame "own". by iApply "obj".
  Qed.

  Lemma acq_mono P Q tid :
    (P ⊢ Q : vProp) → (▽{tid} P ⊢ ▽{tid} Q : vProp).
  Proof.
    iIntros (PQ). iStartProof iProp. iIntros (?) "acq".
    rewrite acq_mod_eq. iDestruct "acq" as (𝓥) "[own P]".
    iExists 𝓥. iFrame "own". by iApply PQ.
  Qed.

  Global Instance rel_mod_ne tid : NonExpansive (rel_mod tid).
  Proof. rewrite rel_mod_eq. solve_proper. Qed.
  Global Instance rel_mod_proper tid : Proper ((≡) ==> (≡)) (rel_mod tid).
  Proof. apply (ne_proper _). Qed.
  Global Instance acq_mod_ne tid : NonExpansive (acq_mod tid).
  Proof. rewrite acq_mod_eq. solve_proper. Qed.
  Global Instance acq_mod_proper tid : Proper ((≡) ==> (≡)) (acq_mod tid).
  Proof. apply (ne_proper _). Qed.

  Global Instance rel_persistent (P : vProp) tid :
    Persistent P → Persistent ((△{tid} P)%I : vProp).
  Proof. rewrite rel_mod_eq. apply _. Qed.
  Global Instance acq_persistent (P : vProp) tid :
    Persistent P → Persistent ((▽{tid} P)%I : vProp).
  Proof. rewrite acq_mod_eq. apply _. Qed.

End RelAcqProp.
