From iris.algebra Require Import frac auth.
From iris.base_logic.lib Require Import invariants.
From iris.proofmode Require Import tactics.
From iris.bi Require Import fractional.
From gpfsl Require Export vprop.
From gpfsl.lang Require Import lattice_cmra.
Set Default Proof Using "Type".

Class view_invG Σ := {
  view_inv_inG :> inG Σ (authR (optionUR (fracLatR view_Lat)))
}.

Definition view_invΣ : gFunctors :=
  #[ GFunctor (constRF (authR (optionUR (fracLatR view_Lat)))) ].

Instance subG_view_invG Σ : subG (view_invΣ) Σ → view_invG Σ.
Proof. solve_inG. Qed.

Definition view_inv_pool_name := gname.

Section defs.
Context `{invG Σ, view_invG Σ}.
Notation vProp := (vProp Σ).

Program Definition view_tok_def (γ : view_inv_pool_name) (q : frac) : vProp :=
  MonPred (λ V : view, ∃ V0 : view, own γ (◯ Some (q, to_latT V0)) ∧ ⌜V0 ⊑ V⌝)%I _.
Next Obligation. solve_proper. Qed.
Definition view_tok_aux : seal (@view_tok_def). by eexists. Qed.
Definition view_tok := unseal (@view_tok_aux).
Definition view_tok_eq : @view_tok = _ := seal_eq _.

Definition view_inv_def γ (N: namespace) (P: vProp) : vProp :=
  ⎡ inv N (∃ V, view_tok γ 1 V ∨ own γ (● Some (1%Qp, to_latT V)) ∗ P V) ⎤%I.
Definition view_inv_aux : seal (@view_inv_def). by eexists. Qed.
Definition view_inv := unseal (@view_inv_aux).
Definition view_inv_eq : @view_inv = _ := seal_eq _.
End defs.

Section view_inv_props.
Context `{invG Σ, view_invG Σ}.
Notation vProp := (vProp Σ).

Global Instance view_tok_timeless γ q : Timeless (view_tok γ q).
Proof.
  rewrite view_tok_eq /Timeless. iStartProof (iProp _). iIntros (V). iApply timeless.
Qed.


Global Instance view_inv_contractive γ N : Contractive (view_inv γ N).
Proof. rewrite view_inv_eq. solve_contractive. Qed.
Global Instance view_inv_ne γ N : NonExpansive (view_inv γ N).
Proof. rewrite view_inv_eq. solve_proper. Qed.
Global Instance view_inv_proper γ N : Proper ((≡) ==> (≡)) (view_inv γ N).
Proof. apply (ne_proper _). Qed.

Global Instance view_inv_persistent γ N (P : vProp) :
  Persistent (view_inv γ N P).
Proof. rewrite view_inv_eq; apply _. Qed.

Global Instance view_tok_fractional γ : Fractional (λ q, view_tok γ q).
Proof.
  rewrite /Fractional view_tok_eq =>p q. split=>V. iSplit.
  - iDestruct 1 as (V1) "[[Hp Hq] %]". iSplitL "Hp"; iExists V1; by iFrame.
  - iDestruct 1 as "[Hp Hq]". iDestruct "Hp" as (V1) "[Hp %]".
    iDestruct "Hq" as (V2) "[Hq %]".
    iCombine "Hp" "Hq" as "H". rewrite lat_op_join'.
    iExists _. iFrame. iPureIntro. solve_lat.
Qed.

Global Instance view_tok_asfractional γ q :
  AsFractional (view_tok γ q) (λ q, view_tok γ q) q.
Proof. split. done. apply _. Qed.

Lemma view_tok_valid γ q:
  view_tok γ q -∗ ⌜(q ≤ 1)%Qc⌝.
Proof.
  iStartProof (iProp _). iIntros (?). rewrite view_tok_eq.
  iDestruct 1 as (V) "[own ?]".
  by iDestruct (own_valid with "own") as %[? ?]%auth_valid_discrete.
Qed.

Lemma view_tok_split_unit γ q q' :
  view_tok γ (q + q') -∗ view_tok γ q ∗ <obj> view_tok γ q'.
Proof.
  iStartProof (iProp _). iIntros (?). rewrite view_tok_eq.
  iDestruct 1 as (V) "[own ?]".
  rewrite -{1}(lat_join_bottom_rightid_L V) -lat_op_join' -frac_op' -pair_op.
  iDestruct "own" as "[own1 own2]".
  iSplitR "own2".
  - iExists _. by iFrame.
  - rewrite monPred_at_objectively. iIntros (?). iExists _. by iFrame "own2".
Qed.

Lemma view_inv_alloc_frac N E (q q': frac) (Eq: (q + q' = 1)%Qp):
  (|={E}=> ∃ γ, view_tok γ q ∗
    ∀ (P: vProp), ▷ P ={E}=∗ view_inv γ N P ∧ view_tok γ q')%I.
Proof.
  iStartProof (iProp _). iIntros (V).
  iMod (own_alloc (● Some (1%Qp, to_latT V) ⋅ ◯ Some (1%Qp, to_latT V)))
    as (γ) "[oA oN]"; [done|]. rewrite -{2}Eq. iDestruct "oN" as "[oN oN']".
  iExists γ. rewrite view_tok_eq. iSplitL "oN".
  { iExists _. by iFrame. }
  iIntros "!>" (P V' ?) "P".
  iCombine "oA" "oN'" as "oA".
  iMod (own_update _ _
        (● Some (1%Qp, to_latT (V ⊔ V')) ⋅ ◯ Some (q', to_latT (V ⊔ V')))
         with "oA") as "[oA tok]".
  { apply auth_update, option_local_update, frac_lat_local_update. }
  iSplitR "tok".
  - rewrite view_inv_eq monPred_at_embed. iApply inv_alloc.
    iExists _. iRight. by iFrame.
  - iExists _. iFrame. iPureIntro. solve_lat.
Qed.

Lemma view_inv_alloc_half N E :
  (|={E}=> ∃ γ, view_tok γ (1/2) ∗
    ∀ (P: vProp), ▷ P ={E}=∗ view_inv γ N P ∧ view_tok γ (1/2))%I.
Proof. apply view_inv_alloc_frac. by rewrite Qp_div_2. Qed.

Lemma view_inv_alloc N E :
  (|={E}=> ∃ γ, ∀ (P: vProp), ▷ P ={E}=∗ view_inv γ N P ∧ view_tok γ 1%Qp)%I.
Proof.
  iMod view_inv_alloc_half as (γ) "[tok Inv]".
  iExists γ. iIntros "!>" (P) "P". iFrame "tok". by iApply ("Inv" with "P").
Qed.

Lemma view_inv_alloc_open N E :
  ↑N ⊆ E →
  (|={E}=> ∃ γ, ∀ P, |={E,E ∖ ↑N}=> view_inv γ N P ∧ (▷ P ={E ∖ ↑N,E}=∗ view_tok γ 1%Qp))%I.
Proof.
  intros. iStartProof (iProp _). iIntros (V).
  iMod (own_alloc (● Some (1%Qp, to_latT V) ⋅ ◯ Some (1%Qp, to_latT V)))
    as (γ) "[oA oN]"; [done|].
  iExists γ. iIntros "!>" (P).
  iMod (inv_alloc_open N E (∃ V, view_tok γ 1 V ∨ own γ (● Some (1%Qp, to_latT V)) ∗ P V)%I)
        as "[Inv Closed]"; [done|].
  rewrite view_inv_eq monPred_at_embed. iFrame "Inv".
  iIntros "!>" (V1 Ext1) "P".
  iCombine "oA" "oN" as "oA".
  iMod (own_update _ _
          (● Some (1%Qp, to_latT (V ⊔ V1)) ⋅ ◯ Some (1%Qp, to_latT (V ⊔ V1)))
          with "oA") as "[oA tok]".
  { apply auth_update, option_local_update, frac_lat_local_update. }
  iMod ("Closed" with "[oA P]").
  + iExists _. iRight. iNext. by iFrame "oA".
  + iModIntro. rewrite view_tok_eq. iExists _. iFrame. iPureIntro. solve_lat.
Qed.

Lemma view_inv_dealloc γ N (P: vProp) E :
  ↑N ⊆ E → view_inv γ N P ∧ view_tok γ 1%Qp ={E}=∗ ▷ P.
Proof.
  intros. iStartProof (iProp _). iIntros (V) "[#Inv tok]".
  rewrite view_inv_eq view_tok_eq /view_inv_def.
  iDestruct "tok" as (V1) "[tok %]".
  iMod (inv_open with "Inv") as "[VI Closed]"; [done|].
  iDestruct "VI" as (V') "[>tok'|[>oA P]]".
  { rewrite view_tok_eq. iDestruct "tok'" as (V2) "[tok' _]".
    by iDestruct (own_valid_2 with "tok tok'") as %[VAL _]%auth_valid_discrete. }
  iDestruct (own_valid_2 with "oA tok")
    as %[[INCL|INCL]%Some_included VAL]%auth_valid_discrete_2;
    last by apply frac_lat_max in INCL.
  inversion INCL as [Eq1 Eq2%to_latT_inj%leibniz_equiv_iff]. subst V'.
  iFrame "P". iMod ("Closed" with "[tok]"); [|done].
  iExists V1. iLeft. rewrite view_tok_eq. iExists V1. by iFrame "tok".
Qed.

Lemma view_inv_acc_base_slow γ N P E :
  ↑N ⊆ E → view_inv γ N P ={E,E∖↑N}=∗
    ∀ Vb q,
    (monPred_in Vb → view_tok γ q) ={∅}=∗
    (monPred_in Vb → view_tok γ q) ∗
    ∃ V (R: vProp), (monPred_in V → ▷ P) ∗ R ∗
    (□ (view_tok γ 1%Qp -∗ R -∗ monPred_in V)) ∗
    (R -∗
      ((monPred_in Vb → view_tok γ q) -∗
      (monPred_in V → ▷ P) ={∅}=∗
      (monPred_in Vb → view_tok γ q) ∗ |={E∖↑N,E}=> True) ∧
      (∀ V' (Q: vProp), ⎡ view_tok γ q V' ⎤ -∗
            (⎡ view_tok γ q V' ⎤ ={E∖↑N}=∗ ⎡ (monPred_in V → ▷ P) V' ⎤ ∗ Q)
            ={E∖↑N,E}=∗ Q) ∧
      (view_tok γ 1%Qp -∗ monPred_in V ∧ |={E∖↑N,E}=> True)).
Proof.
  intros. iStartProof (iProp _). iIntros (?) "#Inv".
  rewrite view_inv_eq view_tok_eq.
  iMod (inv_open with "Inv") as "[VI Closed]"; [done|].
  iIntros "!>" (Vb q V ->) "tok". rewrite monPred_at_impl.
  iDestruct ("tok" $! (V ⊔ Vb) with "[%] [%]") as (V1) "[tok %]"; [solve_lat..|].
  iDestruct "VI" as (V') "[>tok'|[>oA P]]".
  { rewrite view_tok_eq. iDestruct "tok'" as (V2) "[tok' _]".
    iDestruct (own_valid_2 with "tok tok'") as %[VAL _]%auth_valid_discrete.
    rewrite /= frac_op' Qp_plus_comm in VAL.
    by apply Qp_not_plus_q_ge_1 in VAL. }
  iModIntro. iSplitL "tok".
  { iIntros (???). iExists _. iFrame. iPureIntro. solve_lat. }
  iExists V', ⎡ own γ (● Some (1%Qp, to_latT V')) ⎤%I.
  iSplitL "P". { iIntros (???). by iFrame "P". } iFrame "oA". iSplitL "".
  { iIntros (V2 ?) "!>". iDestruct 1 as (V3) "[tok %]".
    iIntros (V4 ?) "oA".
    iDestruct (own_valid_2 with "oA tok")
        as %[[INCL|INCL]%Some_included VAL]%auth_valid_discrete_2;
        last by apply frac_lat_max in INCL.
    inversion INCL as [Eq1 Eq2%to_latT_inj%leibniz_equiv_iff]. subst V'.
    iPureIntro. solve_lat. }
  iIntros (??) "oA". iSplit; last iSplit.
  - iIntros (V2 ?) "tok". rewrite monPred_at_impl.
    iDestruct ("tok" $! (V2 ⊔ Vb) with "[%] [%]") as (V3) "[tok %]"; [solve_lat..|].
    iIntros (V4 ?) "P". rewrite monPred_at_impl.
    iSpecialize ("P" $! (V' ⊔ V4) with "[%] [%]"); [solve_lat..|].
    iCombine "oA" "tok" as "oA".
    iMod (own_update _ _
           (● Some (1%Qp, to_latT (V' ⊔ V4)) ⋅ ◯ Some (q, to_latT (V3 ⊔ V4)))
           with "oA") as "[oA tok]".
    { apply auth_update, option_local_update, frac_lat_local_update. }
    iModIntro. iSplitL "tok".
    + iIntros (???). iExists _. iFrame "tok". iPureIntro. solve_lat.
    + iMod ("Closed" with "[oA P]"); [|done].
      iExists _. iRight. by iFrame.
  - iIntros (V3 Q V4 Ext4) "tok". iDestruct "tok" as (V3') "[tok %]".
    iIntros (??) "P".
    iCombine "oA" "tok" as "oA".
    iMod (own_update _ _
           (● Some (1%Qp, to_latT (V' ⊔ V3)) ⋅ ◯ Some (q, to_latT (V3' ⊔ V3)))
           with "oA") as "[oA tok]".
    { apply auth_update, option_local_update, frac_lat_local_update. }
    iMod ("P" with "[tok]") as "[P $]".
    { iExists _. iFrame "tok". iPureIntro. solve_lat. }
    rewrite monPred_at_impl.
    iSpecialize ("P" $! (V' ⊔ V3) with "[%] [%]"); [solve_lat..|].
    iMod ("Closed" with "[oA P]").
    + iExists _. iRight. by iFrame.
    + iPureIntro. solve_lat.
  - iIntros (V2 ?). iDestruct 1 as (V3) "[tok %]". iSplit.
    + iDestruct (own_valid_2 with "oA tok")
        as %[[INCL|INCL]%Some_included VAL]%auth_valid_discrete_2;
        last by apply frac_lat_max in INCL.
      inversion INCL as [Eq1 Eq2%to_latT_inj%leibniz_equiv_iff]. by subst V'.
    + iMod ("Closed" with "[tok]"); [|done].
      iExists V3. iLeft. rewrite view_tok_eq. iExists V3. by iFrame "tok".
Qed.

Lemma view_inv_acc_base_1 γ N P E :
  ↑N ⊆ E → view_inv γ N P ={E,E∖↑N}=∗
    ∀ Vb q,
    (monPred_in Vb → view_tok γ q) ={∅}=∗
    (monPred_in Vb → view_tok γ q) ∗ ∃ V, (monPred_in V → ▷ P) ∗
    (((monPred_in Vb → view_tok γ q) -∗
      (monPred_in V → ▷ P) ={∅}=∗
      (monPred_in Vb → view_tok γ q) ∗ |={E∖↑N,E}=> True) ∧
    (∀ V' (Q: vProp), ⎡ view_tok γ q V' ⎤ -∗
          (⎡ view_tok γ q V' ⎤ ={E∖↑N}=∗ ⎡ (monPred_in V → ▷ P) V' ⎤ ∗ Q)
          ={E∖↑N,E}=∗ Q) ∧
    (view_tok γ 1%Qp -∗ monPred_in V ∧ |={E∖↑N,E}=> True)).
Proof.
  intros. iIntros "#Inv".
  iMod (view_inv_acc_base_slow with "Inv") as "VI"; [done|].
  iIntros "!>" (Vb q) "I". iMod ("VI" with "I") as "[$ VI]".
  iDestruct "VI" as (V R) "(I & R & IR & VI)".
  iModIntro. iExists V. iFrame "I". iApply ("VI" with "R").
Qed.

Lemma view_inv_acc_base_2 γ N P q E :
  ↑N ⊆ E → view_inv γ N P ∧ view_tok γ q ={E,E∖↑N}=∗
    view_tok γ q ∗ ∃ V (R: vProp), (monPred_in V → ▷ P) ∗ R ∗
    (□ (view_tok γ 1%Qp -∗ R -∗ monPred_in V)) ∗
    (R -∗
    ((view_tok γ q -∗ (monPred_in V → ▷ P) ={E∖↑N,E}=∗ view_tok γ q) ∧
    (∀ V' (Q: vProp),
        ⎡ view_tok γ q V' ⎤ -∗
          (⎡ view_tok γ q V' ⎤ ={E∖↑N}=∗ ⎡ (monPred_in V → ▷ P) V' ⎤ ∗ Q)
          ={E∖↑N,E}=∗ Q) ∧
    (view_tok γ 1%Qp -∗ monPred_in V ∧ |={E∖↑N,E}=> True))).
Proof.
  intros. iIntros "[#Inv tok]".
  iDestruct (monPred_in_intro with "tok") as (Vb) "[#In tok]".
  iMod (view_inv_acc_base_slow with "Inv") as "VS"; [done|].
  iSpecialize ("VS" $! Vb with "[tok]").
  { iIntros "?". iApply (monPred_in_elim with "In tok"). }
  iMod (fupd_mask_mono with "VS") as "[tok VS]"; [set_solver|].
  iDestruct ("tok" with "In") as "$".
  iDestruct "VS" as (V R) "(P & R & IR & Closed)". iExists V, R; iFrame "P R IR".
  iIntros "!> R". iSpecialize ("Closed" with "R").
  iSplit; last iSplit.
  - rewrite bi.and_elim_l. iIntros "tok P".
    iSpecialize ("Closed" with "[$tok] P"); [by iIntros "_"|].
    iMod (fupd_mask_mono with "Closed") as "[tok >_]"; [set_solver|].
    by iDestruct ("tok" with "In") as "$".
  - by rewrite bi.and_elim_r bi.and_elim_l.
  - by rewrite bi.and_elim_r bi.and_elim_r.
Qed.

Lemma view_inv_acc_base γ N P q E :
  ↑N ⊆ E → view_inv γ N P ∧ view_tok γ q ={E,E∖↑N}=∗
    view_tok γ q ∗ ∃ V, (monPred_in V → ▷ P) ∗
    ((view_tok γ q -∗ (monPred_in V → ▷ P) ={E∖↑N,E}=∗ view_tok γ q) ∧
    (∀ V' (Q: vProp),
        ⎡ view_tok γ q V' ⎤ -∗
          (⎡ view_tok γ q V' ⎤ ={E∖↑N}=∗ ⎡ (monPred_in V → ▷ P) V' ⎤ ∗ Q)
          ={E∖↑N,E}=∗ Q) ∧
    (view_tok γ 1%Qp -∗ monPred_in V ∧ |={E∖↑N,E}=> True)).
Proof.
  intros. iIntros "[#Inv tok]".
  iDestruct (monPred_in_intro with "tok") as (Vb) "[#In tok]".
  iMod (view_inv_acc_base_1 with "Inv") as "VS"; [done|].
  iSpecialize ("VS" $! Vb with "[tok]").
  { iIntros "?". iApply (monPred_in_elim with "In tok"). }
  iMod (fupd_mask_mono with "VS") as "[tok VS]"; [set_solver|].
  iDestruct ("tok" with "In") as "$".
  iDestruct "VS" as (V) "[P Closed]". iExists V; iFrame "P". iModIntro.
  iSplit; last iSplit.
  - rewrite bi.and_elim_l. iIntros "tok P".
    iSpecialize ("Closed" with "[$tok] P"); [by iIntros "_"|].
    iMod (fupd_mask_mono with "Closed") as "[tok >_]"; [set_solver|].
    by iDestruct ("tok" with "In") as "$".
  - by rewrite bi.and_elim_r bi.and_elim_l.
  - by rewrite bi.and_elim_r bi.and_elim_r.
Qed.

Lemma view_inv_acc_consume γ N P q E :
  ↑N ⊆ E → view_inv γ N P ∧ view_tok γ q ={E,E∖↑N}=∗
    ∃ V, (monPred_in V → ▷ P) ∗ ((monPred_in V → ▷ P) ={E∖↑N,E}=∗ view_tok γ q).
Proof.
  iIntros (?) "VI".
  iMod (view_inv_acc_base with "VI") as "[tok VI]"; [done|].
  iDestruct "VI" as (V) "[P Closed]". iExists V. iFrame "P".
  rewrite bi.and_elim_l. iIntros "!> P".
  iApply ("Closed" with "tok P").
Qed.

End view_inv_props.
