From iris.base_logic.lib Require Export own.
From iris.base_logic.lib Require Import fancy_updates.
From iris.proofmode Require Import tactics monpred.
From iris.bi Require Export monpred.
From gpfsl.lang Require Export lang.

Canonical Structure lat_bi_index (Lat : latticeT) :=
  {| bi_index_type := Lat |}.
Instance lat_bi_index_bot (Lat : latticeT) (bot : Lat) :
  LatBottom bot → BiIndexBottom bot.
Proof. done. Qed.

Canonical Structure view_bi_index :=
  {| bi_index_type := view |}.

(* Types of view predicates. *)
Definition vProp Σ := monPred view_bi_index (uPredI (iResUR Σ)).
Definition vPropC Σ := monPredC view_bi_index (uPredI (iResUR Σ)).
Definition vPropI Σ := monPredI view_bi_index (uPredI (iResUR Σ)).
Definition vPropSI Σ := monPredSI view_bi_index (uPredSI (iResUR Σ)).

Hint Extern 10 (IsBiIndexRel _ _) => unfold IsBiIndexRel; solve_lat
            : typeclass_instances.

(* SOME big_op lemmas *)
From iris.bi Require Import big_op.

Lemma big_sepM_dup {PROP: sbi} `{Countable K} {A: Type} (φ: K → A → PROP) (m: gmap K A) k a q
  (In: m !! k = Some a)
  : ▷?q (φ k a -∗ φ k a ∗ φ k a) ∗ ▷?q ([∗ map] k' ↦ v' ∈ m, φ k' v')
    ⊢ ▷?q φ k a ∗ ▷?q ([∗ map] k' ↦ v' ∈ m, φ k' v').
Proof.
  iIntros "(Dup & res)". rewrite (big_sepM_delete φ _ _ _ In).
  iDestruct "res" as "(Hφd & res)".
  iDestruct ("Dup" with "Hφd") as "(Hφd1 & ?)".
  iCombine "Hφd1" "res" as "res".
  rewrite -(big_sepM_delete _ _ _ _ In).
  by iFrame.
Qed.


Section ExtraMonpred.
(* TODO: move to Iris *)
Lemma bi_wand_iff_sym `{PROP : bi} (P Q : PROP):
  (P ∗-∗ Q)%I ≡ (Q ∗-∗ P)%I.
Proof.
  apply bi.equiv_spec. split; iIntros "Eq"; iSplit; iIntros "?"; by iApply "Eq".
Qed.

Lemma bi_iff_sym `{PROP : bi} (P Q : PROP):
  (P ↔ Q)%I ≡ (Q ↔ P)%I.
Proof.
  apply bi.equiv_spec.
  split; rewrite /bi_iff; iIntros "Eq";
  iSplit; (by [rewrite bi.and_elim_r|rewrite bi.and_elim_l]).
Qed.

Lemma monPred_objectively_intuitionistically
  `{I : biIndex, PROP : bi} (P: monPred I PROP) :
  <obj> □ P ⊣⊢ □ <obj> P.
Proof.
  apply bi.equiv_spec; split; iIntros "#P"; iModIntro; iModIntro; iApply "P".
Qed.

Lemma monPred_objectively_subjectively
  `{I : biIndex, PROP : bi} (P Q: monPred I PROP) :
  <obj> (<subj> P -∗ Q) -∗ <subj> P -∗ <obj> Q.
Proof.
  iStartProof PROP. iIntros (?) "PQ". iIntros (??) "P".
  rewrite 2!monPred_at_objectively. iIntros (?). iApply "PQ".
  by rewrite 2!monPred_at_subjectively.
Qed.

End ExtraMonpred.

Section monPred_in_vprop.
  Context `{Σ: gFunctors}.
  Implicit Types (P Q: vProp Σ).

  Lemma monPred_in_sep P Q V:
    (monPred_in V → P ∗ Q) ⊣⊢ ((monPred_in V → P) ∗ (monPred_in V → Q)).
  Proof.
    split => V'. iSplit.
    - iIntros "PQ". rewrite monPred_at_impl.
      iDestruct ("PQ" $! (V ⊔ V') with "[%] [%]") as "[P Q]"; [solve_lat..|].
      iSplitL "P"; iIntros (???); iFrame.
    - iIntros "[P Q]". iIntros (V2 -> Ext).
      rewrite 2!monPred_at_impl.
      iDestruct ("P" $! V2 with "[%] [%]") as "$"; [done..|].
      by iApply "Q".
   Qed.

  Lemma monPred_in_exist {A: Type} (Ψ: A → vProp Σ) V :
    (monPred_in V → ∃ x, Ψ x) ⊣⊢ ∃ x, monPred_in V → Ψ x.
  Proof.
    split => V'. iSplit.
    - iIntros "Ex". rewrite monPred_at_impl.
      iDestruct ("Ex" $! (V ⊔ V') with "[%] [%]") as (x) "HΨ"; [solve_lat..|].
      iExists x. iIntros (???); iFrame.
    - iDestruct 1 as (x) "Ex". iIntros (V2 -> Ext). iExists x.
      by iApply "Ex".
  Qed.

  Lemma monPred_in_embed (P: iProp Σ) V:
    (monPred_in V → ⎡ P ⎤) ⊣⊢ (⎡ P ⎤ : vProp Σ).
  Proof.
    split => V'. iSplit.
    - iIntros "Em". rewrite monPred_at_impl.
      iDestruct ("Em" $! (V ⊔ V') with "[%] [%]") as "Em"; [solve_lat ..|].
      by rewrite monPred_at_embed.
    - iIntros "Em" (? _) "_". by rewrite monPred_at_embed.
  Qed.

  Lemma monPred_in_join P Q V:
    (monPred_in V → P) ∗ Q ⊢ (monPred_in V → P ∗ Q).
  Proof.
    split => ?. iIntros "[P Q]" (V' -> Ext). iFrame "Q". by iApply "P".
  Qed.

  Lemma monPred_in_elim_vProp P V:
    (monPred_in V → P) -∗ ⎡ ∃ V', P V' ⎤.
  Proof.
    split=> Vx. iIntros "H". iExists (V ⊔ Vx). iApply "H". iPureIntro. solve_lat.
  Qed.

  Lemma monPred_in_later_timeless `{invG Σ} P `{Timeless _ P} V E:
    (monPred_in V → ▷ P) ={E}=∗ (monPred_in V → P).
  Proof.
    split => V'. iIntros "P". rewrite monPred_at_impl.
    iDestruct ("P" $! (V ⊔ V') with "[%] [%]") as ">P"; [solve_lat..|].
    iIntros "!>" (???). iFrame "P".
  Qed.

End monPred_in_vprop.
